#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/AdlS593X.h"


void S593_ClearController(U32 dwPort)
{
    /*set MWTC a nonzero value*/
    OP_outd(dwPort,MWTC, 256);
    OP_outd(dwPort,MRTC, 256);
    /*reset mailbox flag and FIFO status*/
    OP_outd(dwPort,MCSR, 0x0E002200);
    /*clear all interrupt flags*/
    OP_outd(dwPort, INTCSR, 0x28BF0000);
}


void S593_StartODMA(U32 dwPort, PPCI_ADD pBufAdd)
{
    OP_REG opr;

    OP_outd(dwPort, MRAR, pBufAdd->dwAddPhy);
    /*Set transfer count*/
    OP_outd(dwPort, MRTC, pBufAdd->dwLength);
    /*Enable DMA complete interrupt*/
    opr.r = OP_inpd(dwPort,INTCSR);
    opr.intcsr.C_MRTC_IE = 1;
    OP_outd(dwPort, INTCSR, opr.r);
    /*Enable P2A DMA*/
    opr.r = OP_inpd(dwPort, MCSR);
    opr.mcsr.C_P2A_FLAG_RESET = 1;
    opr.mcsr.C_P2A_EN_TRANSFER = 1;
    OP_outd(dwPort, MCSR, opr.r);
}


void S593_StartIDMA(U32 dwPort, PPCI_ADD pBufAdd)
{
    OP_REG opr;

    OP_outd(dwPort, MWAR, pBufAdd->dwAddPhy);
    /*Set transfer count*/
    OP_outd(dwPort, MWTC, pBufAdd->dwLength);
    /*Enable DMA complete interrupt*/
    opr.r = OP_inpd(dwPort,INTCSR);
    opr.intcsr.C_MWTC_IE = 1;
    OP_outd(dwPort, INTCSR, opr.r);
    /*Enable A2P DMA*/
    opr.r = OP_inpd(dwPort, MCSR);
    opr.mcsr.C_A2P_FLAG_RESET = 1;
    opr.mcsr.C_A2P_EN_TRANSFER = 1;
    OP_outd(dwPort, MCSR, opr.r);
}


U32 S593_StopODMA(U32 dwPort)
{
    OP_REG opr;
    U32 dwLeft;

    dwLeft = OP_inpd(dwPort,MRTC);

    /*Disable and clear DMA complete interrupt*/
    opr.r = OP_inpd(dwPort,INTCSR);
    opr.intcsr.C_MRTC_IE = 0;
    OP_outd(dwPort,INTCSR, opr.r);
    /*Disable DMA transfer, Flush FIFO*/
    opr.r = OP_inpd(dwPort, MCSR);
    opr.mcsr.C_P2A_FLAG_RESET = 1;
    opr.mcsr.C_P2A_EN_TRANSFER = 0;
    OP_outd(dwPort, MCSR, opr.r);

    return dwLeft;
}


U32 S593_StopIDMA(U32 dwPort)
{
    OP_REG opr;
    U32 dwLeft;

    dwLeft = OP_inpd(dwPort,MWTC);

    /*Disable DMA complete interrupt*/
    opr.r = OP_inpd(dwPort,INTCSR);
    opr.intcsr.C_MWTC_IE = 0;
    OP_outd(dwPort,INTCSR, opr.r);
    /*Disable DMA transfer, Flush FIFO*/
    opr.r = OP_inpd(dwPort, MCSR);
    opr.mcsr.C_A2P_FLAG_RESET = 1;
    opr.mcsr.C_A2P_EN_TRANSFER = 0;
    OP_outd(dwPort, MCSR, opr.r);

    return dwLeft;

}


U32 S593_OutCountLeft(U32 dwPort)
{
    return OP_inpd(dwPort,MRTC);
}


U32 S593_InCountLeft(U32 dwPort)
{
    return OP_inpd(dwPort,MWTC);
}


BOOLEAN S593_OutFlushFIFO(U32 dwPort)
{
    OP_REG rOpM;
    U16 wIter;
    BOOLEAN fResult = FALSE;
    /*
     * Use a small loop to check if FIFO empty
     * When PCI transfer complete, there may still data in FIFO
     * So, do not disable Pass-Thru at once.
     *   Must wait FIFO empty
     */
    for(wIter=0; wIter<256; wIter++){
        rOpM.r = OP_inpd(dwPort, MCSR);
        if(rOpM.mcsr.S_P2A_EMPTY){
            fResult = TRUE;
            break;
        }
    }

    return fResult;
}


U32 S593_InFlushFIFO(U32 dwPort, U32 dwAccessCount, U32 dwBufHead1, U32 dwBufHead2)
{
    U32 dwAccessPos, dwLogicPos;
    U8 *pBuf;
    U16 wIter;
    OP_REG rOPM;

    if(dwBufHead1>dwBufHead2)
        dwLogicPos = (dwBufHead1-dwBufHead2);
    else
        dwLogicPos = 0;
    dwAccessPos = dwAccessCount-OP_inpd(dwPort,MWTC);
    dwLogicPos += dwAccessPos;
    pBuf = (U8 *)dwBufHead1;

    for(wIter=0; wIter<12; wIter++){
        if(dwAccessPos>=dwAccessCount){
            pBuf = (U8 *)dwBufHead2;
            dwAccessPos = 0;
            if(dwBufHead1>dwBufHead2)
                dwLogicPos = 0;
            else
                dwLogicPos = (dwBufHead2-dwBufHead1);
        }

        rOPM.r = OP_inpd(dwPort, MCSR);
        if(rOPM.mcsr.S_A2P_EMPTY)
            break;
        *((U32 *)(pBuf+dwAccessPos)) = OP_inpd(dwPort, FIFO);
        dwAccessPos += 4;
        dwLogicPos += 4;
    }

    return dwLogicPos;
}


void S593_StartIMB(U32 dwPort, U16 wMailBox, U16 wMbxByte)
{
    OP_REG opr;

    opr.r = OP_inpd(dwPort, INTCSR);
    opr.intcsr.C_IMB_IE = 1;
    opr.intcsr.C_IMB_MBX = wMailBox & 0x03;
    opr.intcsr.C_IMB_BYTE = wMbxByte & 0x03;
    OP_outd(dwPort, INTCSR, opr.r);
}


void S593_StartOMB(U32 dwPort, U16 wMailBox, U16 wMbxByte)
{
    OP_REG opr;

    opr.r = OP_inpd(dwPort, INTCSR);
    opr.intcsr.C_OMB_IE = 1;
    opr.intcsr.C_OMB_MBX = wMailBox & 0x03;
    opr.intcsr.C_OMB_BYTE = wMbxByte & 0x03;
    OP_outd(dwPort, INTCSR, opr.r);
}


void S593_StopIMB(U32 dwPort)
{
    OP_REG opr;

    opr.r = OP_inpd(dwPort, INTCSR);
    opr.intcsr.C_IMB_IE = 0;
    OP_outd(dwPort, INTCSR, opr.r);
}


void S593_StopOMB(U32 dwPort)
{
    OP_REG opr;

    opr.r = OP_inpd(dwPort, INTCSR);
    opr.intcsr.C_OMB_IE = 0;
    OP_outd(dwPort, INTCSR, opr.r);
}
