#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/siginfo.h> //for POLL_IN
//#include <linux/config.h>

#ifdef CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#if defined(__SMP__)
#include <linux/smp.h>
#endif

#if defined(MODVERSIONS)
#if defined(__SMP__)
#include <linux/modversions-smp.h>
#else
#include <linux/modversions.h>
#endif
#endif

#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,13)
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif


void clear_fasync_queue(struct fasync_struct **ppfa);


void AdlOpenDbx(register PPCI_DBX pDbx)
{
    if(!pDbx)
        return;

    pDbx->dwAccessCount = 0L;
    pDbx->dwAccessPos = 0L;
    pDbx->xFlag.w = 0;
    pDbx->hMapView[0] = pDbx->hMapView[1] = NULL;
    //01/14/08
    pDbx->dwTimeOut = 0;
}


/*
 * close event and view when device close 
 * prevent accident access this while device closed 
 */
//do not need now!
void AdlCloseDbx(register PPCI_DBX pDbx)
{
    if(!pDbx)
        return;
    
    pDbx->xFlag.f.Running = 0;
    if(pDbx->hMapView[0]){
        pDbx->hMapView[0] = pDbx->hMapView[1] = NULL;
    }
    //01/14/08
    pDbx->dwTimeOut = 0;
}


/*DeferProcCall*/
void AdlDpcDbx(register PPCI_DBX pDbx)
{
#ifdef __SMP__
    unsigned long irq_flags;
#endif

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    if(pDbx->xFlag.f.SyncMode){
        /*SyncMode ==> Complete IRP, stop running*/
        pDbx->xFlag.f.Running = 0;
    }
    else{
        /*AsyncMode ==> */
        if(pDbx->xFlag.f.DblBuf){
        /*DoubleBufferMode ==> */
            /*reset access position*/
            pDbx->dwAccessPos = 0L;
            /*change buffer index*/
            if(pDbx->xFlag.f.DrvHfRdyH){
                pDbx->xFlag.f.HalfRdyL = 0;
                pDbx->xFlag.f.HalfRdyH = 1;
                pDbx->xFlag.f.DrvHfRdyH = 0;
            }
            else{
                pDbx->xFlag.f.HalfRdyL = 1;
                pDbx->xFlag.f.HalfRdyH = 0;
                pDbx->xFlag.f.DrvHfRdyH = 1;
            }
        }
        else{
        /*OneShot Mode ==> */
            if(pDbx->wIterations==1){
            /*One Iteration ==> */
                pDbx->xFlag.f.Running = 0;
            }
            else{
            /*Multiple Iterations ==> */
                /*decrement iteration count*/
                pDbx->dwAccessPos = 0L;
                //yuan modify 03/20/02
                if(pDbx->wIterations)
                    pDbx->wIterations--;
            }
        }
    }
    
#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
#endif
}


void AdlCloseDsen(register PPCI_DSEN pDsen)
{
    if(!pDsen)
        return;

    if(pDsen->async_queue1!=NULL)
        clear_fasync_queue(&(pDsen->async_queue1));
    if(pDsen->async_queue2!=NULL)
        clear_fasync_queue(&(pDsen->async_queue2));
}


/*Delete DSEN*/
void AdlDropDsen(register PPCI_DSEN pDsen)
{
    if(!pDsen)
        return;
        
    if(pDsen->async_queue1!=NULL)
        clear_fasync_queue(&(pDsen->async_queue1));
    if(pDsen->async_queue2!=NULL)
        clear_fasync_queue(&(pDsen->async_queue2));

    kfree((void *)pDsen);
}


/*New DSEN and Initialize*/
PPCI_DSEN AdlSetupDsen()
{
    PPCI_DSEN pDsen;
    pDsen = (PPCI_DSEN)kmalloc(sizeof(PCI_DSEN), GFP_KERNEL);
    if(!pDsen)
        return NULL;

    //06/09/00 jeffrey, initial the async_queue1 pointer
    pDsen->async_queue1 = NULL;
    pDsen->async_queue2 = NULL;
    
#ifdef __SMP__
    //KeInitializeSpinLock(&pDsen->kSpinLock);
    spin_lock_init(&pDsen->kSpinLock);
#endif

    return pDsen;
}


/*DeferProcCall ....Wakeup event*/
//the delivery of SIGEVENT1 is put into bottom_half of each device
void AdlDpcDsen(register PPCI_DSEN pDsen)
{
    if(!pDsen)
        return;
}


void adl_kill_fasync(struct fasync_struct *fa, int sig)
{
    struct fown_struct *fown;
    struct fasync_struct *temp_fa;
    int i;

    i = 0;
    temp_fa = fa;
    while(temp_fa){
        fown = &temp_fa->fa_file->f_owner;
        if(fown->pid){
            fown->signum = sig; //replace the fctnl(SIGSET...)
        }
        temp_fa = temp_fa->fa_next;
    }

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
    kill_fasync(&fa, sig, POLL_IN);
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
    kill_fasync(fa, sig);
#endif
}


void clear_fasync_queue(struct fasync_struct **ppfa)
{
    struct fasync_struct *pfa, **fp;

    for(fp=ppfa; (pfa=*fp)!=NULL; fp=ppfa){
        //remove the struct file from fa and free the memory
        fasync_helper(pfa->fa_fd, pfa->fa_file, 0, ppfa);
        //no error will occur ! no more condition check!
    }
    //at the end, the *(ppfa)(or pDsen->async_queue1) will be set to NULL);
}
