#include <linux/pci.h>
#include <asm/io.h>
//#include <linux/config.h>

#ifdef CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#if defined(__SMP__)
#include <linux/smp.h>
#endif

#if defined(MODVERSIONS)
#if defined(__SMP__)
#include <linux/modversions-smp.h>
#else
#include <linux/modversions.h>
#endif
#endif

#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/Adlp9030.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,13)
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

#if !defined(MAP_NR)
#define MAP_NR(a) virt_to_page(a)
#endif

#if !defined(mem_map_reserve)
#define mem_map_reserve(page) SetPageReserved(page)
#endif

#if !defined(mem_map_unreserve)
#define mem_map_unreserve(map) ClearPageReserved(map)
#endif

extern DAS_IOT_CARD_DESP iniCardDesp;
extern PCI_DEVEXT DevExtension[32];

#ifndef DAS_ERR_RESOURCE_SETUP
#define DAS_ERR_RESOURCE_SETUP -204
#endif


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)

void find_valid_port_addr(PPCI_DEVEXT pDevExt, struct pci_dev *pciptr, int port_no)
{
    int i, port_i;

    for(port_i=0, i=0; port_i<DEVICE_COUNT_RESOURCE&&i<port_no; port_i++){
        if(pci_resource_start(pciptr, port_i)!=0x0000){
            //base_address is assigned for some function
            if(pci_resource_flags(pciptr, port_i)&IORESOURCE_IO){
                //IO SPACE
                pDevExt->dwPortAddr[i] = pci_resource_start(pciptr, port_i);
                pDevExt->dwPortLen[i] = pci_resource_len(pciptr, port_i);
                i++;
            }
            /*
            else if(pci_resource_flags(pciptr, port_i)&IORESOURCE_MEM){
                // MEMORY SPACE
                MEMORY_SPACE = pci_resource_start(pciptr, port_i);
                MEMORY_LENGTH = pci_resource_len(pciptr, port_i);
                i++;
            }
            */
        }
    }
    pDevExt->wPortCnt = i;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)

void AdlInitModule(PCI_DRVEXT *pdrv_ext)
{
    int vi;
    DAS_IOT_FUNC_DESP *pFdsp;

    memset(pdrv_ext, '\0', sizeof(PCI_DRVEXT));
    memcpy(&(pdrv_ext->CardDesp), &iniCardDesp, sizeof(DAS_IOT_CARD_DESP));

    pdrv_ext->wPacerMin[DAS_FUNC_AI] = 2;
    pdrv_ext->fTimerRoutine = FALSE;

    pFdsp = pdrv_ext->CardDesp.mFunc;
    for(vi=0; vi<DAS_FUNC_COUNT; vi++, pFdsp++){
        if(pFdsp->wArraySize>DAS_CHANNEL_GAIN_COUNT)
            pFdsp->wArraySize = DAS_CHANNEL_GAIN_COUNT;
    }

    return;
}

int AdlInitDevice(U16 VendorID, U16 DeviceID, U16 PciController, U16 wPort_Count, PPCI_DEVEXT pDevExt, struct pci_dev *pciptr, int bd)
{
    U32 modBaseAddr;
    U16 pci_command, subsystem_id, subvendor_id;
    U8 modLatency;

    if(!pciptr)
        return DAS_ERR_RESOURCE_SETUP;

    pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
    subvendor_id = pciptr->subsystem_vendor;
    subsystem_id = pciptr->subsystem_device;

    switch(PciController){
        case PCI_CTRL_S593:
            if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)){
                if((pci_command&0x4)!=4)
                    pci_command |= 4;
                pci_write_config_word(pciptr, PCI_COMMAND, pci_command);
                pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
                pci_read_config_byte(pciptr, PCI_REVISION_ID, &(pDevExt->wRevId));
                pDevExt->wBus = pciptr->bus->number;
                pDevExt->wSlot = PCI_SLOT(pciptr->devfn);
                pDevExt->wFunc = PCI_FUNC(pciptr->devfn);
                pDevExt->Irq_no = pciptr->irq;
                find_valid_port_addr(pDevExt, pciptr, wPort_Count);
                if(pci_enable_device(pciptr)){
                    printk("[AMCC5933] pci_enable_device failed\n");
                    return DAS_ERR_RESOURCE_SETUP;
                }
                pDevExt->Irq_no = pciptr->irq;
                pDevExt->wPortCnt = wPort_Count;
                pDevExt->wPciController = PciController;
                pDevExt->wCard = bd;
                AdlSetupBuffer(pDevExt);
            }
            break;
        case PCI_CTRL_PLX90:
            if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9050)&&(subsystem_id==DeviceID))||
                ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID))||
                ((pciptr->vendor==VendorID)&&(pciptr->device==0x7224)&&(DeviceID==0x7248))||
                ((pciptr->vendor==VendorID)&&(pciptr->device==0x7348)&&(DeviceID==0x7396)))
            {
                pci_read_config_byte(pciptr, PCI_REVISION_ID, &(pDevExt->wRevId));
                if(subsystem_id==0x63A8){
                    pDevExt->dwDevRA = 0x63A8;
                }
                else if(subsystem_id==0xC111){
                    pDevExt->wRevId +=1;
                }
                if((pci_resource_start(pciptr, 1)&0x0080)&&(pDevExt->wRevId<2)){
                    modBaseAddr = pci_resource_start(pciptr, 5);
                    if(modBaseAddr==0x00)
                        modBaseAddr = (pci_resource_start(pciptr, 2)+0x100)&0xFF00;
                    else
                        modBaseAddr = modBaseAddr&0xfffc;
                    pci_write_config_dword(pciptr, PCI_BASE_ADDRESS_1, (modBaseAddr|PCI_BASE_ADDRESS_SPACE_IO));
                    pci_resource_start(pciptr, 1) = modBaseAddr;
                    pci_resource_end(pciptr, 1) = modBaseAddr+0x7f;
                }
                if(DeviceID==0x7396) //this may be ignored
                    DeviceID = pciptr->device;
                /*additional check is needed*/
                pci_read_config_byte(pciptr, PCI_REVISION_ID,  &(pDevExt->wRevId));
                pDevExt->wBus = pciptr->bus->number;
                pDevExt->wSlot = PCI_SLOT(pciptr->devfn);
                pDevExt->wFunc = PCI_FUNC(pciptr->devfn);
                pDevExt->Irq_no = pciptr->irq;
                find_valid_port_addr(pDevExt, pciptr, wPort_Count);
                if(pci_enable_device(pciptr)){
                    printk("[PLX90] pci_enable_device failed\n");
                    return DAS_ERR_RESOURCE_SETUP;
                }
                pDevExt->Irq_no = pciptr->irq;
                pDevExt->wPortCnt = wPort_Count;
                pDevExt->wPciController = PciController;
                pDevExt->wCard = bd;
                AdlSetupBuffer(pDevExt);
            }
            break;
        case PCI_CTRL_P9080:
            if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9080)&&(subvendor_id==VendorID)&&(subsystem_id==DeviceID))||
                ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)))
            {
                modLatency = 0xff;
                pci_write_config_word(pciptr, PCI_LATENCY_TIMER, modLatency);
                if((pci_command&0x4)!=4)
                    pci_command |= 4;
                pci_write_config_word(pciptr, PCI_COMMAND, pci_command);
                pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
                pci_read_config_byte(pciptr, PCI_REVISION_ID, &(pDevExt->wRevId));
                pDevExt->wBus = pciptr->bus->number;
                pDevExt->wSlot = PCI_SLOT(pciptr->devfn);
                pDevExt->wFunc = PCI_FUNC(pciptr->devfn);
                pDevExt->Irq_no = pciptr->irq;
                find_valid_port_addr(pDevExt, pciptr, wPort_Count);
                if(pci_enable_device(pciptr)){
                    printk("[PLX9080] pci_enable_device failed\n");
                    return DAS_ERR_RESOURCE_SETUP;
                }
                pDevExt->Irq_no = pciptr->irq;
                pDevExt->wPortCnt = wPort_Count;
                pDevExt->wPciController = PciController;
                pDevExt->wCard = bd;
                AdlSetupBuffer(pDevExt);
            }
            break;
        case PCI_CTRL_P9030:
            if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)){
                pci_read_config_byte(pciptr, PCI_REVISION_ID, &(pDevExt->wRevId));
                pDevExt->wBus = pciptr->bus->number;
                pDevExt->wSlot = PCI_SLOT(pciptr->devfn);
                pDevExt->wFunc = PCI_FUNC(pciptr->devfn);
                pDevExt->Irq_no = pciptr->irq;
                find_valid_port_addr(pDevExt, pciptr, wPort_Count);
                if(pci_enable_device(pciptr)){
                    printk("[PLX9030] pci_enable_device failed\n");
                    return DAS_ERR_RESOURCE_SETUP;
                }
                pDevExt->Irq_no = pciptr->irq;
                pDevExt->wPortCnt = wPort_Count;
                pDevExt->wPciController = PciController;
                if((DeviceID==0x7256)||(DeviceID==0x7258)||(DeviceID==0x7260)){
                    U32 dwTmp, card_idx;
                    pci_read_config_dword(pciptr, PCI_BASE_ADDRESS_1, (int *)&modBaseAddr);
                    modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                    P9030_GetUserIoPin(modBaseAddr, 0L, &dwTmp);
                    card_idx = dwTmp;
                    P9030_GetUserIoPin(modBaseAddr, 1L, &dwTmp);
                    card_idx += dwTmp<<1;
                    P9030_GetUserIoPin(modBaseAddr, 2L, &dwTmp);
                    card_idx += dwTmp<<2;
                    P9030_GetUserIoPin(modBaseAddr, 3L, &dwTmp);
                    card_idx += dwTmp<<3;
                    if(DeviceID==0x7260){
                        P9030_GetUserIoPin(modBaseAddr, 8, &dwTmp);
                        card_idx += dwTmp<<4;
                        if(card_idx>=32)
                            pDevExt->wCard = bd;
                        else
                            pDevExt->wCard = card_idx;
                    }
                    else{
                        if(card_idx<16){
                            pDevExt->wCard = card_idx;
                        }
                        else{
                            pDevExt->wCard = bd;
                        }
                    }
                }
                else if((DeviceID==0x7452)||(DeviceID==0x7442)||(DeviceID==0x7443)||(DeviceID==0x7444)){
                    U32 dwTmp, card_idx;
                    pci_read_config_dword(pciptr, PCI_BASE_ADDRESS_1, (int *)&modBaseAddr
                    );
                    modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                    P9030_GetUserIoPin(modBaseAddr, 4L, &dwTmp);
                    card_idx = dwTmp;
                    P9030_GetUserIoPin(modBaseAddr, 5L, &dwTmp);
                    card_idx += dwTmp<<1;
                    P9030_GetUserIoPin(modBaseAddr, 6L, &dwTmp);
                    card_idx += dwTmp<<2;
                    P9030_GetUserIoPin(modBaseAddr, 7L, &dwTmp);
                    card_idx += dwTmp<<3;
                    if(card_idx<16)
                        pDevExt->wCard = card_idx;
                    else
                        pDevExt->wCard = bd;
                    //printk("card id is: %d\n", card_idx);
                }
                else{
                    pDevExt->wCard = bd;
                }
                AdlSetupBuffer(pDevExt);
            }
            break;
        default:
            break;
    }

    return bd;
}

/* Initialize the module - Register the character device */
int AdlFindDevice(U16 VendorID, U16 DeviceID, U16 PciController, U16 wPort_Count)
{
    U32 modBaseAddr;
    U16 pci_command, subsystem_id, subvendor_id;
    U8 modLatency;
    static struct pci_dev *pciptr;
    int pci_info_index = 0;

#ifdef CONFIG_PCI
    /*find the specified device and its configuration*/
    //if(pcibios_present()){
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
    for(pci_info_index=0; pci_info_index<32; ){
    #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
        if(PciController==PCI_CTRL_S593)
            pciptr = pci_get_subsys(VendorID, DeviceID, PCI_ANY_ID, PCI_ANY_ID,pciptr);
        else
            pciptr = pci_get_subsys(PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID,PCI_ANY_ID, pciptr);
    #else
        if(PciController==PCI_CTRL_S593)
            pciptr = pci_find_subsys(VendorID, DeviceID, PCI_ANY_ID, PCI_ANY_ID, pciptr);
        else
            pciptr = pci_find_subsys(PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, pciptr);
    #endif
        if(!pciptr)
            break;
        //printk("%x %x %x \n", pciptr->vendor, pciptr->device, pciptr->subsystem_device);
#else
    if(pcibios_present()){
        pci_for_each_dev(pciptr){
#endif

        pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
        //pci_read_config_word(pciptr, PCI_SUBSYSTEM_VENDOR_ID, &subvendor_id);
        subvendor_id = pciptr->subsystem_vendor;
        //pci_read_config_word(pciptr, PCI_SUBSYSTEM_ID, &subsystem_id);
        subsystem_id = pciptr->subsystem_device;

        switch(PciController){
            /*AMCC controller Match condition*/
            /*Vendor, Device, Command*/
            /*Command ==> bus master, memory access*/
            case PCI_CTRL_S593: //AMCC serial
                if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)){
                    //printk("command : %x\n", pci_command);
                    if((pci_command&0x4)!=4)
                       pci_command |= 4;
                    pci_write_config_word(pciptr, PCI_COMMAND, pci_command);
                    pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
                    //printk("command2 : %x\n", pci_command);
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus = pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                    //DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[AMCC5933] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    DevExtension[pci_info_index].wCard = pci_info_index;
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index ++;
                }
                break;
            /*PLX9050 controller Match condition*/
            /*Vendor, Device place in SubVendor, SubSystem*/
            case PCI_CTRL_PLX90:
                /*Alex 980925*/
                if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9050)&&(subsystem_id==DeviceID))||
                    ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID))||
                    /*Alex 990909 add support of PCI-7224*/
                    ((pciptr->vendor==VendorID)&&(pciptr->device==0x7224)&&(DeviceID==0x7248))||
                    /*Yuan 07/04/00 add support of PCI-7348*/
                    ((pciptr->vendor==VendorID)&&(pciptr->device==0x7348)&&(DeviceID==0x7396)))
                {
                    //yuan add for fix PLX9050 bug
                    //if(pciptr->base_address[1]&0x0080){
                    //yuan add 06/12/03
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    if(subsystem_id==0x63A8){
                        DevExtension[pci_info_index].dwDevRA = 0x63A8;
                    }
                    else if(subsystem_id==0xC111){
                        DevExtension[pci_info_index].wRevId += 1;
                    }
                    if((pci_resource_start(pciptr, 1)&0x0080)&&(DevExtension[pci_info_index].wRevId<2)){
                        modBaseAddr = pci_resource_start(pciptr, 5);
                        /*the PCI_BASE_ADDRESS_SPACE_IO bit will be removed*/
                        //if(!(modBaseAddr &0x0001))
                        if(modBaseAddr==0x00)
                            modBaseAddr = (pci_resource_start(pciptr, 2)+0x100)&0xFF00;
                        else
                            modBaseAddr = modBaseAddr&0xfffc;
                        pci_write_config_dword(pciptr, PCI_BASE_ADDRESS_1, (modBaseAddr|PCI_BASE_ADDRESS_SPACE_IO));
                        /*modify the PCI registry with the pci_dev structure*/
                        pci_resource_start(pciptr, 1) = modBaseAddr;
                        pci_resource_end(pciptr, 1) = modBaseAddr+0x7f;
                        //update the pci bios data , 00/07/20 jeffrey add.
                        //add the IO SPACE bit for the find_valid_port_addr()
                    }
                    if(DeviceID==0x7396) //this may be ignored
                        DeviceID = pciptr->device;
                    /*additional check is needed*/
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus = pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[PLX90] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    DevExtension[pci_info_index].wCard = pci_info_index;
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index++;
                }
                /*Alex 980925*/
                break;
            case PCI_CTRL_P9080:
                //printk("%x %x\n", subsystem_id, subvendor_id);
                if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9080)&&(subvendor_id==VendorID)&&(subsystem_id==DeviceID))||
                    //yuan add 10/08/99
                    ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)))
                {
                    //yuan add 10/08/99 to minimize DMA latency
                    modLatency = 0xff;
                    pci_write_config_word(pciptr, PCI_LATENCY_TIMER, modLatency);
                    //printk("command : %x\n", pci_command);
                    if((pci_command&0x4)!=4)
                        pci_command |= 4;
                    pci_write_config_word(pciptr, PCI_COMMAND, pci_command);
                    pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
                    //printk("command2 : %x\n", pci_command);
                    //update the pci bios data , 00/07/20 jeffrey add.
                    //pcibios_write_config_dword(pciptr->bus->number, pciptr->devfn, PCI_LATENCY_TIMER, modLatency);
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus = pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                    //DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[PLX9080] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    DevExtension[pci_info_index].wCard = pci_info_index;
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index++;
                 }
                 break;
            case PCI_CTRL_P9030:
                if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)){
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus = pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[PLX9030] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    if(DeviceID==0x7256||DeviceID==0x7258||DeviceID==0x7260){
                        U32 dwTmp, card_idx;
                        pci_read_config_dword(pciptr, PCI_BASE_ADDRESS_1, (int *)&modBaseAddr);
                        modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                        P9030_GetUserIoPin(modBaseAddr, 0L, &dwTmp);
                        card_idx = dwTmp;
                        P9030_GetUserIoPin(modBaseAddr, 1L, &dwTmp);
                        card_idx += dwTmp<<1;
                        P9030_GetUserIoPin(modBaseAddr, 2L, &dwTmp);
                        card_idx += dwTmp<<2;
                        P9030_GetUserIoPin(modBaseAddr, 3L, &dwTmp);
                        card_idx += dwTmp<<3;
                        if(DeviceID==0x7260){
                            P9030_GetUserIoPin(modBaseAddr, 8, &dwTmp);
                            card_idx += dwTmp<<4;
                            if(card_idx>=32)
                                DevExtension[pci_info_index].wCard = pci_info_index;
                            else
                                DevExtension[pci_info_index].wCard = card_idx;
                        }
                        else{
                            if(card_idx<16){
                                DevExtension[pci_info_index].wCard = card_idx;
                            }
                            else{
                                DevExtension[pci_info_index].wCard = pci_info_index;
                            }
                        }
                    }
                    else if((DeviceID==0x7452)||(DeviceID==0x7442)||(DeviceID==0x7443)||(DeviceID==0x7444)){
                        U32 dwTmp, card_idx;
                        pci_read_config_dword(pciptr, PCI_BASE_ADDRESS_1, (int *)&modBaseAddr);
                        modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                        P9030_GetUserIoPin(modBaseAddr, 4L, &dwTmp);
                        card_idx = dwTmp;
                        P9030_GetUserIoPin(modBaseAddr, 5L, &dwTmp);
                        card_idx += dwTmp<<1;
                        P9030_GetUserIoPin(modBaseAddr, 6L, &dwTmp);
                        card_idx += dwTmp<<2;
                        P9030_GetUserIoPin(modBaseAddr, 7L, &dwTmp);
                        card_idx += dwTmp<<3;
                        if(card_idx<16)
                            DevExtension[pci_info_index].wCard = card_idx;
                        else
                            DevExtension[pci_info_index].wCard = pci_info_index;
                    }
                    else{
                        DevExtension[pci_info_index].wCard = pci_info_index;
                    }
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index++;
                }
                break;
            default:
                break;
        } //end of switch(PciController)
    } //end of pci_for_each_dev()
    return pci_info_index; //number of the desired device

    #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    }else
        return -1;
    #endif

#else
    return -1;

#endif
}

#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)

/*Initialize the module - Register the character device*/
int AdlFindDevice(U16 VendorID, U16 DeviceID, U16 PciController, U16 wPort_Count)
{
    U32 modBaseAddr;
    U16 pci_command, subsystem_id, subvendor_id;
    U8 modLatency;
    static struct pci_dev *pciptr;
    int pci_info_index = 0;

#ifdef CONFIG_PCI
    //*find the specified device and its configuration*/
    //if(pcibios_present()){
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
    for(pci_info_index=0; pci_info_index<32; ){
    #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
        if(PciController == PCI_CTRL_S593)
            pciptr = pci_get_subsys(VendorID, DeviceID, PCI_ANY_ID, PCI_ANY_ID,pciptr);
        else
            pciptr = pci_get_subsys(PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID,PCI_ANY_ID, pciptr);
    #else
        if(PciController == PCI_CTRL_S593)
            pciptr = pci_find_subsys(VendorID, DeviceID, PCI_ANY_ID, PCI_ANY_ID, pciptr);
        else
            pciptr = pci_find_subsys(PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, pciptr);
    #endif
    if(!pciptr)
        break;
    //printk("%x %x %x \n", pciptr->vendor, pciptr->device, pciptr->subsystem_device);
#else
    if(pcibios_present()){
        pci_for_each_dev(pciptr){
#endif

        pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
        //pci_read_config_word(pciptr, PCI_SUBSYSTEM_VENDOR_ID, &subvendor_id);
        subvendor_id = pciptr->subsystem_vendor;
        //pci_read_config_word(pciptr, PCI_SUBSYSTEM_ID, &subsystem_id);
        subsystem_id = pciptr->subsystem_device;

        switch(PciController){
            /*AMCC controller Match condition*/
            /*Vendor, Device, Command*/
            /*Command ==> bus master, memory access*/
            case PCI_CTRL_S593: //AMCC serial
                if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)){
                    //printk("command : %x\n", pci_command);
                    if((pci_command&0x4)!=4)
                       pci_command |= 4;
                    pci_write_config_word(pciptr, PCI_COMMAND, pci_command);
                    pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
                    //printk("command2 : %x\n", pci_command);
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus= pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot= PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc= PCI_FUNC(pciptr->devfn);
                    //DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[AMCC5933] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    DevExtension[pci_info_index].wCard = pci_info_index;
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index ++;
                }
                break;
            /*PLX9050 controller Match condition*/
            /*Vendor, Device place in SubVendor, SubSystem*/
            case PCI_CTRL_PLX90:
                /*Alex 980925*/
                if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9050)&&(subsystem_id==DeviceID))||
                    ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID))||
                     /*Alex 990909 add support of PCI-7224*/
                    ((pciptr->vendor==VendorID)&&(pciptr->device==0x7224)&&(DeviceID==0x7248))||
                     /*Yuan 07/04/00 add support of PCI-7348*/
                    ((pciptr->vendor==VendorID)&&(pciptr->device==0x7348)&&(DeviceID==0x7396)))
                {
                    //yuan add for fix PLX9050 bug
                    //if(pciptr->base_address[1]&0x0080){
                    //yuan add 06/12/03
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    if(subsystem_id == 0x63A8) {
                        DevExtension[pci_info_index].dwDevRA = 0x63A8;
                    }
                    else if(subsystem_id==0xC111){
                        DevExtension[pci_info_index].wRevId += 1;
                    }
                    if((pci_resource_start(pciptr, 1)&0x0080)&&(DevExtension[pci_info_index].wRevId<2)){
                        modBaseAddr = pci_resource_start(pciptr, 5);
                        /*the PCI_BASE_ADDRESS_SPACE_IO bit will be removed*/
                        //if(!(modBaseAddr &0x0001))
                        if(modBaseAddr==0x00)
                            modBaseAddr = (pci_resource_start(pciptr, 2)+0x100)&0xFF00;
                        else
                            modBaseAddr = modBaseAddr&0xfffc;
                        pci_write_config_dword(pciptr, PCI_BASE_ADDRESS_1, (modBaseAddr|PCI_BASE_ADDRESS_SPACE_IO));
                        /*modify the PCI registry with the pci_dev structure*/
                        pci_resource_start(pciptr, 1) = modBaseAddr;
                        pci_resource_end(pciptr, 1) = modBaseAddr+0x7f;
                        //update the pci bios data , 00/07/20 jeffrey add.
                        //add the IO SPACE bit for the find_valid_port_addr()
                    }
                    if(DeviceID==0x7396) //this may be ignored
                        DeviceID = pciptr->device;
                    /*additional check is needed*/
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus = pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[PXL90] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    DevExtension[pci_info_index].wCard = pci_info_index;
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index ++;
                }
                /*Alex 980925*/
                break;
            case PCI_CTRL_P9080:
                //printk("%x %x\n", subsystem_id, subvendor_id);
                if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9080)&&(subvendor_id==VendorID)&&(subsystem_id==DeviceID))||
                    //yuan add 10/08/99
                    ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)))
                {
                    //yuan add 10/08/99 to minimize DMA latency
                    modLatency = 0xff;
                    pci_write_config_word(pciptr, PCI_LATENCY_TIMER, modLatency);
                    //printk("command : %x\n", pci_command);
                    if((pci_command&0x4)!=4)
                        pci_command|=4;
                    pci_write_config_word(pciptr, PCI_COMMAND, pci_command);
                    pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
                    //printk("command2 : %x\n", pci_command);
                    //update the pci bios data , 00/07/20 jeffrey add.
                    //pcibios_write_config_dword( pciptr->bus->number, pciptr->devfn, PCI_LATENCY_TIMER, modLatency );
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus = pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                    //DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[PLX9080] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    DevExtension[pci_info_index].wCard = pci_info_index;
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index ++;
                 }
                 break;
            case PCI_CTRL_P9030:
                if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)){
                    pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                    DevExtension[pci_info_index].wBus = pciptr->bus->number;
                    DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                    DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                    //printk("irq: %d\n", pciptr->irq);
                    if(pci_enable_device(pciptr)){
                        printk("[PLX9030] pci_enable_device failed\n");
                        return -1;
                    }
                    DevExtension[pci_info_index].Irq_no = pciptr->irq;
                    //printk("irq2: %d\n", pciptr->irq);
                    DevExtension[pci_info_index].wPortCnt = wPort_Count;
                    DevExtension[pci_info_index].wPciController = PciController;
                    if(DeviceID==0x7256||DeviceID==0x7258||DeviceID==0x7260){
                        U32 dwTmp, card_idx;
                        pci_read_config_dword( pciptr, PCI_BASE_ADDRESS_1, (int *)&modBaseAddr);
                        modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                        P9030_GetUserIoPin(modBaseAddr, 0L, &dwTmp);
                        card_idx = dwTmp;
                        P9030_GetUserIoPin(modBaseAddr, 1L, &dwTmp);
                        card_idx += dwTmp<<1;
                        P9030_GetUserIoPin( modBaseAddr, 2L, &dwTmp);
                        card_idx += dwTmp<<2;
                        P9030_GetUserIoPin(modBaseAddr, 3L, &dwTmp);
                        card_idx += dwTmp<<3;
                        if(DeviceID==0x7260){
                            P9030_GetUserIoPin(modBaseAddr, 8, &dwTmp);
                            card_idx += dwTmp<<4;
                            if(card_idx>=32)
                                DevExtension[pci_info_index].wCard = pci_info_index;
                            else
                                DevExtension[pci_info_index].wCard = card_idx;
                        }
                        else{
                            if(card_idx<16){
                                DevExtension[pci_info_index].wCard = card_idx;
                            }
                            else{
                                DevExtension[pci_info_index].wCard = pci_info_index;
                            }
                        }
                    }
                    else if((DeviceID==0x7452)||(DeviceID==0x7442)||(DeviceID==0x7443)||(DeviceID==0x7444)){
                        U32 dwTmp, card_idx;
                        pci_read_config_dword(pciptr, PCI_BASE_ADDRESS_1, (int *)&modBaseAddr);
                        modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                        P9030_GetUserIoPin(modBaseAddr, 4L, &dwTmp);
                        card_idx = dwTmp;
                        P9030_GetUserIoPin(modBaseAddr, 5L, &dwTmp);
                        card_idx += dwTmp<<1;
                        P9030_GetUserIoPin( modBaseAddr, 6L, &dwTmp);
                        card_idx += dwTmp<<2;
                        P9030_GetUserIoPin(modBaseAddr, 7L, &dwTmp);
                        card_idx += dwTmp<<3;
                        if(card_idx<16)
                            DevExtension[pci_info_index].wCard = card_idx;
                        else
                            DevExtension[pci_info_index].wCard = pci_info_index;
                    }
                    else{
                        DevExtension[pci_info_index].wCard = pci_info_index;
                    }
                    AdlSetupBuffer(&DevExtension[pci_info_index]);
                    pci_info_index++;
                }
                break;
            default:
                break;
        } //end of switch(PciController)
    } //end of pci_for_each_dev()
    return pci_info_index; //number of the desired device

    #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    }else
        return -1;
    #endif

#else
    return -1;

#endif
}

#endif


#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)

void find_valid_port_addr(PPCI_DEVEXT pDevExt, struct pci_dev *pciptr, int port_no)
{
    int i, port_i;
    U32 mask, baseAddr;

    for(port_i=0, i=0; port_i<6&&i<port_no; port_i++){
        //if((pciptr->base_address[port_i]&0x0001)==1){
        if(pciptr->base_address[port_i]!=0x0000){
            /*base_address is asigned for some function*/
            if(pciptr->base_address[port_i]&PCI_BASE_ADDRESS_SPACE){
                /*IO SPACE*/
                pDevExt->dwPortAddr[i] = (pciptr->base_address[port_i]&PCI_BASE_ADDRESS_IO_MASK);
                baseAddr = pciptr->base_address[port_i];
                pcibios_write_config_dword(pciptr->bus->number, pciptr->devfn, PCI_BASE_ADDRESS_0+(4*port_i), ~0);
                pcibios_read_config_dword(pciptr->bus->number, pciptr->devfn, PCI_BASE_ADDRESS_0+(4*port_i), (int *)&mask);
                pcibios_write_config_dword(pciptr->bus->number, pciptr->devfn, PCI_BASE_ADDRESS_0+(4*port_i), baseAddr); /*restore the origin ioport address*/
                //pDevExt->dwPortLen[i] = ~mask+1;
                pDevExt->dwPortLen[i] = ~mask+2;
                i++;
            }
            else{
                /*MEMORY SPACE*/
                //MEMORY_SPACE = (pciptr->base_address[port_i]&PCI_BASE_ADDRESS_MEM_MASK);
            }
        }
    }
}

/*Initialize the module - Register the character device*/
int AdlFindDevice(U16 VendorID, U16 DeviceID, U16 PciController, U16 wPort_Count)
{
    U32 modBaseAddr;
    U16 pci_command, subsystem_id, subvendor_id;
    U8 modLatency;
    struct pci_dev *pciptr;
    int pci_info_index = 0;

#ifdef CONFIG_PCI
    /*find the specified device and its configuration*/
    if(pcibios_present()){
        pciptr = pci_find_slot(0, 0);
        /* find the Slot0/Func0 in Bus 0 .... to get the pci_dev List
         * assum this should exist, may find bus0/slot1 , bus0/slot2 ..... , if slot0 absent
         */
        while(pciptr!=NULL){
            pci_read_config_word(pciptr, PCI_COMMAND, &pci_command);
            pci_read_config_word(pciptr, PCI_SUBSYSTEM_VENDOR_ID, &subvendor_id);
            pci_read_config_word(pciptr, PCI_SUBSYSTEM_ID, &subsystem_id);
            switch(PciController){
                /*AMCC controller Match condition*/
                /*Vendor, Device, Command*/
                /*Command ==> bus master, memory access*/
                case PCI_CTRL_S593: //AMCC serial
                    if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)&&(pci_command&0x04)){
                        pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                        DevExtension[pci_info_index].wBus = pciptr->bus->number;
                        DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                        DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                        DevExtension[pci_info_index].Irq_no = pciptr->irq;
                        find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                        DevExtension[pci_info_index].wPortCnt = wPort_Count;
                        DevExtension[pci_info_index].wPciController = PciController;
                        DevExtension[pci_info_index].wCard = pci_info_index;
                        AdlSetupBuffer(&DevExtension[pci_info_index]);
                        pci_info_index++;
                    }
                    break;
                /*PLX9050 controller Match condition*/
                /*Vendor, Device place in SubVendor, SubSystem*/
                case PCI_CTRL_PLX90:
                    /*Alex 980925*/
                    if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9050)&&(subsystem_id==DeviceID))||
                        ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID))||
                        /*Alex 990909 add support of PCI-7224*/
                        ((pciptr->vendor==VendorID)&&(pciptr->device==0x7224)&&(DeviceID==0x7248))||
                        /*Yuan 07/04/00 add support of PCI-7348*/
                        ((pciptr->vendor==VendorID)&&(pciptr->device==0x7348)&&(DeviceID==0x7396)))
                    {
                        //yuan add for fix PLX9050 bug
                        if(pciptr->base_address[1]&0x0080){
                            modBaseAddr = pciptr->base_address[5];
                            if(!(modBaseAddr &0x0001))
                                modBaseAddr = (pciptr->base_address[2]+0x100)&0xFF00;
                            else
                                modBaseAddr = modBaseAddr&0xfffc;
                            /*add the IO SPACE bit for the find_valid_port_addr()*/
                            pciptr->base_address[1] = (modBaseAddr|PCI_BASE_ADDRESS_SPACE_IO);
                            pci_write_config_dword(pciptr, PCI_BASE_ADDRESS_1, (modBaseAddr|PCI_BASE_ADDRESS_SPACE_IO));
                            //update the pci bios data , 00/07/20 jeffrey add.
                            pcibios_write_config_dword(pciptr->bus->number, pciptr->devfn, PCI_BASE_ADDRESS_1, (modBaseAddr|PCI_BASE_ADDRESS_SPACE_IO));
                        }
                        if(DeviceID==0x7396) //this may be ignored
                            DeviceID = pciptr->device;
                        /*additional check is needed*/
                        pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                        DevExtension[pci_info_index].wBus = pciptr->bus->number;
                        DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                        DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                        DevExtension[pci_info_index].Irq_no = pciptr->irq;
                        find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                        DevExtension[pci_info_index].wPortCnt = wPort_Count;
                        DevExtension[pci_info_index].wPciController = PciController;
                        DevExtension[pci_info_index].wCard = pci_info_index;
                        AdlSetupBuffer(&DevExtension[pci_info_index]);
                        pci_info_index++;
                    }
                    /*Alex 980925*/
                    break;
                case PCI_CTRL_P9080:
                    if(((pciptr->vendor==0x10b5)&&(pciptr->device==0x9080)&&(subvendor_id==VendorID)&&(subsystem_id==DeviceID))||
                        //yuan add 10/08/99
                        ((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)))
                    {
                        //yuan add 10/08/99 to minimize DMA latency
                        modLatency = 0xff;
                        pci_write_config_word(pciptr, PCI_LATENCY_TIMER, modLatency);
                        //update the pci bios data , 00/07/20 jeffrey add.
                        pcibios_write_config_dword(pciptr->bus->number, pciptr->devfn, PCI_LATENCY_TIMER, modLatency);
                        pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                        DevExtension[pci_info_index].wBus = pciptr->bus->number;
                        DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                        DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                        DevExtension[pci_info_index].Irq_no = pciptr->irq;
                        find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                        DevExtension[pci_info_index].wPortCnt = wPort_Count;
                        DevExtension[pci_info_index].wPciController = PciController;
                        DevExtension[pci_info_index].wCard = pci_info_index;
                        AdlSetupBuffer(&DevExtension[pci_info_index]);
                        pci_info_index++;
                    }
                    break;
                case PCI_CTRL_P9030:
                    if((pciptr->vendor==VendorID)&&(pciptr->device==DeviceID)){
                        pci_read_config_byte(pciptr, PCI_REVISION_ID, &(DevExtension[pci_info_index].wRevId));
                        DevExtension[pci_info_index].wBus = pciptr->bus->number;
                        DevExtension[pci_info_index].wSlot = PCI_SLOT(pciptr->devfn);
                        DevExtension[pci_info_index].wFunc = PCI_FUNC(pciptr->devfn);
                        DevExtension[pci_info_index].Irq_no = pciptr->irq;
                        find_valid_port_addr(&DevExtension[pci_info_index], pciptr, wPort_Count);
                        DevExtension[pci_info_index].wPortCnt = wPort_Count;
                        DevExtension[pci_info_index].wPciController = PciController;
                        if(DeviceID==0x7256||DeviceID==0x7258||DeviceID==0x7260){
                            U32 dwTmp, card_idx;
                            pci_read_config_dword(pciptr, PCI_BASE_ADDRESS_1, (int *)&modBaseAddr);
                            modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                            P9030_GetUserIoPin(modBaseAddr, 0L, &dwTmp);
                            card_idx = dwTmp;
                            P9030_GetUserIoPin(modBaseAddr, 1L, &dwTmp);
                            card_idx += dwTmp<<1;
                            P9030_GetUserIoPin(modBaseAddr, 2L, &dwTmp);
                            card_idx += dwTmp<<2;
                            P9030_GetUserIoPin(modBaseAddr, 3L, &dwTmp);
                            card_idx += dwTmp<<3;
                            if(DeviceID==0x7260){
                                P9030_GetUserIoPin(modBaseAddr, 8, &dwTmp);
                                card_idx += dwTmp<<4;
                                if(card_idx>=32)
                                    DevExtension[pci_info_index].wCard = pci_info_index;
                                else
                                    DevExtension[pci_info_index].wCard = card_idx;
                            }
                            else{
                                if(card_idx<16){
                                    DevExtension[pci_info_index].wCard = card_idx;
                                    //printk("PCI7256/PCI7258: card_idx[%d]\n", card_idx);
                                }
                                else{
                                    DevExtension[pci_info_index].wCard = pci_info_index;
                                }
                            }
                        }
                        else if((DeviceID==0x7452)||(DeviceID==0x7442)||(DeviceID==0x7443)||(DeviceID==0x7444))
                        {
                            U32 dwTmp, card_idx;
                            pci_read_config_dword(pciptr,PCI_BASE_ADDRESS_1, (int *)&modBaseAddr);
                            modBaseAddr &= PCI_BASE_ADDRESS_IO_MASK;
                            P9030_GetUserIoPin(modBaseAddr, 4L, &dwTmp);
                            card_idx = dwTmp;
                            P9030_GetUserIoPin(modBaseAddr, 5L, &dwTmp);
                            card_idx += dwTmp<<1;
                            P9030_GetUserIoPin(modBaseAddr, 6L, &dwTmp);
                            card_idx += dwTmp<<2;
                            P9030_GetUserIoPin(modBaseAddr, 7L, &dwTmp);
                            card_idx += dwTmp<<3;
                            if(card_idx<16)
                                DevExtension[pci_info_index].wCard = card_idx;
                            else
                                DevExtension[pci_info_index].wCard = pci_info_index;
                        }
                        else{
                            DevExtension[pci_info_index].wCard = pci_info_index;
                        }
                        AdlSetupBuffer(&DevExtension[pci_info_index]);
                        pci_info_index++;
                    }
                    break;
                default:
                    break;
            } //end of switch(PciController)
            pciptr = pciptr->next;
        } //end of while
        return pci_info_index; //number of the desired device
    }
    else
        return -1;
#else
    return -1;
#endif
}

#endif


BOOLEAN AdlSetupBuffer(PPCI_DEVEXT pDevExt)
{
    DAS_IOT_FUNC_DESP *pFdsp;
    PPCI_DBX pDbx;
    PPCI_DRVEXT pDrvExt;
    int vi;

    pDrvExt = pDevExt->pDrvExt;
    pFdsp = pDrvExt->CardDesp.mFunc;

    /*Buffer Allocation*/
    for(vi=0; vi<DAS_FUNC_COUNT; vi++, pFdsp++){
        pDevExt->pDbx[vi] = NULL;
        /*Install a DBX when required in FunctionDescriptor*/
        /* ==> has function, interrupt or DMA, GainArray*/
        if(!pFdsp->wFlag.bFunction)
            continue;
        if((!pFdsp->wFlag.bDMA)&&(!pFdsp->wFlag.bInterrupt))
            continue;
        if(!pFdsp->wArraySize)
            continue;
        pDbx = (PPCI_DBX)kmalloc(sizeof(PCI_DBX), GFP_KERNEL);
        if(!pDbx)
            return FALSE;
        pDevExt->pDbx[vi] = pDbx;
        memset(pDbx, '\0', sizeof(PCI_DBX));
    }

    /*Dual interrupt Event setup*/
    if(pDrvExt->CardDesp.fDualInterruptEventNotify){
        pDevExt->pDin = AdlSetupDsen();
        if(!pDevExt->pDin)
            return FALSE;
    }
    /*Trace Notify Event setup*/
    pDevExt->pTcn = NULL;
    if(pDrvExt->CardDesp.fTraceEventNotify){
        pDevExt->pTcn = AdlSetupDsen();
        if(!pDevExt->pTcn)
            return FALSE;
    }

    return TRUE;
}


unsigned long adl_req_memory(unsigned long bytecnt)
{
    int pagecnt = 0, i = 0, order = 0;
    unsigned long pBaseAdd = 0;

    //printk("[7300: new adl_req_memory");

    pagecnt = (bytecnt%PAGE_SIZE)? (bytecnt/PAGE_SIZE)+1:(bytecnt/PAGE_SIZE);
    bytecnt = PAGE_SIZE*pagecnt;
    if(pagecnt>1){
        order = get_order(bytecnt);
    }
    else
        order = 0;

    while(order>=0){
        pBaseAdd = __get_free_pages(GFP_KERNEL | GFP_DMA32, order);
        if(pBaseAdd)
            break;
        order--;
        pagecnt = 1<<order;
    }

    if(!pBaseAdd){
        return 0; //end of list
    }

    for(i=0; i<pagecnt; i++)
        mem_map_reserve(MAP_NR(pBaseAdd+i*PAGE_SIZE));

    memset((void *)pBaseAdd, 0, PAGE_SIZE<<order);

    return pBaseAdd;
}


int adl_free_mem(unsigned long adr, unsigned long len)
{
    u32 size = 0, order = 0, tmp_adr = 0;
    size = len;
    tmp_adr = adr;
    order = get_order(len);

    //printk("[7300: new adl_free_mem");

    while(size>0){
        mem_map_unreserve(MAP_NR(tmp_adr));
        tmp_adr+=PAGE_SIZE;
        size-=PAGE_SIZE;
    }
    free_pages(adr, order);

    return 0;
}
