#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/Adlp9030.h"


void P9030_SetInt(U32 dwPort, U16 wPin, U16 wEnable, U16 act_lev, U16 trig, U16 int_clr)
{
    PLX90_WORDREG rPlx;
    U8 bEnable = wEnable? 1:0;

    rPlx.r = OP_inpw(dwPort, LINTCSR);
    switch(wPin){
        case 1:
            rPlx.intcsr.Int1Enable = bEnable;
            if(act_lev){
                rPlx.intcsr.Int1ActHigh = 1;
                if(trig)
                    rPlx.intcsr.Int1EdgeTrig = 1; //only operate in active high mode
                else
                    rPlx.intcsr.Int1EdgeTrig = 0;
            }
            else{
                rPlx.intcsr.Int1ActHigh = 0;
                rPlx.intcsr.Int1EdgeTrig = 0; //bit[8]=0 in level-active mode
            }
            if(int_clr)
                rPlx.intcsr.Int1Clear = 1;
            else
                rPlx.intcsr.Int1Clear = 0;
            break;
        case 2:
            rPlx.intcsr.Int2Enable = bEnable;
            if(act_lev){
                rPlx.intcsr.Int2ActHigh = 1;
                if(trig)
                    rPlx.intcsr.Int2EdgeTrig = 1; //only operate in active high mode
                else
                    rPlx.intcsr.Int2EdgeTrig = 0;
            }
            else{
                rPlx.intcsr.Int2ActHigh = 0;
                rPlx.intcsr.Int2EdgeTrig = 0; //bit[8]=0 in level-active mode
            }
            if(int_clr)
                rPlx.intcsr.Int2Clear = 1;
            else
                rPlx.intcsr.Int2Clear = 0;
            break;
        case 0:
        default:
            rPlx.intcsr.PciIntEnable = bEnable;
            break;
    }
    if(wPin){
        if(rPlx.intcsr.Int2Enable|rPlx.intcsr.Int1Enable)
            rPlx.intcsr.PciIntEnable = 1;
        else
            rPlx.intcsr.PciIntEnable = 0;
    }
    OP_outw(dwPort, LINTCSR, rPlx.r);
}


void P9030_IocSetDualInt(PPCI_DEVEXT pDevExt, DAS_IOT_DUALINT *pSbuf)
{
    PLX90_WORDREG rPlx;
    U32 dwPort;
#ifdef __SMP__
    unsigned long irq_flags;
#endif
    PPCI_DSEN pDsen;
    //KIRQL kIrql;

    pDsen = pDevExt->pDin;
    dwPort = OP_Port(pDevExt);
    pSbuf->wErrorCode = DAS_ERR_NO;

#ifdef __SMP__
    //KeAcquireSpinLock(&pDsen->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDsen->kSpinLock, irq_flags);
#endif

    pDsen->bSrc1Enable = pSbuf->bInt1Enable? 1:0;
    pDsen->bSrc2Enable = pSbuf->bInt2Enable? 1:0;
    pDsen->wSrc1Mode = pSbuf->bInt1Mode;
    pDsen->wSrc2Mode = pSbuf->bInt2Mode;

    rPlx.r = OP_inpw(dwPort, LINTCSR);
    rPlx.intcsr.Int1ActHigh = pSbuf->bInt1ActiveHigh? 1:0;
    if(rPlx.intcsr.Int1ActHigh){
        if(pSbuf->bInt1Mode&0x1)
            rPlx.intcsr.Int1EdgeTrig = 1; //only operate in active high mode
    }
    else
        rPlx.intcsr.Int1EdgeTrig = 0;

    rPlx.intcsr.Int2ActHigh = pSbuf->bInt2ActiveHigh? 1:0;
    if(rPlx.intcsr.Int2ActHigh){
        if(pSbuf->bInt2Mode&0x1)
            rPlx.intcsr.Int2EdgeTrig = 1; //only operate in active high mode
    }
    else
        rPlx.intcsr.Int2EdgeTrig = 0;

    rPlx.intcsr.Int1Enable = pSbuf->bInt1Enable;
    rPlx.intcsr.Int2Enable = pSbuf->bInt2Enable;
    rPlx.intcsr.Int1Clear = (pSbuf->bInt1Mode>>4)&0x1;
    rPlx.intcsr.Int2Clear = (pSbuf->bInt2Mode>>4)&0x1;
    if(pSbuf->bInt1Enable|pSbuf->bInt2Enable)
        rPlx.intcsr.PciIntEnable = 1;
    else
        rPlx.intcsr.PciIntEnable = 0;
    OP_outw(dwPort, LINTCSR, rPlx.r);

#ifdef __SMP__
    //KeReleaseSpinLock(&pDsen->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDsen->kSpinLock, irq_flags);
#endif
}


void P9030_SetUserIoPin(U32 dwPort, U32 dwPin, U32 dwDat)
{
    U32 dwShf;
    U32 dwVal;

    dwShf = dwPin*3;
    dwVal = OP_inpd(dwPort, GPIOC)&(~(0x00000007<<dwShf));
    dwVal |= (dwDat&0x7)<<dwShf;
    OP_outd(dwPort, GPIOC, (U32)dwVal);
}


void P9030_GetUserIoPin(U32 dwPort, U32 dwPin, U32 *dwpDat)
{
    U32 dwShf;
    U32 dwVal;

    dwShf = dwPin*3;
    dwVal = (OP_inpd(dwPort, GPIOC)&(0x00000007<<dwShf))>>dwShf;
    *dwpDat = (dwVal>>2)&1;
}
