#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/Adlplx90.h"


void PLX90_ClearController(U32 dwPort)
{
    OP_outb(dwPort,LINTCSR,0x00);
    OP_outb(dwPort,EEPROM,0x40);
    OP_outb(dwPort,EEPROM,0x00);
}


void PLX90_SoftReset(U32 dwPort)
{
    OP_outb(dwPort,EEPROM,0x40);
    OP_outb(dwPort,EEPROM,0x00);
}


void PLX90_SetInt(U32 dwPort, U16 wPin, U16 wEnable)
{
    PLX90_BYTEREG rPlx;
    U8  bEnable = wEnable? 1:0;

    rPlx.r = OP_inpb(dwPort, LINTCSR);
    switch(wPin){
        case 0:
            rPlx.intcsr.PciIntEnable = bEnable;
            break;
        case 1:
            rPlx.intcsr.Int1Enable  = bEnable;
            rPlx.intcsr.Int1ActHigh = 1;
            break;
        case 2:
            rPlx.intcsr.Int2Enable  = bEnable;
            rPlx.intcsr.Int2ActHigh = 1;
            break;
        default:
            break;
    }
    OP_outb(dwPort, LINTCSR, rPlx.r);
}

void PLX90_IocSetDualInt(PPCI_DEVEXT pDevExt, DAS_IOT_DUALINT *pSbuf)
{
    PLX90_BYTEREG rPlx;
    U32 dwPort;
#ifdef __SMP__
    unsigned long irq_flags;
#endif
    PPCI_DSEN pDsen;
    //KIRQL kIrql;

    pDsen = pDevExt->pDin;
    dwPort = OP_Port(pDevExt);
    pSbuf->wErrorCode = DAS_ERR_NO;

#ifdef __SMP__
    //KeAcquireSpinLock(&pDsen->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDsen->kSpinLock, irq_flags);
#endif

    pDsen->bSrc1Enable = pSbuf->bInt1Enable? 1:0;
    pDsen->bSrc2Enable = pSbuf->bInt2Enable? 1:0;
    pDsen->wSrc1Mode = pSbuf->bInt1Mode;
    pDsen->wSrc2Mode = pSbuf->bInt2Mode;

    rPlx.r = OP_inpb(dwPort, LINTCSR);
    rPlx.intcsr.Int1ActHigh = pSbuf->bInt1ActiveHigh? 1:0;
    rPlx.intcsr.Int2ActHigh = pSbuf->bInt2ActiveHigh? 1:0;
    rPlx.intcsr.Int1Enable  = pSbuf->bInt1Enable;
    rPlx.intcsr.Int2Enable  = pSbuf->bInt2Enable;
    rPlx.intcsr.PciIntEnable = 1;
    OP_outb(dwPort, LINTCSR, rPlx.r);

#ifdef __SMP__
    //KeReleaseSpinLock(&pDsen->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDsen->kSpinLock, irq_flags);
#endif
}


void PLX90_IocGetDualInt(PPCI_DEVEXT pDevExt, DAS_IOT_DUALINT *pSbuf)
{
    PLX90_BYTEREG rPlx;
    U32 dwPort;
#ifdef __SMP__
    unsigned long irq_flags;
#endif
    PPCI_DSEN pDsen;

    dwPort = OP_Port(pDevExt);
    pDsen = pDevExt->pDin;
    pSbuf->wErrorCode = DAS_ERR_NO;

#ifdef __SMP__
    //KeAcquireSpinLock(&pDsen->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDsen->kSpinLock, irq_flags);
#endif

    pSbuf->bInt1Mode = (U8)pDsen->wSrc1Mode;
    pSbuf->bInt2Mode = (U8)pDsen->wSrc2Mode;
    rPlx.r = OP_inpb(dwPort, LINTCSR);
    pSbuf->bInt1Enable = rPlx.intcsr.Int1Enable;
    pSbuf->bInt2Enable = rPlx.intcsr.Int2Enable;
    pSbuf->bInt1ActiveHigh = rPlx.intcsr.Int1ActHigh;
    pSbuf->bInt2ActiveHigh = rPlx.intcsr.Int2ActHigh;
    pSbuf->bInt1Status = rPlx.intcsr.Int1Status;
    pSbuf->bInt2Status = rPlx.intcsr.Int2Status;

#ifdef __SMP__
    //KeReleaseSpinLock(&pDsen->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDsen->kSpinLock, irq_flags);
#endif
}


void PLX90_SetUserIoPin(U32 dwPort, U16 wPin, U16 wDat)
{
    U16 wShf;
    U16 wVal;

    wShf = (wPin&3)*3;
    wVal = OP_inpw(dwPort, CNTRL)&(~(0x7<<wShf));
    wVal |= (wDat&0x7)<<wShf;
    OP_outw(dwPort, CNTRL, wVal);
}


void PLX90_GetUserIoPin(U32 dwPort, U16 wPin, U16 *wpDat)
{
    U16 wShf;
    U16 wVal;

    wShf = (wPin&3)*3;
    wVal = OP_inpw(dwPort, CNTRL)>>wShf;
    *wpDat = (wVal>>2)&1;
}


U16 PLX90_IsrCheckSource(U32 dwPort)
{
    PLX90_BYTEREG rPlx;
    U16 wRetCode = 0;

    rPlx.r = OP_inpb(dwPort, LINTCSR);
    if(rPlx.intcsr.SoftIntrrupt){
        wRetCode |= 0x0200;
        /*must clear soft interrupt before return*/
        rPlx.intcsr.SoftIntrrupt = 0;
        OP_outb(dwPort, LINTCSR, rPlx.r);
    }
    if(rPlx.intcsr.PciIntEnable)
        wRetCode |= 0x0100;
    if(rPlx.intcsr.Int1Enable&&rPlx.intcsr.Int1Status)
        wRetCode |= 0x0001;
    if(rPlx.intcsr.Int2Enable&&rPlx.intcsr.Int2Status)
        wRetCode |= 0x0002;

    return wRetCode;
}
