#include <linux/version.h>
#include <linux/kernel.h>
//#include <linux/config.h>

#ifdef CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#if defined(__SMP__)
#include <linux/smp.h>
#endif

#if defined(MODVERSIONS)
#if defined(__SMP__)
#include <linux/modversions-smp.h>
#else
#include <linux/modversions.h>
#endif
#endif

#include <linux/sched.h>
//#include <linux/module.h>
#include <linux/mm.h>
#include <asm/page.h>
#include <asm/io.h>

#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,13)
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

extern U32 BufSize[DAS_FUNC_COUNT];


BOOLEAN AdlDbxAllocate(PPCI_DEVEXT pDevExt)
{

    DAS_IOT_FUNC_DESP *pFdsp;
    PPCI_DBX pDbx;
    PPCI_DRVEXT pDrvExt;
    int vi;

    pDrvExt = pDevExt->pDrvExt;
    pFdsp = pDrvExt->CardDesp.mFunc;

    for(vi=0; vi<DAS_FUNC_COUNT; vi++, pFdsp++){
        pDbx = pDevExt->pDbx[vi];
        if(pDbx==NULL)
            continue;

        if(BufSize[vi]==0){ //buffer size = 0, no AdlSetupDbxBuffer is needed
            pDbx->xAddDbx[0].dwAddPhy = 0;
            pDbx->dwBufferSize = 0;
        //01/10/08 added
        #ifdef __SMP__
            //KeInitializeSpinLock( & pDbx->kSpinLock );
            spin_lock_init(&pDbx->kSpinLock);
        #endif
            /*setup array size*/
            pDbx->wCGArraySize = pFdsp->wArraySize;
            pDbx->xAddDbx[0].dwAddMap = 0L;
            continue;
        }

        /*setup the buffer from the Adl_Mem_Mgr module*/
        //additional page (4 Kbyte) for PCI_DBXORUN structure
        if(!AdlSetupDbxBuffer(pDbx, pFdsp->wArraySize, BufSize[vi]+4)){
            kfree((void *)pDbx);
            pDbx = NULL;
            vi--;
            for( ; vi>=0; vi--){ //clear allocated Dbx buffer
                pDbx = pDevExt->pDbx[vi];
                AdlDropDbx(pDbx);
                pDevExt->pDbx[vi] = NULL;
            }
            return FALSE;
        }
    }

    return TRUE;
}


/*
 * use get_free_page to allocate memory temprary , Allocator module will replace
 * the Highest memory process , and "ioremap" will map the Range-Out memory into memory space
 * "remap_page_range" will release View of kernel memory space to user application
 * NOW!! just use get_free_page for memory allocation , max. size 128k byte
 */
BOOLEAN AdlSetupDbxBuffer(register PPCI_DBX pDbx, U16 wArraySize,U32 buffer_kbyte)
{
    U32 pHighMemAdd;
    void *pBaseAdd;
    U32 dwLength;

    /*Initialize Spin Lock of DBX*/
#ifdef __SMP__
    //KeInitializeSpinLock( & pDbx->kSpinLock );
    spin_lock_init( &pDbx->kSpinLock );
#endif

    /*setup array size*/
    pDbx->wCGArraySize = wArraySize;
    pDbx->dwBufferSize = 0L;
    pDbx->xAddDbx[0].dwAddMap = 0L;

    /*Try to find the nearest maximum value*/
    //yuan 12/01/99
    pHighMemAdd = 0L;
    pHighMemAdd = adl_request_memory(buffer_kbyte, GFP_KERNEL);
    if(pHighMemAdd==0) //can't allocate enough memory
        return FALSE;

    dwLength = buffer_kbyte<<10;
    //map the out-space physical memory into kernel space
    pBaseAdd = ioremap(pHighMemAdd, dwLength);
    //printk("alloc: %x %x\n",pHighMemAdd, pBaseAdd);
    memset(pBaseAdd, 0, dwLength);
    dwLength -= DbxORunStructSize;
    /*setup dual buffer memory address and length*/
    //pDbx->dwBufferSize = dwLength;
    pDbx->dwBufferSize = dwLength;
    pDbx->xAddDbx[0].dwLength = dwLength;
    pDbx->xAddDbx[1].dwLength = dwLength/2;
    //pDbx->xAddDbx[0].dwAddMap = (U32)pBaseAdd;
    pDbx->xAddDbx[0].dwAddMap = (U32)pBaseAdd+DbxORunStructSize;
    pDbx->xAddDbx[1].dwAddMap = (U32)pBaseAdd+DbxORunStructSize+dwLength/2;
    /*get Physical Address of contigous memory*/
    //leave for DMA , but that will change by allocator module
    //pDbx->xAddDbx[0].dwAddPhy = pHighMemAdd;
    pDbx->xAddDbx[0].dwAddPhy = pHighMemAdd+DbxORunStructSize;
    pDbx->xAddDbx[1].dwAddPhy = pHighMemAdd+DbxORunStructSize+dwLength/2;
    pDbx->hMapView[2] = (void *)pBaseAdd;

    return TRUE;
}


/*Delete DBX object*/
void AdlDropDbx(register PPCI_DBX pDbx)
{
    if(!pDbx)
        return;

    /*free DoubleBuferr Memory*/
    if(pDbx->xAddDbx[0].dwAddMap){
        iounmap((void *)pDbx->xAddDbx[0].dwAddMap-DbxORunStructSize);
        adl_free_memory(pDbx->xAddDbx[0].dwAddPhy-DbxORunStructSize);
        pDbx->xAddDbx[0].dwAddMap = 0L;
    }

    kfree((void *)pDbx);
}
