#ifndef _ADLS593X_H
#define _ADLS593X_H

void          S593_ClearController(unsigned long dwPort);
void          S593_StartIDMA(unsigned long dwPort, PPCI_ADD pBufAdd);
void          S593_StartODMA(unsigned long dwPort, PPCI_ADD pBufAdd);
unsigned long S593_StopIDMA(unsigned long dwPort);
unsigned long S593_StopODMA(unsigned long dwPort);
unsigned long S593_InCountLeft(unsigned long dwPort);
unsigned long S593_OutCountLeft(unsigned long dwPort);
unsigned long S593_InFlushFIFO(unsigned long dwPort, unsigned long dwAccessCount, unsigned long dwBufHead1, unsigned long dwBufHead2);
BOOLEAN       S593_OutFlushFIFO(unsigned long dwPort);
void          S593_StartIMB(U32 dwPort, U16 wMailBox, U16 wMbxByte);
void          S593_StartOMB(U32 dwPort, U16 wMailBox, U16 wMbxByte);
void          S593_StopIMB(unsigned long dwPort);
void          S593_StopOMB(unsigned long dwPort);


/*
 * OMB  -- Outgoing MailBox register
 * IMB  -- Incoming MailBox register
 * FIFO -- FIFO register port (bidirectional)
 * MWAR -- Master Write Address Register
 * MRAR -- Master Read  Address Register
 * MWTC -- Master Write Transfer Count Register
 * MRTC -- Master Read  Transfer Count Register
 * MBEF -- MailBox Empty/Full status
 * INTCSR -- Interrupt Control/Status register
 * MCSR -- Bus Master Control/Status register
 */
#define ADL_OP_OMB1   0x00
#define ADL_OP_OMB2   0x04
#define ADL_OP_OMB3   0x08
#define ADL_OP_OMB4   0x0c
#define ADL_OP_IMB1   0x10
#define ADL_OP_IMB2   0x14
#define ADL_OP_IMB3   0x18
#define ADL_OP_IMB4   0x1c

#define ADL_OP_FIFO   0x20
#define ADL_OP_MWAR   0x24
#define ADL_OP_MWTC   0x28
#define ADL_OP_MRAR   0x2c
#define ADL_OP_MRTC   0x30
#define ADL_OP_MBEF   0x34
#define ADL_OP_INTCSR 0x38
#define ADL_OP_MCSR   0x3c

/**/
#define MAILBOX1     0
#define MAILBOX2     1
#define MAILBOX3     2
#define MAILBOX4     3
#define FIFOBYTE0    0
#define FIFOBYTE1    1
#define FIFOBYTE2    2
#define FIFOBYTE3    3
#define ENDIANCON_NO 0
#define ENDIANCON_16 1
#define ENDIANCON_32 2
#define ENDIANCON_64 3

#define INTCSR_CLEAR_STATUS 0x00bf0000
#define MCSR_CLEAR_STATUS   0x0e000000

// #pragma pack(push,1)

typedef struct
{
    unsigned long C_OMB_BYTE              : 2;
    unsigned long C_OMB_MBX               : 2;
    unsigned long C_OMB_IE                : 1;
    unsigned long Dummy1                  : 3;
    unsigned long C_IMB_BYTE              : 2;
    unsigned long C_IMB_MBX               : 2;
    unsigned long C_IMB_IE                : 1;
    unsigned long Dummy2                  : 1;
    unsigned long C_MWTC_IE               : 1;
    unsigned long C_MRTC_IE               : 1;
    unsigned long S_OMB_INT               : 1;
    unsigned long S_IMB_INT               : 1;
    unsigned long S_MWTC_INT              : 1;
    unsigned long S_MRTC_INT              : 1;
    unsigned long S_MASTER_ABORT_INT      : 1;
    unsigned long S_TARGET_ABORT_INT      : 1;
    unsigned long Dummy3                  : 1;
    unsigned long S_INT_ASSERT            : 1;
    unsigned long S_ENDIAN_CON            : 2;
    unsigned long S_P2A_FIFO_ADVANCE_BYTE : 2;
    unsigned long S_A2P_FIFO_ADVANCE_BYTE : 2;
    unsigned long S_A2P_FIFO_TOGGLE       : 1;
    unsigned long S_P2A_FIFO_TOGGLE       : 1;
}
S593X_INTCSR;

typedef struct
{
    unsigned long S_P2A_ALLFULL       : 1;
    unsigned long S_P2A_4FULL         : 1;
    unsigned long S_P2A_EMPTY         : 1;
    unsigned long S_A2P_ALLFULL       : 1;
    unsigned long S_A2P_4FULL         : 1;
    unsigned long S_A2P_EMPTY         : 1;
    unsigned long S_P2A_TC_0          : 1;
    unsigned long S_A2P_TC_0          : 1;
    unsigned long C_A2P_HI_PRIORITY   : 1;
    unsigned long C_A2P_REQ_AT_4FULL  : 1;
    unsigned long C_A2P_EN_TRANSFER   : 1;
    unsigned long Dummy1              : 1;
    unsigned long C_P2A_HI_PRIORITY   : 1;
    unsigned long C_P2A_REQ_AT_4EMPTY : 1;
    unsigned long C_P2A_EN_TRANSFER   : 1;
    unsigned long Dummy2              : 1;
    unsigned long C_NV_OP             : 8;
    unsigned long C_ADDON_RESET       : 1;
    unsigned long C_P2A_FLAG_RESET    : 1;
    unsigned long C_A2P_FLAG_RESET    : 1;
    unsigned long C_MBX_FLAG_RESET    : 1;
    unsigned long Dummy3              : 1;
    unsigned long C_NV_ACCESS         : 3;
}
S593X_MCSR;

typedef union _OP_REG
{
    S593X_INTCSR  intcsr;
    S593X_MCSR    mcsr;
    unsigned long r;
}
OP_REG, *POP_REG;

//#pragma pack(pop)

#endif
