#ifndef _ADLDEV_H_
#define _ADLDEV_H_

#include <linux/pci.h> //struct pci_dev
#include <linux/timer.h>
#include <linux/version.h>

#ifdef CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a)<16)|((b)<8)|(c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
#include <linux/sched.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
#include <linux/moduleparam.h>
#include <linux/device.h>
#include <linux/workqueue.h>
#else
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
#include <linux/tqueue.h>
#endif
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
//#include <linux/tqueue.h>
#include <linux/wait.h> //for DECLARE_WAIT_QUEUE_HEAD
#include <linux/spinlock.h> //struct spinlock_t
#else
#include <asm/spinlock.h> //struct spinlock_t
#endif

#define MAXMBUFCNT 64


/*
 * Constant for structures
 */
/*Port report from PCI bios*/
#define PCI_PORT_MAX   4
#define PCI_CTRL_S593  0
#define PCI_CTRL_PLX90 1
#define PCI_CTRL_P9080 2
//yuan add 09/24/01
#define PCI_CTRL_P9030 3
/*FPGA download register definition*/
#define FPGA_CONFIG      0x0100
#define FPGA_CLR_RST     0x0200
#define FPGA_STATUS      0x0001
#define FPGA_CONFIG_DONE 0x0002
#define FPGA_INIT_DONE   0x0004
#define FPGA_DATA_BUSY   0x0008

//#pragma pack(push,1)

#ifndef basic_type
#define basic_type

typedef unsigned char      U8;
typedef short              I16;
typedef unsigned short     U16;
typedef long               I32;
typedef unsigned long      U32;
typedef unsigned long long U64;
typedef float              F32;
typedef double             F64;
typedef enum{FALSE, TRUE}  BOOLEAN;

#define FIRSTBYTE(VALUE)   (VALUE&0x00FF)
#define SECONDBYTE(VALUE)  ((VALUE>>8)&0x00FF)
#define THIRDBYTE(VALUE)   ((VALUE>>16)&0x00FF)
#define FOURTHBYTE(VALUE)  ((VALUE>>24)&0x00FF)

#endif //basic_type


/*PCI Address / Port Struct -- Physical Address, Base Address, Length*/
typedef struct pci_buffer_adderss
{
    U32 dwAddPhy;
    U32 dwAddMap;
    U32 dwLength;
}
PCI_ADD, *PPCI_ADD;


typedef union pci_dbx_flag
{
    struct{
        U16 Running    : 1;
        U16 DMA        : 1;
        U16 AutoScan   : 1;
        U16 Output     : 1;
        U16 SyncMode   : 1;
        U16 DblBuf     : 1;
        U16 DrvHfRdyH  : 1;
        U16 UsrHfRdyH  : 1;
        U16 HalfRdyL   : 1;
        U16 HalfRdyH   : 1;
        U16 Tracing    : 1;
        U16 TimerStart : 1;
        U16 IdleTime   : 1;
        U16 PostTrig   : 1;
        U16 ChainDMA   : 1;
        //yuan add 09/29/00
        U16 DlyTrig    : 1;
        //U16 Resrv2     : 1;
    }f;
    U16 w;
}
PCI_DBXFLAG,*PPCI_DBXFLAG;


/*Double buffer transfer use struct*/
typedef struct pci_double_buffer_transfer
{
    void                 *hMapView[MAXMBUFCNT];
    //jeffrey 00/04/14, mark for unusefull event definition
    //jeffrey 00/04/14, No Interrupt Request Package in Linux
    U32                  dwBufferSize;
    U32                  dwAccessCount;
    U32                  dwAccessPos;
#ifdef __SMP__
    //jeffrey 01/01/04, Setup spinlock for SMP kernel
    spinlock_t           kSpinLock;
#endif
    PCI_ADD              xAddDbx[2];
    U16                  wCGArraySize;
    U16                  wBytesPerData;
    U16                  wIterations;
    U16                  wPacerMin;
    PCI_DBXFLAG          xFlag;
    //yuan test 01/18/01
    //PMDL                 pmdl;
    PCI_ADD              xAddMbx[MAXMBUFCNT];
    //U64                  UsrSel;
    //U64                  IsrSel;
    U32                  UsrSel;
    U32                  IsrSel;
    U64                  DrvRdy;
    int                  sig_no;
    struct fasync_struct *async_queue;
    //01/14/08 add
    U32                  dwTimeOut;
}
PCI_DBX, *PPCI_DBX;


/*
 * Driver extension -- linked by each device extension
 */
//jeffrey 00/04/14 , too many WinNT-kernel structure , ignore that temprary
typedef struct pci_driver_extension
{
    DAS_IOT_CARD_DESP CardDesp;
    U16               wPacerMin[DAS_FUNC_COUNT];
    BOOLEAN           fTimerRoutine;
}
PCI_DRVEXT, *PPCI_DRVEXT;


/*jeffrey 00/04/14, mark the EVENT structure in Linux*/
typedef struct pci_Dual_Src_Event_notify
{
    //06/09/00 jeffrey , async_queue replace the Event mechenism
    int                  sig_no1;
    int                  sig_no2;
    struct fasync_struct *async_queue1;
    struct fasync_struct *async_queue2;
#ifdef __SMP__
    //jeffrey 01/01/04, Setup spinlock for SMP kernel
    spinlock_t           kSpinLock;
#endif
    U16                  wSrc1Mode;
    U16                  wSrc2Mode;
    U8                   bSrc1Enable;
    U8                   bSrc2Enable;
}
PCI_DSEN, *PPCI_DSEN;


/*
 * Device extension -- create by IO Manger
 *
 * hDevice -- Pointer to the device object
 * pDrvExt -- Pointer to the Driver extension
 * wPortCnt -- Count of Port report by BIOS
 * xDbx -- Dbl Buf Txf struct for each function
 */
typedef struct pci_device_extension
{
    //jeffrey 00/04/14, mark the WinNT native Object in Linux
    PPCI_DRVEXT       pDrvExt;
    PPCI_DBX          pDbx[DAS_FUNC_COUNT];
    PPCI_DSEN         pDin;
    PPCI_DSEN         pTcn;
    U8                wFunc;
    U8                wBus;
    U8                wSlot;
    U8                Irq_no;
    U32               dwSlot;
    U32               dwDevRA;
    U32               dwDevRB;
    U32               dwPortAddr[PCI_PORT_MAX];
    U32               dwPortLen[PCI_PORT_MAX];
    U16               wPciController;
    U16               wCard;
    U8                wRevId;
    U16               wPortCnt;
    U16               Reference;
    struct timer_list adl_timer; //for IoStartTimer
    BOOLEAN           openable;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
    wait_queue_head_t sync_wq;
    wait_queue_head_t di_sync_wq;
    wait_queue_head_t do_sync_wq;
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
    struct wait_queue *sync_wq;
    struct wait_queue *di_sync_wq, *do_sync_wq;
#endif
}
PCI_DEVEXT, *PPCI_DEVEXT;


//#pragma pack(pop)

/*PCI Context -- struct use to communicate with this lib*/
typedef struct pci_context_for_loading
{
    /*Fill By Caller*/
    U16               wVendorID;
    U16               wDeviceID;
    //jeffrey 00/04/14 , mark for linux driver
    U16               wPciController;
    U16               wPortCount;
    U16               wPacerMin[DAS_FUNC_COUNT];
    DAS_IOT_CARD_DESP *pCardDesp;
    /*Reserved*/
    PPCI_DEVEXT       pDevExt;
    U16               wCard;
    U16               wBus;
    U16               wDevice;
    U16               wFunc;
    U8                wRevId;
    //yuan add for testing status 01/07/00
}
PCI_CONTEXT,*PPCI_CONTEXT;

#endif
