#include <linux/version.h>
#include <linux/module.h>   /* Specifically, a module */
#include <linux/kernel.h>   /* We're doing kernel work */
//#include <linux/config.h>   // for PCI, SMP configuration
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/interrupt.h>

#ifndef MODULE
#define MODULE
#endif

#ifdef CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#if defined(__SMP__)
#include <linux/smp.h>
#endif

#if defined(MODVERSIONS)
#if defined(__SMP__)
#include <linux/modversions-smp.h>
#else
#include <linux/modversions.h>
#endif
#endif


#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,13) )
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#include <unistd.h>
#endif
#endif

#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/Adlp9030.h"
#include "pci7260.h"

/* The name for our device, as it will appear in
 * /proc/devices */
#define DEVICE_NAME     "p7260"
#define ADL_VENDOR_ID   0x144A
#define ADL_DEVICE_ID   0x7260

#define PCI_CS_HOTSWAP_CTRL 0x4A

extern DAS_IOT_CARD_DESP  iniCardDesp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
extern irqreturn_t p7260_interrupt_handler(int irq, void *dev_id, struct pt_regs *regs);
#else
extern irqreturn_t p7260_interrupt_handler(int irq, void *dev_id);
#endif

PCI_DEVEXT DevExtension[32]; // instance for every device_open
static PCI_DRVEXT drv_ext;
int device_number = 0; // return by AdlFindDevice(...) in init_module

static int MAJOR = 0;
int irq_result =0;
// static int READ_PORT,WRITE_PORT;
#define ADL_PCI_CTL     PCI_CTRL_P9030

static int pci7260_init (struct pci_dev *, const struct pci_device_id *);
void pci7260_remove (struct pci_dev *);

static struct pci_device_id pci7260_tbl[] = {
    {ADL_VENDOR_ID, ADL_DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID, },
    {0, }
};

static struct pci_driver pci7260_driver = {
    name     : "PCI-7260 Driver",
    id_table : pci7260_tbl,
    probe    : pci7260_init,
    remove   : pci7260_remove,
};

MODULE_DEVICE_TABLE(pci, pci7260_tbl);
#define DUMMY_NAME "PCI7260W"
static struct class *pci7260_class;

int p7260_open(struct inode *inode, struct file *file)
{
    int dev_index, i;
    PPCI_DEVEXT dev_ext;
    U8 ctrl=0;
    U16 intsts = 0;

    //printk ("DRIVE: device_open(%p)\n", file);
    dev_index = MINOR(inode->i_rdev);
    printk ("DRIVE: device_index:%d\n", dev_index);
    for( i=0; i< device_number; i++ ){
        if( DevExtension[i].openable ){
            if( DevExtension[i].wCard == dev_index )
                break;
        }
    }

    if( i == device_number )
        return DAS_ERR_NO_DEVICE;
    else
        dev_index = i;

//  if( dev_index > device_number )
//  return DAS_ERR_NO_DEVICE;

//  if( !DevExtension[dev_index].openable )
//  return DAS_ERR_NO_DEVICE;

    dev_ext = &(DevExtension[dev_index]);
    if( !AdlOpenDevice( dev_ext ) )
        return DAS_ERR_NO_DEVICE;
    if(!dev_ext->Reference) {
        intsts = PT_inpw(PT_Port(dev_ext), INTCSR);
        ctrl = (intsts>>3) & 0x01;
        if(ctrl) {
            ctrl = (intsts>>11) & 0x01;
            PT_outb(PT_Port(dev_ext), EMG_CTL, 0x80|ctrl);
        }
        dev_ext->dwDevRA = 0;
        dev_ext->dwDevRB = 0;
    }
    file->private_data = dev_ext;
    dev_ext->Reference ++;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_INC_USE_COUNT;
#endif

    return DAS_ERR_NO;
}

int p7260_release(struct inode *inode,struct file *file)
{
    PPCI_DEVEXT pDevExt = file->private_data;

    printk ("DRIVE: device_release(%p,%p)\n", inode, file);

    if(pDevExt->Reference == 1) {
        if( pDevExt->pDin != NULL ){
            if( pDevExt->pDin->async_queue1 != NULL )
                fasync_helper( (int)inode, file, 0, &(pDevExt->pDin->async_queue1));
            if( pDevExt->pDin->async_queue2 != NULL )
                fasync_helper( (int)inode, file, 0, &(pDevExt->pDin->async_queue2));
        }
        if( pDevExt->pTcn ) {
            if( pDevExt->pTcn->async_queue1 != NULL )
                fasync_helper( (int)inode, file, 0, &(pDevExt->pTcn->async_queue1));
            if( pDevExt->pTcn->async_queue2 != NULL )
                fasync_helper( (int)inode, file, 0, &(pDevExt->pTcn->async_queue2));
        }
    }

    pDevExt->Reference --;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_DEC_USE_COUNT;
#endif

    return DAS_ERR_NO;
}

BOOLEAN
AdlPioWriteDIO(PPCI_DEVEXT pDevExt, DAS_IOT_PIO* pSbuf)
{
    U32 dwPort;
    U16 /*High_Byte, */Low_Byte;
    /*---------------------------------------*/
    dwPort = PT_Port(pDevExt) + ADL_PT_DIO;
    AdlIocPioWrite(1, pSbuf, inpW(dwPort));
    Low_Byte = (U16) (pSbuf->dwValue & 255);     // Read the low byte
    //reverse
    Low_Byte = (~Low_Byte) & 0xff;
    dwPort = PT_Port(pDevExt);
    PT_outw(dwPort, DO, (U16)Low_Byte);

    return TRUE;
}

void AdlSetDualInt( PPCI_DEVEXT pDevExt, DAS_IOT_DUALINT* pSbuf)
{
    U32 dwPort;
    U16 wTmp = 0;

    dwPort = PT_Port(pDevExt);
    pSbuf->bInt1Mode |= pSbuf->bInt2Mode;
    pSbuf->bInt1Mode &= 7;
    pSbuf->bInt2Mode = 0;
    pSbuf->bInt1Enable |= pSbuf->bInt2Enable;
    pSbuf->bInt1ActiveHigh = 1;
    pSbuf->bInt2Enable = 0;
    P9030_IocSetDualInt( pDevExt, pSbuf );
    wTmp = ( pSbuf->bInt1Mode ) << 8;
    PT_outw( dwPort, INTCSR, wTmp );
    wTmp |= 0x0007;
    PT_outw( dwPort, INTCSR, wTmp );
    P9030_SetInt( OP_Port(pDevExt), 1, 1, 1, 0, 1 );
    P9030_SetInt( OP_Port(pDevExt), 1, 1, 1, 0, 0 );
}

void AdlSetCOSInt(PPCI_DEVEXT pDevExt, DAS_IOT_COSINT* pSbuf)
{
    U16  wTmp;
    /*---------------------------------------*/
    wTmp = (U16) ((pSbuf->mCOS.wValue) & 0xff);
    PT_outw(PT_Port(pDevExt), COSCSR, wTmp);
}

BOOLEAN WDT_IocCtrSet( struct inode *inode, struct file *file, PPCI_DEVEXT pDevExt, DAS_IOT_CTR* pCtr )
{
    U32 dwPort = PT_Port (pDevExt);
    U16 crtval = 0;
    U16 wIntSts = 0;

    switch(pCtr->wState) {
        case 0:
            crtval = pCtr->dwCounterValue & 0xffff;
            outW(dwPort+ADL_PT_WDT, crtval);
            pDevExt->dwDevRB &= 0xffff0000;
            pDevExt->dwDevRB |= crtval;
            pDevExt->dwDevRB |= (pCtr->wMode<<16);
            break;
        case 1:
            wIntSts = PT_inpw(PT_Port(pDevExt), INTCSR);
            switch(pCtr->wMode) {
                case 0://disable
                    PT_outw(dwPort, INTCSR, wIntSts & 0x0700);
                    if(pDevExt->pTcn)
                    fasync_helper( (int)inode, file, 0, &(pDevExt->pTcn->async_queue1) );
                    break;
                case 1://enable
                case 2://restart
                    if(!(pDevExt->dwDevRB & 0xffff))
                        return 0;
                    if(pDevExt->pTcn) {
                        if((pDevExt->dwDevRB>>16) & 0x01)
                            fasync_helper( (int)inode, file, 1, &(pDevExt->pTcn->async_queue1) );
                    }
                    PT_outw(dwPort, INTCSR, wIntSts | 0x8000);
                    P9030_SetInt(OP_Port(pDevExt), 1, 1, 1, 0, 1);
                    P9030_SetInt(OP_Port(pDevExt), 1, 1, 1, 0, 0);
                    break;
                default:
                    return 0;
            }
            break;
    }
    return TRUE;
}
/*-------------------------------------------------------*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
int p7260_ioctl( 
	struct inode *inode,  
	struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
#else
long p7260_ioctl( 
	struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
	struct inode  *inode = file->f_mapping->host;
#endif
    PPCI_DEVEXT  pDevExt = file->private_data;
    void* temp_param;
    short error_code;
    U32 dwPort, vi, dwTemp;

    dwPort = PT_Port(pDevExt);

    switch ( ioctl_num ){
        case DAS_IOC_GET_DESP :
            temp_param = (void *)kmalloc( sizeof( DAS_IOT_CARD_DESP ), GFP_ATOMIC );
            if (! AdlIocGetDesp(pDevExt, (DAS_IOT_CARD_DESP*)temp_param )){
                error_code = ((DAS_IOT_CARD_DESP*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_CARD_DESP))){
                    kfree(temp_param);
                    return -EFAULT;
                 }
                 kfree( temp_param );
                 return error_code;
            }
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_CARD_DESP))){
                 kfree(temp_param);
                 return -EFAULT;
            }
            kfree( temp_param );
            break;
        //yuan add 05/10/01
        case DAS_IOC_GET_RESRC :
            temp_param = (void *)kmalloc( sizeof( DAS_IOT_RESRC ), GFP_ATOMIC );
            for(vi=0; vi<4; vi++)
                ((DAS_IOT_RESRC*)temp_param)->PortAddr[vi] = (U16) (pDevExt->dwPortAddr[vi]);
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_RESRC)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_DUALINT_SET :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_DUALINT), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_DUALINT)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ! pDevExt->pDin ) {
               ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_NO_FUNCTION;
               if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT)))
               {
                    kfree(temp_param);
                    return -EFAULT;
               }
               kfree(temp_param);
               return -EFAULT;
            }

            // chained the Signal request to fasync_queue , before the AdlSetDualInt
            if( ((DAS_IOT_DUALINT *) temp_param)->bInt1Status )
            {
                // change of SIGNAL async_queue is needed , fa->magic will be set to FASYNC_MAGIC
                if( fasync_helper( (int)inode, file, ((DAS_IOT_DUALINT *)temp_param)->bInt1Enable? 1:0, &(pDevExt->pDin->async_queue1) ) == -ENOMEM )
                {
                    // no enough memory for kmalloc or no matched
                    ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_POOL_ALLOC;
                    if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT))){
                        kfree(temp_param);
                        return -EFAULT;
                    }
                    kfree(temp_param);
                    return -1;
                }
            }

            if( ((DAS_IOT_DUALINT *) temp_param)->bInt2Status )
            { // change of SIGNAL async_queue is needed
                if( (fasync_helper( (int)inode, file, ((DAS_IOT_DUALINT *) temp_param)->bInt2Enable? 1:0, \
                    &(pDevExt->pDin->async_queue2)) ) == -ENOMEM ){
                    // no enough memory for kmalloc or no matched
                    // clear the linked fasync_queue before
                    if( ((DAS_IOT_DUALINT *) temp_param)->bInt1Status && ((DAS_IOT_DUALINT *)temp_param)->bInt1Enable )
                    {
                        fasync_helper( (int)inode, file, 0, &(pDevExt->pDin->async_queue1) );
                    }

                    ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_POOL_ALLOC;
                    if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT))){
                        kfree(temp_param);
                        return -EFAULT;
                    }
                    kfree(temp_param);
                    return -1;
                }
            }
            // yuan modify 11/30/01
            AdlSetDualInt(pDevExt, (DAS_IOT_DUALINT *) temp_param);
            //  P9030_IocSetDualInt(pDevExt, (DAS_IOT_DUALINT *) temp_param);
            //  PT_outw(dwPort, CLRIRQ, 0);
            kfree( temp_param );
            break;
        case DAS_IOC_DUALINT_GET :
           temp_param = (void *)kmalloc( sizeof(DAS_IOT_DUALINT), GFP_ATOMIC );
           if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_DUALINT)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ! pDevExt->pDin )
            {
                ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_NO_FUNCTION;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT)))
                {
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree(temp_param);
                return -EFAULT;
            }
            PLX90_IocGetDualInt(pDevExt, (DAS_IOT_DUALINT*)temp_param);

            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;

        case DAS_IOC_COSINT_SET :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_COSINT), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_COSINT)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            AdlSetCOSInt(pDevExt,(DAS_IOT_COSINT*)temp_param);
            kfree( temp_param );
            break;
        /*-----------------------------------------------------------*/
        case DAS_IOC_DI_PIO_READ :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ((DAS_IOT_PIO*)temp_param)->wChannel == PCI7260_CHANNEL_DI )
            {
                dwTemp = ADL_PT_DI;
                ((DAS_IOT_PIO*)temp_param)->dwValue = inpB( dwPort + dwTemp );
            }
            else if ( ((DAS_IOT_PIO*)temp_param)->wChannel == PCI7260_Init_DI ) {
                dwTemp = ADL_PT_INITPAT;
                ((DAS_IOT_PIO*)temp_param)->dwValue = inpB( dwPort + dwTemp );
                ((DAS_IOT_PIO*)temp_param)->dwValue = (~((DAS_IOT_PIO*)temp_param)->dwValue) & 0xff;
            }
            else if ( ((DAS_IOT_PIO*)temp_param)->wChannel == PCI7260_EMG_DI ) {
                dwTemp = ADL_PT_EMG_PAT;
                ((DAS_IOT_PIO*)temp_param)->dwValue = inpB( dwPort + dwTemp );
                ((DAS_IOT_PIO*)temp_param)->dwValue = (~(((DAS_IOT_PIO*)temp_param)->dwValue)) & 0xff;
            }
            else if ( ((DAS_IOT_PIO*)temp_param)->wChannel == PCI7260_COS_DI ) {
                //wTemp = ADL_PT_COSCSR;
                ((DAS_IOT_PIO*)temp_param)->dwValue = pDevExt->dwDevRA;
            }
            else
            {
                // dwTemp = ADL_PT_COSCSR;
                //((DAS_IOT_PIO*)temp_param)->dwValue = pDevExt->dwDevRA;
            }

            if ( ! AdlIocPioRead( 1, (DAS_IOT_PIO*)temp_param) )
            {
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO)))
                {
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_DO_PIO_READ :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            ((DAS_IOT_PIO*)temp_param)->dwValue = PT_inpb(dwPort, DIO);
            if ( ! AdlIocPioRead(1, (DAS_IOT_PIO*)temp_param) ){
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO)))
                {
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_DO_PIO_WRITE :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO)))
            {
                kfree(temp_param);
                return -EFAULT;
            }
            if(((DAS_IOT_PIO*)temp_param)->wByteCount==0x100) {
                //U32         ofs;
                U8         devid;
                struct pci_dev *pdev;
//20110629, workaround by Tim
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)                
		pdev = pci_find_slot (pDevExt->wBus, PCI_DEVFN(pDevExt->wSlot,pDevExt->wFunc));
#else
		pdev = pci_get_bus_and_slot (pDevExt->wBus, PCI_DEVFN(pDevExt->wSlot,pDevExt->wFunc));
#endif
                devid = (U8) (((DAS_IOT_PIO*)temp_param)->wChannel<<3);
                pci_write_config_byte( pdev, PCI_CS_HOTSWAP_CTRL, devid );
            }
            else if(((DAS_IOT_PIO*)temp_param)->wByteCount==0x200) {
                //yuan modify 06/21/05
                U8 ctrl=0;
                U16 intsts = PT_inpw(PT_Port(pDevExt), INTCSR);

                ctrl = (intsts>>11) & 0x01;
                switch(((DAS_IOT_PIO*)temp_param)->wChannel&0x0f) {
                    case 0:
                        ctrl= 0;
                        break;
                    case 1:
                        ctrl= 1;
                        break;
                    case 2:
                    default:
                        ctrl |= 0x80;
                        break;
                }
                PT_outb(dwPort, EMG_CTL, ctrl);
            }
            else {
                //if ( ! AdlPioWriteDIO(pDevExt, PSB_PIO) )
                    // goto IoctlError;
            //}
                if ( ! AdlPioWriteDIO(pDevExt, (DAS_IOT_PIO*)temp_param) )
                {
                    error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                    if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO)))
                    {
                        kfree(temp_param);
                        return -EFAULT;
                    }
                    kfree( temp_param );
                    return error_code;
                }
            }
            kfree( temp_param );
            break;
        case DAS_IOC_CTR_CONFIG :
            temp_param = (void *)kmalloc( sizeof( DAS_IOT_CTR ), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_CTR))){
                kfree(temp_param);
                return -EFAULT;
            }
            if (! WDT_IocCtrSet(inode, file, pDevExt, (DAS_IOT_CTR *) temp_param)) {
                error_code = ((DAS_IOT_CTR*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_CTR))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_CTR_READ :
            temp_param = (void *)kmalloc( sizeof( DAS_IOT_CTR ), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_CTR))){
                kfree(temp_param);
                return -EFAULT;
            }
            //yuan modify 06/26/05
            {
                U16 wIntSts = PT_inpw(PT_Port(pDevExt), INTCSR);
                switch(((DAS_IOT_CTR*)temp_param)->wMode) {
                    case 1:
                    ((DAS_IOT_CTR *)temp_param)->dwCounterValue = (wIntSts>>3) & 0x01;
                        break;
                    case 0:
                    default:
                        ((DAS_IOT_CTR *)temp_param)->dwCounterValue = (wIntSts>>4) & 0x01;
                        break;
                }
            }
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_CTR))){
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_PORT_READ :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PORTIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PORTIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            AdlIocPortIoRead(pDevExt, 1,(DAS_IOT_PORTIO*)temp_param);
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PORTIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_PORT_WRITE :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PORTIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PORTIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            AdlIocPortIoWrite(pDevExt,1,(DAS_IOT_PORTIO*)temp_param );
            kfree( temp_param );
            break;
        /*-----------------------------------------------------------*/
        default :
            return DAS_ERR_NO_FUNCTION;
    };

    return DAS_ERR_NO;

}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
struct file_operations fops = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
    ioctl: p7260_ioctl,   /*ioctl */
#else
    unlocked_ioctl: p7260_ioctl,   /*ioctl */
#endif
    open: p7260_open,
    release: p7260_release  /* a.k.a. close */
};

#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
struct file_operations fops = {
    NULL,   /* seek */
    NULL, // p7260_read,
    NULL, // p7260_write,
    NULL,   /* readdir */
    NUL0L,   /* select */
    p7260_ioctl,   /*ioctl */
    NULL,   /* mmap */
    p7260_open,
    NULL,  /* flush */
    p7260_release  /* a.k.a. close */
};

#endif

static int __devinit pci7260_init(struct pci_dev *pciptr, const struct pci_device_id *ent)
{
    int err=0;
    int t;
    int irq_result = 0;
    //PPCI_DBX pDbx;

    memset(&(DevExtension[device_number]), 0, sizeof(PCI_DEVEXT));
    DevExtension[device_number].pDrvExt = &drv_ext;

    err = AdlInitDevice(ADL_VENDOR_ID, ADL_DEVICE_ID, ADL_PCI_CTL, 2, \
        &(DevExtension[device_number]), pciptr, device_number
    );
    if(err<0){
        printk("[PCI-7260] pci7260_init: AdlInitDevice failed!\n");
        goto err_out;
    }

        for( t=0; t< PCI_PORT_MAX; t++ ){
            if( DevExtension[device_number].dwPortAddr[t]>0 && DevExtension[device_number].dwPortLen[t]>0 )
            {
                request_region( DevExtension[device_number].dwPortAddr[t], DevExtension[device_number].dwPortLen[t], "pci7260" );
            }
        }
        if( DevExtension[device_number].Irq_no != 0 )
        {
            irq_result = request_irq(DevExtension[device_number].Irq_no, p7260_interrupt_handler, SA_INTERRUPT | SA_SHIRQ, "pci7260", &DevExtension[device_number]);
            if( irq_result != 0 )
            { // unsuccess irq request
                DevExtension[device_number].openable = FALSE;
                //continue; // jump the Timer initialize , because this device is invalidate!!
                return -35;
            }
        }
        //if request IRQ failed !! what's the handler : exit or reject INT operation
        DevExtension[device_number].openable = TRUE;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
    // replace the code segment for __WAIT_QUEUE_HEAD_INITIALIZER
    init_waitqueue_head(&(DevExtension[device_number].sync_wq));
#endif

#ifdef CONFIG_DEVFS_FS
    sprintf(buf, "pci7260W%ld", device_number);
    #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
        err = devfs_mk_cdev(MKDEV(MAJOR, device_number), \
            S_IFCHR|S_IRUGO|S_IWUGO, buf
        );
        //printk("devfs_mk_cdev error: %d\n", err);
    #else if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
        DevExtension[device_number].devfs_handle = \
            devfs_register(NULL, buf, DEVFS_FL_NONE, MAJOR, device_number, \
                S_IFCHR|S_IRUSR|S_IWUSR, &pci7260_fops, NULL
            );
        if(DevExtension[device_number].devfs_handle==NULL){
            printk("[PCI-7260] pci7260_init: devfs_register( pci7260W%ld ) failed!!\n", device_number);
            //goto iounmap_out;
        }
    #endif
#endif
    //pciptr->driver_data = &(DevExtension[device_number];
    pci_set_drvdata(pciptr, &(DevExtension[device_number]));
    printk(KERN_ALERT "[PCI-7260] pci7260_init: Inserting pci7260 device: %d \n", DevExtension[device_number].wCard);

    /* register your own device in sysfs, and this will cause udevd to create corresponding device node */
    device_create(pci7260_class, NULL, MKDEV(MAJOR, DevExtension[device_number].wCard), NULL, \
        DUMMY_NAME "%d", DevExtension[device_number].wCard
    );
    device_number++;
    return 0;

err_out:
    pci_disable_device(pciptr);
    return -ENODEV;
}

void pci7260_remove(struct pci_dev *pdev)
{
    int t;
    PPCI_DEVEXT pDevExt;
    //PPCI_DBX pDbx;

    pDevExt = pci_get_drvdata(pdev);

        // free_irq
        if( pDevExt->Irq_no != 0 && pDevExt->openable == TRUE )
            // pci ios asign some IRQ for 7248 , free the IRQ
            free_irq( pDevExt->Irq_no , pDevExt);
        for( t=0; t< PCI_PORT_MAX; t++ )
        {
            if( pDevExt->dwPortAddr[t]>0 && pDevExt->dwPortLen[t]>0 )
            {
                release_region( pDevExt->dwPortAddr[t], pDevExt->dwPortLen[t] );
            }
        }
        if( pDevExt->pDin )
            AdlDropDsen( pDevExt->pDin );
        if( pDevExt->pTcn )
            AdlDropDsen( pDevExt->pTcn );

    pci_set_drvdata(pdev, NULL);
#ifdef CONFIG_DEVFS_FS
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
    devfs_remove("pci7260W%d", pDevExt->wCard);
#else if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
    if(pDevExt->devfs_handle != NULL){
        devfs_unregister((devfs_handle_t)(pDevExt->devfs_handle));
        pDevExt->devfs_handle = NULL;
    }
#endif
#endif

    pci_disable_device(pdev);
    printk(KERN_ALERT "[PCI-7260] Removing PCI-7260 device: %d\n", pDevExt->wCard);
    device_destroy(pci7260_class, MKDEV(MAJOR, pDevExt->wCard));
    device_number--;
}

int pci7260_module_init(void)
{
    int rel_val;
    int vi;
    DAS_IOT_FUNC_DESP * pFdsp;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
    EXPORT_NO_SYMBOLS;
#endif

    memset( &drv_ext, '\0', sizeof( PCI_DRVEXT ));
    memcpy(&(drv_ext.CardDesp),&iniCardDesp, sizeof( DAS_IOT_CARD_DESP ) );

    pFdsp = drv_ext.CardDesp.mFunc;
    for (vi=0; vi< DAS_FUNC_COUNT; vi++, pFdsp++) {
        if ( pFdsp->wArraySize > DAS_CHANNEL_GAIN_COUNT )
            pFdsp->wArraySize = DAS_CHANNEL_GAIN_COUNT;
    }

    rel_val = register_chrdev(MAJOR, DEVICE_NAME, &fops);
    if(rel_val<0){
        printk("[PCI-7260] pci7260_module_init: Sorry, registering the character device");
        return rel_val;
    }
    MAJOR = rel_val;

    printk("[PCI-7260] pci7260_module_init: Registering the character device: %s successful \n", DEVICE_NAME);
    pci7260_class = class_create(THIS_MODULE, "pci7260");
    if(IS_ERR(pci7260_class)){
        printk("Err: failed in creating class.\n");
    #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
        rel_val = unregister_chrdev(MAJOR, DEVICE_NAME);
        if(rel_val<0)
            printk(KERN_ALERT "[PCI-7260] pci7260_module_init: Error in module_unregister_chrdev: %d\n", rel_val);
        else
    #else
        unregister_chrdev(MAJOR, DEVICE_NAME);
    #endif
        return -1;
    }
    return pci_module_init(&pci7260_driver);
}

void pci7260_module_exit(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
    int ret;
#endif
    device_number = 0;

    pci_unregister_driver(&pci7260_driver);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
    ret = unregister_chrdev(MAJOR, DEVICE_NAME);
    if(ret<0)
        printk(KERN_ALERT "[PCI-7260] pci7260_module_exit: Error in module_unregister_chrdev: %d\n", ret);
    else
#else
    unregister_chrdev(MAJOR, DEVICE_NAME);
#endif

    printk("[PCI-7260] pci7260_module_exit: Unregistering the character device: %s\n", DEVICE_NAME);
    class_destroy(pci7260_class);
}

module_init(pci7260_module_init);
module_exit(pci7260_module_exit);
