#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"


extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 DIO_PortConfig(
    U16 wCardNumber,
    U16 wPort,
    U16 wDirection
    )
{
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    Check_If_Card_Registered(wCardNumber);
    //Check wPort
    if((wDirection!=INPUT_PORT)&&(wDirection!=OUTPUT_PORT))
        return ErrorInvalidPortDirection;

    if(CurrentCard.wCardType==PCI_7442)
        iotConf.mData.mDIO.wChannel = wPort-2;
    else
        iotConf.mData.mDIO.wChannel = wPort;
    iotConf.mData.mDIO.fDirectionIn = ((wDirection==INPUT_PORT)? 1:0);
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DIO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return ErrorConfigFailed;
        //return iotConf.wErrorCode;
    }

    if(CurrentCard.wCardType==PCI_7350){
        CurrentCard.DI_Status.CurrentConfig.mData.m7350DIO.PortConfig = \
        CurrentCard.DO_Status.CurrentConfig.mData.m7350DIO.PortConfig = \
            iotConf.mData.m7350DIO.PortConfig;
    }

    return NoError;
}


//06/09/06 add for 7442 TTL line config
I16 DIO_LineConfig(
    U16 wCardNumber,
    U16 wPort,
    U16 wLine,
    U16 wDirection
    )
{
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    Check_If_Card_Registered(wCardNumber);
    if((wDirection!=INPUT_PORT)&&(wDirection!=OUTPUT_PORT))
        return ErrorInvalidPortDirection;

    //yuan 07/31/00
    if(CurrentCard.wCardType==PCI_7442)
        iotConf.mData.m7442LINECONFIG.wChannel = wPort-2;
    else
        iotConf.mData.m7442LINECONFIG.wChannel = wPort;
    iotConf.mData.m7442LINECONFIG.wLine = wLine;
    iotConf.mData.m7442LINECONFIG.wDirectionIn = ((wDirection==INPUT_PORT)?1:0);
    iotConf.mData.m7442LINECONFIG.fLine = 1; //set to line config (01)
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DIO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return ErrorConfigFailed;
    }

    return NoError;
}


//06/09/06 add for 7442 TTL lines config
I16 DIO_LinesConfig(
    U16 wCardNumber,
    U16 wPort,
    U16 wLinesmap
    )
{
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    Check_If_Card_Registered(wCardNumber);

    if(CurrentCard.wCardType==PCI_7442)
        iotConf.mData.m7442LINECONFIG.wChannel = wPort-2;
    else
        iotConf.mData.m7442LINECONFIG.wChannel = wPort;
    iotConf.mData.m7442LINECONFIG.wDirectionIn = wLinesmap;
    iotConf.mData.m7442LINECONFIG.fLine = 3; //set to lines config (11)
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DIO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return ErrorConfigFailed;
    }

    return NoError;
}


#ifdef WIN32
I16 DIO_INT1_EventMessage(
    U16    wCardNumber,
    I16    wInt1Mode,
    HANDLE windowHandle,
    U32    message,
    void   *callbackAddr()
    )
{
    U16 err;
    U32 threadID;
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_DUALINT iotDint;

    memset(&iotDint, '\0', sizeof(DAS_IOT_DUALINT));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(VerID){
        case 2:
        case 1://Window NT system
            CurrentCard.Event_Msg[0].bEnabled = 0;
            if(wInt1Mode>=0){
                CurrentCard.DIOINT_Status.bInt1Enable = iotDint.bInt1Enable= 1;
                if((CurrentCard.wCardType!=PCI_7233)&&
                    (CurrentCard.wCardType!=PCI_7396))
                {
                    CurrentCard.DIOINT_Status.bInt1ActiveHigh = \
                        iotDint.bInt1ActiveHigh = 1;
                }
                else{
                    CurrentCard.DIOINT_Status.bInt1ActiveHigh = \
                        iotDint.bInt1ActiveHigh = 0;
                }
                if(CurrentCard.wCardType==PCI_7396){
                    switch(wInt1Mode){
                        case INT1_FP1C0:
                            wInt1Mode = INT1_RP1C0_FP1C3;
                            break;
                        case INT1_RP1C0_FP1C3:
                            wInt1Mode = INT1_FP1C0;
                            break;
                        default:
                            break;
                    }
                }
                //yuan add 11/30/01
                if((CurrentCard.wCardType==PCI_7256)||
                    (CurrentCard.wCardType==PCI_7260))
                {
                    switch(wInt1Mode){
                        case INT1_COS:
                            wInt1Mode = 1;
                            break;
                        case INT1_CH0:
                            wInt1Mode = 2;
                            break;
                        default:
                            break;
                    }
                }
                CurrentCard.DIOINT_Status.bInt1Mode = iotDint.bInt1Mode = \
                    (U8)wInt1Mode;
            }
            else{
                CurrentCard.DIOINT_Status.bInt1Enable = iotDint.bInt1Enable = 0;
                if(CurrentCard.wCardType!=PCI_7396)
                    CurrentCard.DIOINT_Status.bInt1Mode = iotDint.bInt1Mode = 0;
                else
                    CurrentCard.DIOINT_Status.bInt1Mode = iotDint.bInt1Mode = 4;
            }
            iotDint.bInt2Enable = CurrentCard.DIOINT_Status.bInt2Enable;
            iotDint.bInt2ActiveHigh = \
                CurrentCard.DIOINT_Status.bInt2ActiveHigh;
            if((!(CurrentCard.DIOINT_Status.bInt2Enable))&&
                (CurrentCard.wCardType==PCI_7396))
            {
                CurrentCard.DIOINT_Status.bInt2Mode = 4;
            }
            iotDint.bInt2Mode = CurrentCard.DIOINT_Status.bInt2Mode;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DUALINT_SET, &(iotDint))
                !=DAS_ERR_NO)
            {
                return iotDint.wErrorCode;
            }

            if(CurrentCard.hIntThread[0]){
                if(RegisterCards[wCardNumber].hIntEvent[0])
                    SetEvent(RegisterCards[wCardNumber].hIntEvent[0]);
                while(!CurrentCard.Event_Msg[0].bCompleted);
                CloseHandle(CurrentCard.hIntThread[0]);
                CurrentCard.hIntThread[0] = NULL;
            }
            if(wInt1Mode>=0){
                CurrentCard.Event_Msg[0].wCardNumber = wCardNumber;
                CurrentCard.Event_Msg[0].wInt1or2 = 0;  //INT1
                CurrentCard.Event_Msg[0].windowHandle = windowHandle;
                CurrentCard.Event_Msg[0].message = message;
                CurrentCard.Event_Msg[0].callbackAddr = callbackAddr;
                CurrentCard.Event_Msg[0].bEnabled = \
                    CurrentCard.DIOINT_Status.bInt1Enable;
                CurrentCard.hIntThread[0] = \
                    (HANDLE)CreateThread(NULL, 0, IntThread,
                        &(CurrentCard.Event_Msg[0]), 0, &threadID
                    );
            }
            break;
     }
     return NoError;
}


I16 DIO_INT2_EventMessage(
    U16    wCardNumber,
    I16    wInt2Mode,
    HANDLE windowHandle,
    U32    message,
    void  *callbackAddr()
    )
{
    U16 err;
    U32 threadID;
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_DUALINT iotDint;

    memset(&iotDint, '\0', sizeof(DAS_IOT_DUALINT));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(VerID){
        case 2:
        case 1: //Windows NT system
            CurrentCard.Event_Msg[1].bEnabled = 0;
            if(wInt2Mode>=0){
                CurrentCard.DIOINT_Status.bInt2Enable = iotDint.bInt2Enable =1;
                if((CurrentCard.wCardType!=PCI_7233)&&
                    (CurrentCard.wCardType!=PCI_7396))
                {
                    CurrentCard.DIOINT_Status.bInt2ActiveHigh = \
                        iotDint.bInt2ActiveHigh = 1;
                }
                else{
                    CurrentCard.DIOINT_Status.bInt2ActiveHigh = \
                        iotDint.bInt2ActiveHigh = 0;
                }
                if(CurrentCard.wCardType==PCI_7396){
                    switch(wInt2Mode){
                        case INT2_FP2C0:
                            wInt2Mode = INT2_RP2C0_FP2C3;
                            break;
                        case INT2_RP2C0_FP2C3:
                            wInt2Mode = INT2_FP2C0;
                            break;
                        default:
                            break;
                    }
                }
                //yuan add 11/30/01
                if(CurrentCard.wCardType==PCI_7256){
                    switch(wInt2Mode){
                        case INT2_CH1:
                            wInt2Mode = 4;
                            break;
                        default:
                            break;
                    }
                }
                CurrentCard.DIOINT_Status.bInt2Mode = iotDint.bInt2Mode = \
                    (U8)wInt2Mode;
            }
            else{
                CurrentCard.DIOINT_Status.bInt2Enable = \
                    iotDint.bInt2Enable = 0;
                if(CurrentCard.wCardType!=PCI_7396)
                    CurrentCard.DIOINT_Status.bInt2Mode = iotDint.bInt2Mode =0;
                else
                    CurrentCard.DIOINT_Status.bInt2Mode = iotDint.bInt2Mode =4;
            }
            iotDint.bInt1Enable = CurrentCard.DIOINT_Status.bInt1Enable;
            iotDint.bInt1ActiveHigh = CurrentCard.DIOINT_Status.bInt1ActiveHigh;
            if((!(CurrentCard.DIOINT_Status.bInt1Enable))&&
                (CurrentCard.wCardType==PCI_7396))
            {
                CurrentCard.DIOINT_Status.bInt1Mode = 4;
            }
            iotDint.bInt1Mode = CurrentCard.DIOINT_Status.bInt1Mode;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DUALINT_SET, &(iotDint))
                !=DAS_ERR_NO)
            {
                return iotDint.wErrorCode;
            }

            if(CurrentCard.hIntThread[1]){
                if(RegisterCards[wCardNumber].hIntEvent[1])
                    SetEvent(RegisterCards[wCardNumber].hIntEvent[1]);
                while(!CurrentCard.Event_Msg[1].bCompleted);
                CloseHandle(CurrentCard.hIntThread[1]);
            }
            if(wInt2Mode>=0){
                CurrentCard.Event_Msg[1].wCardNumber = wCardNumber;
                CurrentCard.Event_Msg[1].wInt1or2 = 1;  //INT2
                CurrentCard.Event_Msg[1].windowHandle = windowHandle;
                CurrentCard.Event_Msg[1].message = message;
                CurrentCard.Event_Msg[1].callbackAddr = callbackAddr;
                CurrentCard.Event_Msg[1].bEnabled = \
                    CurrentCard.DIOINT_Status.bInt2Enable;
                CurrentCard.hIntThread[1] = \
                    (HANDLE)CreateThread(NULL, 0, IntThread,
                        &(CurrentCard.Event_Msg[1]), 0, &threadID
                    );
            }
            break;
    }
    return NoError;
}


long WINAPI IntThread(
    EVENT_MESSAGE_TYPE *pArg
    )
{
    DAS_IOT_WAITEVT waitevt;
    OVERLAPPED lap;
    int returnval ;
    HANDLE hEvent;

    switch(VerID){
        case 2:
        case 1:
            pArg->bCompleted =0;
            while(pArg->bEnabled){
                WaitForSingleObject(
                    RegisterCards[pArg->wCardNumber].hIntEvent[pArg->wInt1or2],
                    INFINITE
                );
                if(pArg->bEnabled){
                    if((pArg->windowHandle)&&(pArg->message))
                        PostMessage(pArg->windowHandle, pArg->message, 0, 0);
                    if(pArg->callbackAddr){
                        pArg->bCompleted =1;
                        (*(pArg->callbackAddr))();
                    }
                }
                else{
                    ResetEvent(RegisterCards[pArg->wCardNumber].hIntEvent[\
                        pArg->wInt1or2]
                    );
                }
            }
            pArg->bCompleted =1;
            break;
    }
}
#endif


I16 DIO_SetDualInterrupt(
    U16  wCardNumber,
    I16  wInt1Mode,
    I16  wInt2Mode,
    void (*event1_handler)(int),
    void (*event2_handler)(int)
    )
{
#define LOWBYTE(wValue) (BYTE)(wValue&0xff)
    U16 err;
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_DUALINT iotDint;

    memset(&iotDint, '\0', sizeof(DAS_IOT_DUALINT));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    if(wInt1Mode>=0){ //enable the INT1
        iotDint.bInt1Enable = 1;
        if((CurrentCard.wCardType!=PCI_7233)&&
            (CurrentCard.wCardType!=PCI_7396))
        {
            iotDint.bInt1ActiveHigh = 1;
        }
        else
            iotDint.bInt1ActiveHigh = 0;

        if(CurrentCard.wCardType==PCI_7396){
            switch(wInt1Mode){
                case INT1_FP1C0:
                    wInt1Mode = INT1_RP1C0_FP1C3;
                    break;
                case INT1_RP1C0_FP1C3:
                    wInt1Mode = INT1_FP1C0;
                    break;
                default:
                    break;
            }
        }
        //yuan add 11/30/01
        if((CurrentCard.wCardType==PCI_7256)||
            (CurrentCard.wCardType==PCI_7260))
        {
            switch(wInt1Mode){
                case INT1_COS:
                    wInt1Mode = 1;
                    break;
                case INT1_CH0:
                    wInt1Mode = 2;
                    break;
                default:
                    break;
            }
        }

        iotDint.bInt1Mode = (U8)wInt1Mode;
        //process signal handler
        if(event1_handler!=NULL){
            if(CurrentCard.wCardType==PCI_7452){
                if(wInt1Mode&0x1){
                    if(CurrentCard.sig_act1.sa_handler==SIG_IGN){ //
                        iotDint.bInt1Status = TRUE; //fasync_helper is needed
                        CurrentCard.sig_act1.sa_handler = event1_handler;
                        CurrentCard.sig_act1.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)
                            ==-1)
                        {
                            //ignore the signal SIGEVENT1
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        //signal_handle is asigned , so ignore the signal
                        iotDint.bInt1Status = FALSE; //ignore this signal
                    }
                }
                else if(wInt1Mode&0x2){
                    if(CurrentCard.sig_act2.sa_handler==SIG_IGN){ //
                        iotDint.bInt1Status = TRUE; //fasync_helper is needed
                        CurrentCard.sig_act2.sa_handler = event1_handler;
                        CurrentCard.sig_act2.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)
                            ==-1)
                        {
                            //ignore the signal SIGEVENT1
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        //signal_handle is asigned , so ignore the signal
                        iotDint.bInt1Status = FALSE; //ignore this signal
                    }
                }
                else if(wInt1Mode&0x4){
                    if(CurrentCard.i_sig_act.sa_handler==SIG_IGN){//
                        iotDint.bInt1Status = TRUE; //fasync_helper is needed
                        CurrentCard.i_sig_act.sa_handler = event1_handler;
                        CurrentCard.i_sig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act),
                            NULL)==-1)
                        {
                            //ignore the signal SIGEVENT1
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        //signal_handle is asigned , so ignore the signal
                        iotDint.bInt1Status = FALSE; //ignore this signal
                    }
                }
                else if(wInt1Mode&0x8){
                    if(CurrentCard.o_sig_act.sa_handler==SIG_IGN){ //
                        iotDint.bInt1Status = TRUE; //fasync_helper is needed
                        CurrentCard.o_sig_act.sa_handler = event1_handler;
                        CurrentCard.o_sig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act),
                            NULL)==-1)
                        {
                            //ignore the signal SIGEVENT1
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        //signal_handle is asigned , so ignore the signal
                        iotDint.bInt1Status = FALSE; //ignore this signal
                    }
                }
                else if(wInt1Mode&0x10){
                    if(CurrentCard.i_dbsig_act.sa_handler==SIG_IGN){ //
                        iotDint.bInt1Status = TRUE; //fasync_helper is needed
                        CurrentCard.i_dbsig_act.sa_handler = event1_handler;
                        CurrentCard.i_dbsig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL)
                            ==-1)
                        {
                            //ignore the signal SIGEVENT1
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        //signal_handle is asigned , so ignore the signal
                        iotDint.bInt1Status = FALSE; //ignore this signal
                    }
                }
            }
            else if(CurrentCard.wCardType==PCI_7442){
                if(wInt1Mode&0x1){
                    if(CurrentCard.sig_act1.sa_handler==SIG_IGN){
                        iotDint.bInt1Status = TRUE;
                        CurrentCard.sig_act1.sa_handler = event1_handler;
                        CurrentCard.sig_act1.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        iotDint.bInt1Status = FALSE;
                    }
                }
                else if(wInt1Mode&0x2){
                    if(CurrentCard.sig_act2.sa_handler==SIG_IGN){
                        iotDint.bInt1Status = TRUE;
                        CurrentCard.sig_act2.sa_handler = event1_handler;
                        CurrentCard.sig_act2.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        iotDint.bInt1Status = FALSE;
                    }
                }
            }
            else if(CurrentCard.wCardType==PCI_7443){
                if(wInt1Mode&0x1){
                    if(CurrentCard.sig_act1.sa_handler==SIG_IGN){
                        iotDint.bInt1Status = TRUE;
                        CurrentCard.sig_act1.sa_handler = event1_handler;
                        CurrentCard.sig_act1.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        iotDint.bInt1Status = FALSE;
                    }
                }
                else if(wInt1Mode&0x2){
                    if(CurrentCard.sig_act2.sa_handler==SIG_IGN){
                        iotDint.bInt1Status = TRUE;
                        CurrentCard.sig_act2.sa_handler = event1_handler;
                        CurrentCard.sig_act2.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        iotDint.bInt1Status = FALSE;
                    }
                }
                else if(wInt1Mode&0x4){
                    if(CurrentCard.i_sig_act.sa_handler==SIG_IGN){
                        iotDint.bInt1Status = TRUE;
                        CurrentCard.i_sig_act.sa_handler = event1_handler;
                        CurrentCard.i_sig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        iotDint.bInt1Status = FALSE;
                    }
                }
                else if(wInt1Mode&0x8){
                    if(CurrentCard.o_sig_act.sa_handler==SIG_IGN){
                        iotDint.bInt1Status = TRUE;
                        CurrentCard.o_sig_act.sa_handler = event1_handler;
                        CurrentCard.o_sig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        iotDint.bInt1Status = FALSE;
                    }
                }
            }
            else{
                if(CurrentCard.sig_act1.sa_handler==SIG_IGN){
                    iotDint.bInt1Status = TRUE; //fasync_helper is needed
                    CurrentCard.sig_act1.sa_handler = event1_handler;
                    //for termio read in function kb_hit()
                    CurrentCard.sig_act1.sa_flags |= SA_RESTART;
                    if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)==-1)
                        return ErrorOpenEventFailed;
                }
                else{ //signal_handle is asigned , so ignore the signal setting
                    iotDint.bInt1Status = FALSE; //ignore this signal setting
                }
            }
        }
        else{ //only ENABLE INT , no signal is needed, no handler
            iotDint.bInt1Status = FALSE; //ignore this signal setting
        }
    }
    else{ //disable the INT1, Int1Mode < 0
        iotDint.bInt1Enable = 0;
        iotDint.bInt1ActiveHigh = 1;
        if(CurrentCard.wCardType!=PCI_7396)
            iotDint.bInt1Mode = 0;
        else
            iotDint.bInt1Mode = 4;
        //reset the signal handler
        if(CurrentCard.wCardType==PCI_7452){
            if(CurrentCard.sig_act1.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.sig_act1.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.sig_act2.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.sig_act2.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.i_sig_act.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.i_sig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.o_sig_act.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.o_sig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.i_dbsig_act.sa_handler != SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.i_dbsig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
        }
        else if(CurrentCard.wCardType==PCI_7442){
            if(CurrentCard.sig_act1.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.sig_act1.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL )==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.sig_act2.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.sig_act2.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
        }
        else if(CurrentCard.wCardType==PCI_7443){
            if(CurrentCard.sig_act1.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.sig_act1.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.sig_act2.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.sig_act2.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.i_sig_act.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.i_sig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
            if(CurrentCard.o_sig_act.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.o_sig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
        }
        else {
            if(CurrentCard.sig_act1.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.sig_act1.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)==-1){
                    //ignore the signal SIGEVENT1
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
        }
    }

    if(wInt2Mode>=0){ //enable the INT2
        iotDint.bInt2Enable = 1;
        if((CurrentCard.wCardType!=PCI_7233)&&
            (CurrentCard.wCardType!=PCI_7396))
        {
            iotDint.bInt2ActiveHigh = 1;
        }
        else
            iotDint.bInt2ActiveHigh = 0;
        if(CurrentCard.wCardType==PCI_7396){
            switch(wInt2Mode){
                case INT2_FP2C0:
                    wInt2Mode = INT2_RP2C0_FP2C3;
                    break;
                case INT2_RP2C0_FP2C3:
                    wInt2Mode = INT2_FP2C0;
                    break;
                default:
                    break;
            }
        }
        //yuan add 11/30/01
        if((CurrentCard.wCardType==PCI_7256)||
            (CurrentCard.wCardType==PCI_7260))
        {
            switch(wInt2Mode){
                case INT2_CH1:
                    wInt2Mode = 4;
                    break;
                default:
                    break;
            }
        }
        iotDint.bInt2Mode = (U8)wInt2Mode;
        //process signal handler

        if(event2_handler!=NULL){
            if((CurrentCard.wCardType==PCI_7442)||
                (CurrentCard.wCardType==PCI_7444))
            {
                if(wInt2Mode&0x04){
                    if(CurrentCard.i_dbsig_act.sa_handler==SIG_IGN){ //
                        iotDint.bInt2Status = TRUE;
                        CurrentCard.i_dbsig_act.sa_handler = event2_handler;
                        CurrentCard.i_dbsig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    else{
                        iotDint.bInt2Status = FALSE;
                    }
                }
            }
            else{
                if(CurrentCard.sig_act2.sa_handler==SIG_IGN){
                    iotDint.bInt2Status = TRUE;
                    CurrentCard.sig_act2.sa_handler = event2_handler;
                    CurrentCard.sig_act2.sa_flags |= SA_RESTART;
                    CurrentCard.sig_act2.sa_flags |= SA_NODEFER;
                    if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1)
                        return ErrorOpenEventFailed;
                }
                else{
                    iotDint.bInt2Status = FALSE;
                }
            }
        }
        else{
            iotDint.bInt2Status = FALSE;
        }
    }
    else{ //disable the INT2
        iotDint.bInt2Enable = 0;
        iotDint.bInt2ActiveHigh = 1;
        if(CurrentCard.wCardType!=PCI_7396)
            iotDint.bInt2Mode = 0;
        else
            iotDint.bInt2Mode = 4;

        //reset the signal handler
        if((CurrentCard.wCardType==PCI_7442)||
            (CurrentCard.wCardType==PCI_7444))
        {
            if(CurrentCard.i_dbsig_act.sa_handler!=SIG_IGN){
                iotDint.bInt2Status = TRUE;
                CurrentCard.i_dbsig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL)==-1){
                    return ErrorOpenEventFailed;
                }
            }
            else{
                iotDint.bInt2Status = FALSE;
            }
        }
        else{
            if(CurrentCard.sig_act2.sa_handler!=SIG_IGN){
                iotDint.bInt2Status = TRUE;
                CurrentCard.sig_act2.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1){
                    return ErrorOpenEventFailed;
                }
            }
            else{
                iotDint.bInt2Status = FALSE;
            }
        }
    }

    //send SIGIO to user also ...
    //asign the handler of SIGIO to be SIG_IGN
    //(kernel will send SIGIO when number of queued signal>1024)
    /*
    if(iotDint.bInt1Status||iotDint.bInt2Status){
        if((CurrentCard.sig_act1.sa_handler!=SIG_IGN||
            (CurrentCard.sig_act2.sa_handler!=SIG_IGN))
        {
            signal(SIGIO, SIG_IGN);
        }
        else
            signal(SIGIO, SIG_DFL);
    }
    */
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DUALINT_SET, &(iotDint))
        !=DAS_ERR_NO)
    {
        return iotDint.wErrorCode;
    }

    return NoError;
}


I16 DIO_7300SetInterrupt(
    U16  wCardNumber,
    I16  AuxDIEn,
    I16  T2En,
    void (*event1_handler)(int),
    void (*event2_handler)(int)
    )
{
#define LOWBYTE(wValue) (BYTE)(wValue&0xff)
    U16 err;
    DAS_IOT_DUALINT iotDint;

    memset(&iotDint, '\0', sizeof(DAS_IOT_DUALINT));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotDint.bInt1Enable = AuxDIEn;
    iotDint.bInt2Enable = T2En;

    if(AuxDIEn>0){ //enable the INT1
        if(event1_handler!=NULL){
            if(CurrentCard.sig_act1.sa_handler==SIG_IGN){
                //fasync_helper for INT1 is needed (in kernel level)
                iotDint.bInt1Status = TRUE;
                CurrentCard.sig_act1.sa_handler = event1_handler;
                //for termio read in function kb_hit()
                CurrentCard.sig_act1.sa_flags |= SA_RESTART;
                if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)==-1){
                    //some thing error
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE;
                //tell the driver ignore this signal setting
            }
        }
        else{ //only ENABLE INT , no signal is needed
            iotDint.bInt1Status = FALSE;
            //tell the driver ignore this signal setting
        }
    }
    else{ //disable the INT1
        //reset the signal handler
        if(CurrentCard.sig_act1.sa_handler!=SIG_IGN){ //
            //fasync_helper for INT1 is needed (in kernel level)
            iotDint.bInt1Status = TRUE;
            CurrentCard.sig_act1.sa_handler = SIG_IGN;
            if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)==-1){
                //ignore the signal SIGEVENT1
                return ErrorOpenEventFailed;
            }
        }
        else{ //signal_handle is asigned , so ignore the signal setting
            iotDint.bInt1Status = FALSE;
            //tell the driver ignore this signal setting
        }
    }
    if(T2En>0){ //enable the INT2
        //process signal handler
        if(event2_handler!=NULL){
            if(CurrentCard.sig_act2.sa_handler==SIG_IGN){
                //fasync_helper for INT1 is needed (in kernel level)
                iotDint.bInt2Status = TRUE;
                CurrentCard.sig_act2.sa_handler = event2_handler;
                //for termio read in function kb_hit()
                CurrentCard.sig_act2.sa_flags |= SA_RESTART;
                if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1){
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt2Status = FALSE;
                //tell the driver ignore this signal setting
            }
        }
        else{ //only ENABLE INT , no signal is needed
            iotDint.bInt2Status = FALSE;
            //tell the driver ignore this signal setting
        }
    }
    else{ //disable the INT2
        //reset the signal handler
        if(CurrentCard.sig_act2.sa_handler!=SIG_IGN){ //
            //fasync_helper for INT1 is needed (in kernel level)
            iotDint.bInt2Status = TRUE;
            CurrentCard.sig_act2.sa_handler = SIG_IGN;
            if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1){
                //ignore the signal SIGEVENT1
                return ErrorOpenEventFailed;
            }
        }
        else{ //signal_handle is asigned , so ignore the signal setting
            iotDint.bInt2Status = FALSE;
            //tell the driver ignore this signal setting
        }
    }

    //send SIGIO to user also ...
    //asign the handler of SIGIO to be SIG_IGN
    //(kernel will send SIGIO when number of queued signal>1024)
    /*
    if(iotDint.bInt1Status||iotDint.bInt2Status){
        if((CurrentCard.sig_act1.sa_handler!=SIG_IGN)||
            (CurrentCard.sig_act2.sa_handler!=SIG_IGN))
        {
            signal(SIGIO, SIG_IGN);
        }
        else
            signal(SIGIO, SIG_DFL);
    }
    */
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DUALINT_SET, &(iotDint))
        !=DAS_ERR_NO)
    {
        return iotDint.wErrorCode;
    }

    return NoError;
}


I16 DIO_SetCOSInterrupt(
    U16 wCardNumber,
    U16 Port,
    U16 ctlA,
    U16 ctlB,
    U16 ctlC
    )
{
    U16 err, cosPort;
    DAS_IOT_COSINT iotCInt;

    memset(&iotCInt, '\0', sizeof(DAS_IOT_COSINT));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType==PCI_7256){
        iotCInt.wPort = ctlA&0xff;
        iotCInt.mCOS.wValue = (ctlA>>8)&0xff;
    }
    else if(CurrentCard.wCardType==PCI_7260){
        iotCInt.wPort = 0;
        iotCInt.mCOS.wValue = ctlA&0xff;
    }
    else{
        switch(Port){
            case Channel_P1:
            case COS_P1:
                cosPort = COS_P1;
                break;
            case Channel_P2:
            case COS_P2:
                cosPort = COS_P2;
                break;
            case Channel_P3:
            case COS_P3:
                cosPort = COS_P3;
                break;
            case Channel_P4:
            case COS_P4:
                cosPort = COS_P4;
                break;
            default:
                return ErrorInvalidDioPort;
        }
        iotCInt.wPort = cosPort;
        iotCInt.mCOS.cosRb.P_3_A = ctlA;
        iotCInt.mCOS.cosRb.P_3_B = ctlB;
        iotCInt.mCOS.cosRb.P_3_C = ctlC;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_COSINT_SET, &(iotCInt))
        !=DAS_ERR_NO)
    {
        return iotCInt.wErrorCode;
    }

    return NoError;
}


I16 DIO_SetCOSInterrupt32(
    U16  wCardNumber,
    U8   Port,
    U32  ctl,
    void (*event_handler)(int)
    )
{
    U16 err;
    DAS_IOT_COSINT32 iotCInt;

    memset(&iotCInt, '\0', sizeof(DAS_IOT_COSINT32));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotCInt.bPort = Port;
    iotCInt.dwValue = ctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_COSINT_SET32, &(iotCInt))
        !=DAS_ERR_NO)
    {
        return iotCInt.wErrorCode;
    }

    if(!ctl){ //free signal
        //reset the signal handler
        switch(Port){
            case 0:
                if(CurrentCard.wCardType==PCI_7350){
                    if(CurrentCard.o_sig_act.sa_handler!=SIG_IGN){
                        CurrentCard.o_sig_act.sa_handler = SIG_IGN;
                        if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                }
                else{
                    if(CurrentCard.sig_act1.sa_handler!=SIG_IGN){
                        CurrentCard.sig_act1.sa_handler = SIG_IGN;
                        if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                }
                break;
            case 1:
                if(CurrentCard.sig_act2.sa_handler!=SIG_IGN){
                    CurrentCard.sig_act2.sa_handler = SIG_IGN;
                    if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)==-1)
                        return ErrorOpenEventFailed;
                }
                break;
            case 2:
                if(CurrentCard.i_sig_act.sa_handler!=SIG_IGN){
                    CurrentCard.i_sig_act.sa_handler = SIG_IGN;
                    if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL)
                        ==-1)
                    {
                        return ErrorOpenEventFailed;
                    }
                }
                break;
            case 3:
                if(CurrentCard.o_sig_act.sa_handler!=SIG_IGN){
                    CurrentCard.o_sig_act.sa_handler = SIG_IGN;
                    if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act), NULL)
                        ==-1)
                    {
                        return ErrorOpenEventFailed;
                    }
                }
                break;
        }
    }
    else{
        if(event_handler!=NULL){
            switch(Port){
                case 0:
                    if(CurrentCard.wCardType==PCI_7350){
                        if(CurrentCard.o_sig_act.sa_handler==SIG_IGN){
                            CurrentCard.o_sig_act.sa_handler = event_handler;
                            CurrentCard.o_sig_act.sa_flags |= SA_RESTART;
                            if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act),
                                NULL)==-1)
                            {
                                return ErrorOpenEventFailed;
                            }
                        }
                    }
                    else{
                        if(CurrentCard.sig_act1.sa_handler==SIG_IGN){
                            CurrentCard.sig_act1.sa_handler = event_handler;
                            CurrentCard.sig_act1.sa_flags |= SA_RESTART;
                            if(sigaction(SIGEVENT1, &(CurrentCard.sig_act1),
                                NULL)==-1)
                            {
                                return ErrorOpenEventFailed;
                            }
                        }
                    }
                    break;
                case 1:
                    if(CurrentCard.sig_act2.sa_handler==SIG_IGN){
                        CurrentCard.sig_act2.sa_handler = event_handler;
                        CurrentCard.sig_act2.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT2, &(CurrentCard.sig_act2), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    break;
                case 2:
                    if(CurrentCard.i_sig_act.sa_handler == SIG_IGN){
                        CurrentCard.i_sig_act.sa_handler = event_handler;
                        CurrentCard.i_sig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    break;
                case 3:
                    if(CurrentCard.o_sig_act.sa_handler==SIG_IGN){
                        CurrentCard.o_sig_act.sa_handler = event_handler;
                        CurrentCard.o_sig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT4, &(CurrentCard.o_sig_act), NULL)
                            ==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                    break;
            }
        }
    }

    return NoError;
}


I16 DIO_GetCOSLatchData(
    U16 wCardNumber,
    U16 *CosLData
    )
{
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((CurrentCard.wCardType!=PCI_7256)&&
        (CurrentCard.wCardType!=PCI_7260))
    {
        return ErrorFuncNotSupport;
    }

    iotPio.wChannel = 1;
    //iotPio.wByteCount = 2;
    iotPio.wByteCount = (U16)((CurrentCard_DIFunc.wResoluation+7)/8);
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    //printf("result: %x\n", iotPio.dwValue);
    *CosLData = (U16)iotPio.dwValue;

    return NoError;
}


I16 DIO_GetCOSLatchDataInt32(
    U16 wCardNumber,
    U8  Port,
    U32 *CosLData
    )
{
#define PCI7452_COS_DI 4
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    switch(CurrentCard.wCardType){
        case PCI_7452:
            iotPio.wChannel = Port+PCI7452_COS_DI;
            break;
        case PCI_7442:
            iotPio.wChannel = Port+8;
            break;
        case PCI_7443:
            iotPio.wChannel = Port+6; //6, 7, 8, or 9
            break;
        case PCI_7350:
            iotPio.wChannel = Port|0x8000;
            break;
        default:
            return ErrorFuncNotSupport;
    }
    iotPio.wByteCount = (U16)((CurrentCard_DIFunc.wResoluation+7)/8);
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    //printf("result: %x\n", iotPio.dwValue);
    *CosLData = (U32)iotPio.dwValue;

    return NoError;
}


//10/01/08 meet old declartion
I16 DIO_GetCOSLatchData32(
    U16 wCardNumber,
    U8  Port,
    U32 *CosLData
    )
{
    return DIO_GetCOSLatchDataInt32(wCardNumber, Port, CosLData);
}


I16 DIO_COSInterruptCounter(
    U16 wCardNumber,
    U16 Counter_Num,
    U16 Counter_Mode,
    U16 DI_Port,
    U16 DI_Line,
    U32 *Counter_Value
    )
{
    I16 err;
    DAS_IOT_COSINT_COUNT iotCInt = {0};

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    iotCInt.wCounter = Counter_Num;
    iotCInt.wMode = Counter_Mode; /*0: Reset, 1: Setup, 2: Start, 3: Stop, 4: Read*/
    iotCInt.wPort = DI_Port;
    iotCInt.wLine = DI_Line;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_COSINT_COUNT, &(iotCInt))
        !=DAS_ERR_NO)
    {
        return iotCInt.wErrorCode;
    }
    if(Counter_Mode==4)
        *Counter_Value = iotCInt.dwValue;

    return NoError;
}


I16 DIO_VoltLevelConfig(
    U16 wCardNumber,
    U16 wPortType,
    U16 wVoltLevel
    )
{
    I16 err;

    /*Error Check*/
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType==PCI_7350){
        U32 LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);
        U32 RegOffset = 0x90; //0x90: DIO&AFI Voltage Control Register
        U32 RegData = 0;

        /*Error Check*/
        if((wPortType!=P7350_PortDIO)&&(wPortType!=P7350_PortAFI))
            return ErrorInvalidDioPort;
        if(wVoltLevel>VoltLevel_1R8)
            return ErrorUndefinedParameter;

        /*Read Back DIO&AFI Voltage Control Register*/
        RegData = *((volatile U32 *)(LcrAddr+RegOffset));
        /*
         * bit 0~2 - DIO Port: 0(3.3V), 1(2.5V), 2(1.8V)
         * bit 4~6 - AFI Port: 0(3.3V), 1(2.5V), 2(1.8V)
         */
        RegData &= ~(0x7L<<(wPortType*4));
        RegData |= (wVoltLevel<<(wPortType*4));

        /*Set new configuration*/
        *((U32 *)(LcrAddr+RegOffset)) = RegData;
    }
    else
        return ErrorFuncNotSupport;

    return NoError;
}


I16 DIO_7350_AFIConfig(
    U16 wCardNumber,
    U16 wAFI_Port,
    U16 wAFI_Enable,
    U16 wAFI_Mode,
    U32 dwAFI_TrigOutLen
    )
{
    I16 err;
    U16 AfiDir = 0;
    U32 LcrAddr;
    U32 ConfRegOffset;
    U32 LengthRegOffset;
    U32 ConfRegData = 0;

    /*Error Check*/
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if(wAFI_Port<4){
        ConfRegOffset = 0x98; //AFI 0~3 Configuration Register
    }
    else if(wAFI_Port<8){
        ConfRegOffset = 0x9C; //AFI4~7 Configuration Register
        wAFI_Port -= 4;
    }
    else
        return ErrorInvalidDioPort;

    /*Get mapped IO memory address*/
    LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);
    /*Read back AFI configuration register*/
    ConfRegData = *((volatile U32 *)(LcrAddr+ConfRegOffset));
    /*Reset configurations of the specified AFI port*/
    ConfRegData &= ~(0xffL<<(8*wAFI_Port));
    if(wAFI_Enable==0){
        /*Disable*/
        ConfRegData |= (0x1L<<((8*wAFI_Port)+6));
        *((volatile U32 *)(LcrAddr+ConfRegOffset)) = ConfRegData;
    }
    else{
        /*Enable*/
        switch(wAFI_Mode){
            case 0:  /*P7350_AFI_DIStartTrig*/
            case 1:  /*P7350_AFI_DOStartTrig*/
            case 2:  /*P7350_AFI_DIPauseTrig*/
            case 3:  /*P7350_AFI_DOPauseTrig*/
            case 8:  /*P7350_AFI_HSDIREQ*/
            case 10: /*P7350_AFI_HSDITRIG*/
            case 12: /*P7350_AFI_HSDOACK*/
            case 13: /*P7350_AFI_HSDOTRIG*/
            case 16: /*P7350_POLL_DI*/
                /*Input*/
                AfiDir = 0;
                break;
            case 4:  /*P7350_AFI_DISWTrigOut*/
            case 5:  /*P7350_AFI_DOSWTrigOut*/
            case 6:  /*P7350_AFI_COSTrigOut*/
            case 7:  /*P7350_AFI_PMTrigOut*/
                /*Output*/
                AfiDir = 1;
                /*
                 * should set trigger pulse out length
                 * offset 0x110 - DI Software Trigger Out Pulse Length
                 * offset 0x114 - DO Software Trigger Out Pulse Length
                 * offset 0x118 - COS Event Trigger Out Pulse Length
                 * offset 0x11C - PM Event Trigger Out Pulse Length
                 */
                LengthRegOffset = 0x110+(wAFI_Mode-4)*0x04;
                *((volatile U32 *)(LcrAddr+LengthRegOffset)) = \
                    dwAFI_TrigOutLen;
                break;
            case 9:  /*P7350_AFI_HSDIACK*/
            case 11: /*P7350_AFI_HSDOREQ*/
            case 17: /*P7350_POLL_DO*/
                /*Output*/
                AfiDir = 1;
                break;
            case 14: /*P7350_AFI_SPI*/
            case 15: /*P7350_AFI_I2C*/
                break;
            default:
                return ErrorUndefinedParameter;
        }
        /*Set AFI mode*/
        ConfRegData |= ((wAFI_Mode&0x1f)<<(8*wAFI_Port));
        /*Enable AFI*/
        ConfRegData &= (~(0x1L<<((8*wAFI_Port)+6)));
        /*Configure AFI Direction*/
        ConfRegData |= (AfiDir<<((8*wAFI_Port)+7));

        *((volatile U32 *)(LcrAddr+ConfRegOffset)) = ConfRegData;
    }

    return NoError;
}


I16 DIO_PMConfig(
    U16 wCardNumber,
    U16 wChannel,
    U16 wPM_ChnEn,
    U16 wPM_ChnType
    )
{
    I16 err;
    DAS_IOT_PATTERN_MATCH iotPM;

    memset(&iotPM, '\0', sizeof(DAS_IOT_PATTERN_MATCH));

    /*Error Check*/
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }
    if(wPM_ChnEn){
        if(wPM_ChnType>PATMATCH_Edge_F)
            return ErrorUndefinedParameter;
    }

    iotPM.wPort = wChannel;
    iotPM.wMode = 1; //Let Kernel Know DIO_PMConfig will be performed
    iotPM.dwPattern1 = wPM_ChnEn;
    iotPM.dwPattern2 = wPM_ChnType;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_PMINT_SET, &(iotPM))
        !=DAS_ERR_NO)
    {
        return iotPM.wErrorCode;
    }

    return NoError;
}


I16 DIO_PMControl(
    U16  wCardNumber,
    U16  wPort,
    U16  wPM_Start,
    /*HANDLE  *hEvent,
    BOOLEAN bManualReset*/
    void (*event_handler)(int)
    )
{
    I16 err;
    DAS_IOT_PATTERN_MATCH iotPM;

    memset(&iotPM, '\0', sizeof(DAS_IOT_PATTERN_MATCH));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPM.wPort = wPort;
    //if(bManualReset)
    //    iotPM.wPort |= 0x8000;
    iotPM.wMode = 2; //Let Kernel Know DIO_PMControl will be performed
    iotPM.dwCtrl = wPM_Start;
    //iotPM.wErrorCode = ErrorConfigIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_PMINT_SET, &(iotPM))
        !=DAS_ERR_NO)
    {
        return iotPM.wErrorCode;
    }

    switch(CurrentCard.wCardType){
        case PCI_7350:
            if(!wPM_Start){ //free
                if(CurrentCard.i_sig_act.sa_handler!=SIG_IGN){
                    CurrentCard.i_sig_act.sa_handler = SIG_IGN;
                    if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL)
                        ==-1)
                    {
                        return ErrorOpenEventFailed;
                    }
                }
            }
            else{
                if(event_handler!=NULL){
                    if(CurrentCard.i_sig_act.sa_handler==SIG_IGN){
                        CurrentCard.i_sig_act.sa_handler = event_handler;
                        CurrentCard.i_sig_act.sa_flags |= SA_RESTART;
                        if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act),
                            NULL)==-1)
                        {
                            return ErrorOpenEventFailed;
                        }
                    }
                }
            }
            break;
        default:
            break;
    }

    return NoError;
}


I16 DIO_GetPMLatchData32(
    U16 wCardNumber,
    U16 wPort,
    U32 *PMLData
    )
{
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if(CurrentCard.wCardType==PCI_7350)
        iotPio.wChannel = wPort|0x4000;

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    *PMLData = (U32)iotPio.dwValue;

    return NoError;
}
