#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"


extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 CTR_Setup(
    U16 wCardNumber,
    U16 wCtr,
    U16 wMode,
    U32 dwCount,
    U16 wBinBcd
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    //yuan add 01/31/01 for eror checking
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(CurrentCard.wCardType==PCI_8554)
        wCtr -= 1;

    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;
    if((wMode!=TOGGLE_OUTPUT)&&(wMode!=PROG_ONE_SHOT)&&(wMode!=RATE_GENERATOR)&&
        (wMode!=SQ_WAVE_RATE_GENERATOR)&&(wMode!=SOFT_TRIG)&&(wMode!=HARD_TRIG))
    {
        return ErrorInvalidCounterMode;
    }
    if((wBinBcd!=BIN)&&(wBinBcd!=BCD))
        return ErrorInvalidBinBcdParam;

    iotCtr.wCounterIndex = wCtr;
    iotCtr.wMode = wMode;
    iotCtr.wBinBcd = wBinBcd;
    iotCtr.dwCounterValue = dwCount;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_CONFIG, &(iotCtr))!=DAS_ERR_NO)
        return iotCtr.wErrorCode;

    return NoError;
}


I16 CTR_Clear(
    U16 wCardNumber,
    U16 wCtr,
    U16 wState
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    //yuan add 01/31/01 for eror checking
    if(wCardNumber>= MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(CurrentCard.wCardType==PCI_8554)
        wCtr -= 1;

    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;
    if(wState!=0&&wState!=1)
        return ErrorInvalidCounterState;

    iotCtr.wCounterIndex = wCtr;
    iotCtr.wState = wState;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_RESET, &(iotCtr))!=DAS_ERR_NO)
        return iotCtr.wErrorCode;

    return NoError;
}


I16 CTR_Read(
    U16 wCardNumber,
    U16 wCtr,
    U32 *pValue
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    //yuan add 01/31/01 for eror checking
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(CurrentCard.wCardType==PCI_8554)
        wCtr -= 1;

    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotCtr.wCounterIndex = wCtr;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_READ, &(iotCtr))!=DAS_ERR_NO)
        return iotCtr.wErrorCode;
    *pValue = iotCtr.dwCounterValue;

    return NoError;
}


I16 CTR_Status(
    U16 wCardNumber,
    U16 wCtr,
    U32 *pValue
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    //yuan add 01/31/01 for eror checking
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(CurrentCard.wCardType==PCI_8554)
        wCtr -= 1;

    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotCtr.wCounterIndex = wCtr;
    iotCtr.wMode = 2;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_READ, &(iotCtr))!=DAS_ERR_NO)
        return iotCtr.wErrorCode;
    *pValue = iotCtr.dwCounterValue;

    return NoError;
}


//yuan add 03/11/03
I16 CTR_Update(
    U16 wCardNumber,
    U16 wCtr,
    U32 dwCount
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    //yuan add 01/31/01 for eror checking
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(CurrentCard.wCardType==PCI_8554)
        wCtr -= 1;

    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotCtr.wCounterIndex = wCtr;
    iotCtr.wMode = 1;
    iotCtr.dwCounterValue = dwCount;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_READ, &(iotCtr))!=DAS_ERR_NO)
        return iotCtr.wErrorCode;
    //*pValue = iotCtr.dwCounterValue;

    return NoError;
}


I16 CTR_8554_ClkSrc_Config(
    U16 wCardNumber,
    U16 wCtr,
    U16 ClockSource
    )
{
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(CurrentCard.wCardType!=PCI_8554)
        return ErrorFuncNotSupport;

    wCtr -= 1;

    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotConf = CurrentCard.CTR_Status.CurrentConfig;
    switch(wCtr){
        case 0:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN1_SRC = ClockSource;
            break;
        case 1:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN2_SRC = ClockSource;
            break;
        case 2:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN3_SRC = ClockSource;
            break;
        case 3:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN4_SRC = ClockSource;
            break;
        case 4:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN5_SRC = ClockSource;
            break;
        case 5:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN6_SRC = ClockSource;
            break;
        case 6:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN7_SRC = ClockSource;
            break;
        case 7:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN8_SRC = ClockSource;
            break;
        case 8:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN9_SRC = ClockSource;
            break;
        case 9:
            iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCN10_SRC = ClockSource;
            break;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_8554CTR_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    CurrentCard.CTR_Status.CurrentConfig = iotConf;

    return NoError;
}


I16 CTR_8554_CK1_Config(
    U16 wCardNumber,
    U16 ClockSource
    )
{
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(CurrentCard.wCardType!=PCI_8554)
        return ErrorFuncNotSupport;
    if(ClockSource>1)
        return ErrorInvalidCtrSource;

    iotConf = CurrentCard.CTR_Status.CurrentConfig;
    iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bCK1SEL = ClockSource;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_8554CTR_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    CurrentCard.CTR_Status.CurrentConfig = iotConf;

    return NoError;
}


I16 CTR_8554_Debounce_Config(
    U16 wCardNumber,
    U16 DebounceClock
    )
{
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(CurrentCard.wCardType!=PCI_8554)
        return ErrorFuncNotSupport;
    if(DebounceClock>1)
        return ErrorInvalidCtrSource;

    iotConf = CurrentCard.CTR_Status.CurrentConfig;
    iotConf.mData.m8554CTR.mCMCdata.mCMCbit.bDBCSEL = DebounceClock;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_8554CTR_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    CurrentCard.CTR_Status.CurrentConfig = iotConf;

    return NoError;
}


//yuan add 10/06/00
I16 GCTR_Setup(
    U16 wCardNumber,
    U16 wGCtr,
    U16 wGCtrCtrl,
    U32 dwCount
    )
{
    DAS_IOT_GPTC iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC));

    //yuan add 01/31/01 for error checking
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotGCtr.wCounterIndex = wGCtr;
    iotGCtr.Ctrldata.wValue = wGCtrCtrl;
    iotGCtr.dwCounterValue = dwCount;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_CONFIG, &(iotGCtr))
        !=DAS_ERR_NO)
    {
        return iotGCtr.wErrorCode;
    }
    return NoError;
}


I16 GCTR_Clear(
    U16 wCardNumber,
    U16 wGCtr
    )
{
    DAS_IOT_GPTC iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC));

    //yuan add 01/31/01 for eror checking
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    //yuan 07/31/00
    iotGCtr.wCounterIndex = wGCtr;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_RESET, &(iotGCtr))
        !=DAS_ERR_NO)
    {
        return iotGCtr.wErrorCode;
    }

    return NoError;
}


I16 GCTR_Read(
    U16 wCardNumber,
    U16 wGCtr,
    U32 *pValue
    )
{
    DAS_IOT_GPTC iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC));

    //yuan add 01/31/01 for eror checking
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotGCtr.wCounterIndex = wGCtr;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_READ, &(iotGCtr))
        !=DAS_ERR_NO)
    {
        return iotGCtr.wErrorCode;
    }
    *pValue = iotGCtr.dwCounterValue;

    return NoError;
}


//yuan add 06/22/05 for WDT
//action 0: set count ; 1 (op): disarm/arm/reload
//value for op : 0: disarm, 1: arm, 2: reload
I16 WDT_Control(
    U16 wCardNumber,
    U16 wCtr,
    U16 action
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    if((CurrentCard.wCardType!=PCI_7260)&&(CurrentCard.wCardType!=PCI_7452)&&
        (CurrentCard.wCardType!=PCI_7442)&&(CurrentCard.wCardType!=PCI_7444))
    {
        return ErrorFuncNotSupport;
    }

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotCtr.wCounterIndex = wCtr;
    iotCtr.wMode = action;
    iotCtr.wState = 1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_CONFIG, &(iotCtr))
        !=DAS_ERR_NO)
    {
        return iotCtr.wErrorCode;
    }
    if(!action){
        CurrentCard.i_dbsig_act.sa_handler = SIG_IGN;
        sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL);
    }

    return NoError;
}


I16 WDT_Setup(
    U16 wCardNumber,
    U16 wCtr,
    F32 ovflowSec,
    F32 *actualSec,
    void (*event_handler)(int)
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    if((CurrentCard.wCardType!=PCI_7260)&&(CurrentCard.wCardType!=PCI_7452)&&
        (CurrentCard.wCardType!=PCI_7442)&&(CurrentCard.wCardType!=PCI_7444))
    {
        return ErrorFuncNotSupport;
    }
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if((CurrentCard.wCardType==PCI_7442)||(CurrentCard.wCardType==PCI_7444)){
        iotCtr.wBinBcd = (wCtr>>15)&0x1;
        wCtr &= 0x7fff;
    }

    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    switch(CurrentCard.wCardType){
        case PCI_7260:
            if((ovflowSec<0.0002)||(ovflowSec>31.999))
                return ErrorInvalidCounterValue;
            *actualSec = 0;
            iotCtr.wCounterIndex = wCtr;
            if(event_handler)
                iotCtr.wMode = 1;
            iotCtr.dwCounterValue = (U32)(ovflowSec*2048);
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_CONFIG, &(iotCtr))
                !=DAS_ERR_NO)
            {
                return iotCtr.wErrorCode;
            }
            *actualSec = ((F32)iotCtr.dwCounterValue)/((F32)2048);
            break;
        case PCI_7452:
            if((ovflowSec<(F32)1.0)||(ovflowSec>(F32)3600.0))
                return ErrorInvalidCounterValue;
            *actualSec = 0;
            iotCtr.wCounterIndex = wCtr;
            if(event_handler)
                iotCtr.wMode = 1;
            iotCtr.dwCounterValue = (U32)(ovflowSec*520833);
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_CONFIG, &(iotCtr))
                !=DAS_ERR_NO)
            {
                return iotCtr.wErrorCode;
            }
            *actualSec = ((F32)iotCtr.dwCounterValue)/((F32)520833);
            break;
        case PCI_7442:
        case PCI_7444:
            if((ovflowSec<(F32)0.0000001)||(ovflowSec>(F32)429.0))
                return ErrorInvalidCounterValue;
            *actualSec = 0;
            iotCtr.wCounterIndex = (wCtr);
            if(event_handler)
                iotCtr.wMode = 1;
            iotCtr.dwCounterValue = (U32)(ovflowSec*10000000);
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_CONFIG, &(iotCtr))
                !=DAS_ERR_NO)
            {
                return iotCtr.wErrorCode;
            }
            *actualSec = ((F32)iotCtr.dwCounterValue)/((F32)10000000);
            break;
    }

    if(event_handler){
        if(CurrentCard.i_dbsig_act.sa_handler == SIG_IGN){
            CurrentCard.i_dbsig_act.sa_handler = event_handler;
            CurrentCard.i_dbsig_act.sa_flags |= SA_RESTART;
            if(sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL)==-1){
                return ErrorOpenEventFailed;
            }
        }
    }

    return NoError;
}


I16 WDT_Status(
    U16 wCardNumber,
    U16 Ctr,
    U32 *pValue
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    if((CurrentCard.wCardType!=PCI_7260)&&(CurrentCard.wCardType!=PCI_7452)&&
        (CurrentCard.wCardType!=PCI_7442)&&(CurrentCard.wCardType!=PCI_7444))
    {
        return ErrorFuncNotSupport;
    }
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(Ctr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotCtr.wCounterIndex = Ctr;
    iotCtr.wMode = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_READ, &iotCtr)!=DAS_ERR_NO)
        return iotCtr.wErrorCode;
    *pValue = iotCtr.dwCounterValue;

    return NoError;
}


I16 WDT_Reload(
    U16 wCardNumber,
    U16 wCtr,
    F32 ovflowSec,
    F32 *actualSec
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    if((CurrentCard.wCardType!=PCI_7442)&&(CurrentCard.wCardType!=PCI_7444))
        return ErrorFuncNotSupport;
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotCtr.wState = 2;
    iotCtr.wCounterIndex = wCtr;
    switch(CurrentCard.wCardType){
        case PCI_7442:
        case PCI_7444:
            if((ovflowSec<(F32)0.0000001)||(ovflowSec>(F32)429.0))
                return ErrorInvalidCounterValue;
            *actualSec = 0;
            iotCtr.dwCounterValue = (U32)(ovflowSec*10000000);
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_CONFIG, &(iotCtr))
                !=DAS_ERR_NO)
            {
                return iotCtr.wErrorCode;
            }
            *actualSec = ((F32)iotCtr.dwCounterValue)/((F32)10000000);
            break;
    }

    return NoError;
}


I16 GPTC_Control(
    U16 wCardNumber,
    U16 wGCtr,
    U16 ParamID,
    U16 Value
    )
{
    DAS_IOT_GPTC_9221 iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC_9221));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    switch(CurrentCard.wCardType){
        case PCI_9221:
        case PCI_6202:
        case PCI_9222:
        case PCI_9223:
            if(ParamID>4)
                return ErrorUndefinedParameter;
            iotGCtr.wCounterIndex = wGCtr;
            iotGCtr.dwCounterValue = Value;
            iotGCtr.dwCounterLength = ParamID;
            break;
        case PCI_9524:
            if(ParamID!=0)
                return ErrorUndefinedParameter;
            iotGCtr.wCounterIndex = wGCtr;
            iotGCtr.Ctrldata.wValue[1] = 1; //GPTC_Control
            iotGCtr.dwCounterValue = Value;
            break;
        default:
            return ErrorFuncNotSupport;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_CONFIG_9221, &(iotGCtr))
        !=DAS_ERR_NO)
    {
        return iotGCtr.wErrorCode;
    }

    return NoError;
}


I16 GPTC_Setup(
    U16 wCardNumber,
    U16 wGCtr,
    U16 wMode,
    U16 wSrcCtrl,
    U16 wPolCtrl,
    U32 LReg1_Val,
    U32 LReg2_Val
    )
{
    DAS_IOT_GPTC_9221 iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC_9221));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    switch(CurrentCard.wCardType){
        case PCI_9221:
        case PCI_6202:
        case PCI_9222:
        case PCI_9223:
            iotGCtr.wCounterIndex = wGCtr;
            iotGCtr.Ctrldata.Ctrlbit.fconfig = 1;
            switch(wMode){
                case 1://SimpleGatedEventCNT
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 0;
                    break;
                case 2://SinglePeriodMSR
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 2;
                    break;
                case 3://SinglePulseWidthMSR
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 1;
                    break;
                case 4://SingleGatedPulseGen
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 4;
                    break;
                case 5://SingleTrigPulseGen
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 4;
                    iotGCtr.Ctrldata.Ctrlbit.fEnTrigPulseOutput = 1;
                    break;
                case 6://RetrigSinglePulseGen
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 4;
                    iotGCtr.Ctrldata.Ctrlbit.fEnRetrigger = 1;
                    iotGCtr.Ctrldata.Ctrlbit.fEnTrigPulseOutput = 1;
                    break;
                case 7://SingleTrigContPulseGen
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 5;
                    iotGCtr.Ctrldata.Ctrlbit.fEnTrigPulseOutput = 1;
                    break;
                case 8://ContGatedPulseGen
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 5;
                    break;
                case 9://EdgeSeparationMSR
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 3;
                    break;
                case 10://SingleTrigContPulseGenPWM
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 5;
                    iotGCtr.Ctrldata.Ctrlbit.fEnTrigPulseOutput = 1;
                    iotGCtr.Ctrldata.Ctrlbit.fPWM = 1;
                    break;
                case 11://ContGatedPulseGenPWM
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 5;
                    iotGCtr.Ctrldata.Ctrlbit.fPWM = 1;
                    break;
                case 12://CW_CCW_Encoder
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 8;
                    break;
                case 13://x1_AB_Phase_Encoder
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 9;
                    break;
                case 14://x2_AB_Phase_Encoder
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 10;
                    break;
                case 15://x4_AB_Phase_Encoder
                    iotGCtr.Ctrldata.Ctrlbit.bMode = 11;
                    break;
                case 16://Phase_Z
                    iotGCtr.Ctrldata.Ctrlbit.fEnEZReset = 1;
                    iotGCtr.Ctrldata.Ctrlbit.fEZClearPhase = LReg1_Val&0x3;
                    iotGCtr.Ctrldata.Ctrlbit.fEZClearMode = LReg2_Val&0x3;
                    break;
                default:
                    return ErrorUndefinedParameter;
            }

            if(wMode!=16){
                iotGCtr.Ctrldata.Ctrlbit.fExternalClkSrc = wSrcCtrl&0x1;
                iotGCtr.Ctrldata.Ctrlbit.fExternalGateSrc = (wSrcCtrl>>1)&0x1;
                iotGCtr.Ctrldata.Ctrlbit.fExternalUpDownSrc = (wSrcCtrl>>2)&0x1;

                iotGCtr.Ctrldata.Ctrlbit.fClkenLActive = wPolCtrl&0x1;
                iotGCtr.Ctrldata.Ctrlbit.fGateLActive = (wPolCtrl>>1)&0x1;
                iotGCtr.Ctrldata.Ctrlbit.fUpDownLactive = (wPolCtrl>>2)&0x1;
                iotGCtr.Ctrldata.Ctrlbit.fOutputLactive = (wPolCtrl>>3)&0x1;

                iotGCtr.dwCounterValue = LReg1_Val;
                iotGCtr.dwCounterLength = LReg2_Val;
            }
            break;
        case PCI_9524:
            if(wGCtr<3){
                if(wMode>2)
                    return ErrorUndefinedParameter;
                //08/25/09
                if((LReg1_Val==0)||(LReg2_Val==0)||
                    ((LReg1_Val+LReg2_Val)>/*0xfffffe*/0xfffffffe))
                {
                    return ErrorUndefinedParameter;
                }
            }
            else if(wGCtr<6){
                if(wMode!=3)
                    return ErrorUndefinedParameter;
            }
            else if(wGCtr<7){
                //11/24/08 add for 9524 new feature (Internal Counter for Timer)
                if(wMode!=4)
                    return ErrorUndefinedParameter;
                /*
                 * 11/27/08 fix FPGA bug: counter adds extra one clock
                 */
                if((LReg1_Val==0)||(LReg1_Val==1))
                    return ErrorUndefinedParameter;
                else
                    LReg1_Val--;
            }
            else{
                return ErrorInvalidCounter;
            }
            iotGCtr.wCounterIndex = wGCtr;
            iotGCtr.Ctrldata.wValue[0] = wMode;
            iotGCtr.Ctrldata.wValue[1] = 0; //GPTC_Setup
            iotGCtr.dwCounterValue = LReg1_Val;
            iotGCtr.dwCounterLength = LReg2_Val;
            break;
        default:
            return ErrorFuncNotSupport;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_CONFIG_9221, &(iotGCtr))
        !=DAS_ERR_NO)
    {
        return iotGCtr.wErrorCode;
    }

    return NoError;
}


I16 GPTC_Clear(
    U16 wCardNumber,
    U16 wGCtr
    )
{
    DAS_IOT_GPTC_9221 iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC_9221));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    switch(CurrentCard.wCardType){
        case PCI_9221:
        case PCI_9524:
        case PCI_6202:
        case PCI_9222:
        case PCI_9223:
            iotGCtr.wCounterIndex = wGCtr;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_RESET_9221,
                &(iotGCtr))!=DAS_ERR_NO)
            {
                return iotGCtr.wErrorCode;
            }
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 GPTC_Read(
    U16 wCardNumber,
    U16 wGCtr,
    U32 *pValue
    )
{
    DAS_IOT_GPTC_9221 iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC_9221));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    //03/19/09 add for reading EZ clear latched data
    if(CurrentCard.wCardType==PCI_6202){
        if(wGCtr>=CurrentCardInfo.wCtrChannelCount+3)
           return ErrorInvalidCounter;
    }
    else{
        if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
           return ErrorInvalidCounter;
    }

    switch(CurrentCard.wCardType){
        case PCI_9221:
        case PCI_6202:
        case PCI_9524:
        case PCI_9222:
        case PCI_9223:
            iotGCtr.wCounterIndex = wGCtr;
            iotGCtr.dwCounterLength = 0; //Read counter value
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_READ_9221,
                &(iotGCtr))!=DAS_ERR_NO)
            {
                return iotGCtr.wErrorCode;
            }
            *pValue = iotGCtr.dwCounterValue;
            break;
        default:
            return ErrorFuncNotSupport;
    }
    return NoError;
}


I16 GPTC_Status(
    U16 wCardNumber,
    U16 wGCtr,
    U16 *pValue
    )
{
    DAS_IOT_GPTC_9221 iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC_9221));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    iotGCtr.wCounterIndex = wGCtr;
    switch(CurrentCard.wCardType){
        case PCI_9221:
        case PCI_6202:
            iotGCtr.dwCounterLength = 1;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_READ_9221, &(iotGCtr))
                !=DAS_ERR_NO)
            {
                return iotGCtr.wErrorCode;
            }
            /*bit24: Counting in progress; bit25: Counter Done*/
            *pValue = ((iotGCtr.dwCounterValue)>>24)&0x3;
            break;
        case PCI_9222:
        case PCI_9223:
            iotGCtr.dwCounterLength = 1;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_READ_9221, &(iotGCtr))
                !=DAS_ERR_NO)
            {
                return iotGCtr.wErrorCode;
            }
            /*
             * GPTC 0~3: (wGCtr<=3)
             * bit0: Counting in progress; bit1: Counter Done;
             * Encoder 0~1: (wGCtr<=5)
             * bit0: EA; bit1: EB; bit2: EZ; bit3: EORG;
             */
            *pValue = (iotGCtr.dwCounterValue)&0xf;
            break;
        case PCI_9524:
            iotGCtr.dwCounterLength = 1;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_READ_9221, &(iotGCtr))
                !=DAS_ERR_NO)
            {
                return iotGCtr.wErrorCode;
            }
            *pValue = (iotGCtr.dwCounterValue)&0xff;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 GPTC_9524_PG_Config(
    U16 wCardNumber,
    U16 wGCtr,
    U32 dwPulseGenNum
    )
{
    DAS_IOT_GPTC_9221 iotGCtr;

    memset(&iotGCtr, '\0', sizeof(DAS_IOT_GPTC_9221));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    switch(CurrentCard.wCardType){
        case PCI_9524:
            //08/25/09
            //if(dwPulseGenNum>0xffffff)
            //    return ErrorUndefinedParameter;
            iotGCtr.wCounterIndex = wGCtr;
            iotGCtr.Ctrldata.wValue[1] = 2; //GPTC_9524_PG_Config
            iotGCtr.dwCounterValue = dwPulseGenNum;
            break;
        default:
            return ErrorFuncNotSupport;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_CONFIG_9221, &(iotGCtr))
        !=DAS_ERR_NO)
    {
        return iotGCtr.wErrorCode;
    }

    return NoError;
}


I16 PWM_Output(
    U16 wCardNumber,
    U16 Channel,
    U32 high_interval,
    U32 low_interval
    )
{
    U32 LcrAddr;
    U32 *p_pwm_high;
    U32 *p_pwm_low;
    U32 tmp_value;

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_6202:
            LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
            Channel = Channel&0x3;
            p_pwm_high = (U32 *)(LcrAddr+0x4b8);
            p_pwm_low = (U32 *)(LcrAddr+0x4bc);
            tmp_value = low_interval&0xffffff;
            *p_pwm_high = tmp_value|(0x1<<(Channel+24));
            tmp_value = (*p_pwm_low&0xf0000000)|(high_interval&0xffffff);
            *p_pwm_low = tmp_value|(0x1<<(Channel+24))|(0x1<<(Channel+28));
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 PWM_Stop(
    U16 wCardNumber,
    U16 Channel
    )
{
    U32 LcrAddr;
    U32 *p_pwm_low;
    U32 tmp_value;

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_6202:
            LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
            Channel = Channel&0x3;
            p_pwm_low = (U32 *)(LcrAddr+0x4bc);
            tmp_value = *p_pwm_low;
            *p_pwm_low = tmp_value&(!(0x1<<(Channel+28)));
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 GPTC_EventSetup(
    U16 wCardNumber,
    U16 wGCtr,
    U16 wMode,
    U16 wCtrl,
    U32 LVal_1,
    U32 LVal_2
    )
{
    U32 LcrAddr;
    U32 *p_reg_addr;
    U32 tmp_value;
    DAS_IOT_GPTC_9221 iotGCtr = {0}; //03/13/08

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);
    if(wGCtr>=CurrentCardInfo.wCtrChannelCount)
        return ErrorInvalidCounter;

    switch(CurrentCard.wCardType){
        case PCI_6202:
            LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
            switch(wGCtr){
                case P6202_ENCODER0:
                    p_reg_addr = (U32 *)(LcrAddr+0x4e4);
                    break;
                case P6202_ENCODER1:
                    p_reg_addr = (U32 *)(LcrAddr+0x4e8);
                    break;
                case P6202_ENCODER2:
                    p_reg_addr = (U32 *)(LcrAddr+0x4ec);
                    break;
                default:
                    return ErrorInvalidCounter;
            }
            tmp_value = (LVal_1&0x0007ffff)|
                ((wCtrl&0x0000000f)<<24)|0x80000000;
            *p_reg_addr = tmp_value;
            return NoError;
        case PCI_9222:
        case PCI_9223:
            LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
            if(wGCtr<=3){
                /*GPTC 0~3 has no Event Setup*/
                return ErrorInvalidCounter;
            }
            else{
                iotGCtr.wCounterIndex = wGCtr;
                //The Highest Bit (bit 9) of fReserved2 Lets Kernel Know that
                //the Event Setup to be performed
                iotGCtr.Ctrldata.Ctrlbit.fReserved2 = ((1<<9)|(wCtrl&0xf));
                iotGCtr.dwCounterValue = LVal_1;
            }
            break;
        //do not support at this moment
        /*case PCI_9524:
            if(((wGCtr<3)&&(wMode!=0))||((wGCtr>=3)&&(wMode!=1)))
                return ErrorUndefinedParameter;
            if((LVal_1==0)||(LVal_1>0xffffff))
                return ErrorUndefinedParameter;
            iotGCtr.wCounterIndex = wGCtr;
            iotGCtr.Ctrldata.wValue[0] = wMode;
            iotGCtr.Ctrldata.wValue[1] = 3;
            iotGCtr.dwCounterValue = LVal_1;
            break;*/
        default:
            return ErrorFuncNotSupport;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_GPTC_CONFIG_9221, &(iotGCtr))
        !=DAS_ERR_NO)
    {
        return iotGCtr.wErrorCode;
    }

    return NoError;
}
