#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"
#include "DAQHeader.h"


I16 write_header_to_file(U16 wCardNumber, FILE *fp, U16 data_width,
    U16 ad_range,  U16 num_chan, U8 channel_no, U32 count,
    F64 fSampleRate, U16 num_of_channel_range);


extern int write_data_to_file(U16 *buf, FILE *fp, U16 byte_cnt,
    BOOLEAN exchange, U32 count);


extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 DI_7200_Config(
    U16 wCardNumber,
    U16 wTrigSource,
    U16 wExtTrigEn,
    U16 wTrigPol,
    U16 wI_REQ_Pol
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m7200DIO.wTrigSource = wTrigSource;
    iotConf.mData.wValue[1] = wExtTrigEn|wTrigPol|wI_REQ_Pol;
    //iotConf.mData.m7200DIO.fInTriggerRising=wTrigPol;
    //iotConf.mData.m7200DIO.fInWaitTrigger=wExtTrigEn;
    //iotConf.mData.m7200DIO.fInRequestRising=wI_REQ_Pol;
    iotConf.mData.m7200DIO.fTimerCascade = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    CurrentCard.DI_Status.CurrentConfig = iotConf;

    return NoError;
}


I16 DI_7300A_Config(
    U16     wCardNumber,
    U16     wPortWidth,
    U16     wTrigSource,
    U16     wWaitStatus,
    U16     wTerminator,
    U16     wI_REQ_Pol,
    BOOLEAN clear_fifo,
    BOOLEAN disable_di
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(wPortWidth){
        case 0:
            wPortWidth = DAS_PORT_WIDTH_0;
            break;
        case 8:
            wPortWidth = DAS_PORT_WIDTH_8;
            break;
        case 16:
            wPortWidth = DAS_PORT_WIDTH_16;
            break;
        case 32:
            wPortWidth = DAS_PORT_WIDTH_32;
            break;
        default:
            return ErrorInvalidPortWidth;
    }
    iotConf.mData.m7300DIO.fPortWidth = wPortWidth;
    iotConf.mData.m7300DIO.wTrigSource = wTrigSource;
    iotConf.mData.m7300DIO.fStartWait = wWaitStatus;
    iotConf.mData.m7300DIO.fTerminator = wTerminator;
    iotConf.mData.m7300DIO.fRequestRising = wI_REQ_Pol;
    iotConf.mData.m7300DIO.clear_fifo = (clear_fifo? 1:0);
    iotConf.mData.m7300DIO.disable_di = (disable_di? 1:0);
    //yuan add 11/30/99
    iotConf.mData.m7300DIO.fConfig = 1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    CurrentCard.DI_Status.CurrentConfig = iotConf;

    return NoError;
}


I16 DI_7300B_Config(
    U16     wCardNumber,
    U16     wPortWidth,
    U16     wTrigSource,
    U16     wWaitStatus,
    U16     wTerminator,
    U16     wI_Cntrl_Pol,
    BOOLEAN clear_fifo,
    BOOLEAN disable_di
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(wPortWidth){
        case 0:
            wPortWidth = DAS_PORT_WIDTH_0;
            break;
        case 8:
            wPortWidth = DAS_PORT_WIDTH_8;
            break;
        case 16:
            wPortWidth = DAS_PORT_WIDTH_16;
            break;
        case 32:
            wPortWidth = DAS_PORT_WIDTH_32;
            break;
        default:
            return ErrorInvalidPortWidth;
    }
    iotConf.mData.m7300bDIO.fPortWidth = wPortWidth;
    iotConf.mData.m7300bDIO.wTrigSource = wTrigSource;
    iotConf.mData.m7300bDIO.fStartWait = wWaitStatus;
    iotConf.mData.m7300bDIO.fTerminator = wTerminator;
    iotConf.mData.m7300bDIO.fRequestRising = wI_Cntrl_Pol;
    iotConf.mData.m7300bDIO.clear_fifo = (clear_fifo? 1:0);
    iotConf.mData.m7300bDIO.disable_di = (disable_di? 1:0);
    //yuan add 11/30/99
    iotConf.mData.m7300bDIO.fConfig = 1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    CurrentCard.DI_Status.CurrentConfig = iotConf;

    return NoError;
}


I16 DI_AsyncCheck(
    U16     wCardNumber,
    BOOLEAN *bStopped,
    U32     *dwAccessCnt
    )
{
    I16 err;
    DAS_IOT_ASYNC iotAsync;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber,0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_ASY_CHECK, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    *bStopped = (BOOLEAN)iotAsync.wStopped;
    *dwAccessCnt = iotAsync.dwAccessCount;

    return NoError;
}


I16 DI_AsyncClear(
    U16 wCardNumber,
    U32 *dwAccessCnt
    )
{
    I16 err, i;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    U32 byte_width, tmpCount, BufId, tmp;
    U32 cur_pos, ch_scan, data_res;
    BOOLEAN HalfReady, fstop;
    U16 RdyTrigCnt;
    FILE *fp;
    DAS_IOT_ASYNC iotAsync;
    DAS_IOT_CONFIG iotConf;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_ASY_CLEAR, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    *dwAccessCnt = iotAsync.dwAccessCount;

    //Data Process
    if((CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223))
    {
        iotConf = CurrentCard.DI_Status.CurrentConfig;
        BufId = (U32)(CurrentCard.DI_Status.pDAQbufHandle);
        *dwAccessCnt = iotAsync.dwAccessCount;

        if(!CurrentCard.DI_Status.bDblBufferMode){
            if(!CurrentCard.DI_Status.bToFile){
                //return NoError;
            }
            else{
                //07/31/08
                /*Infinite Trigger*/
                if(((CurrentCard.wCardType==PCI_9222)||
                    (CurrentCard.wCardType==PCI_9223))&&
                    (iotConf.mData.m922xAI.fEnReTrig)&&
                    (iotConf.mData.m922xAI.ReTriggerCnt==0)&&
                    (CurrentCard.DI_Status.pDAQbuf[1]))
                {
                    //Get the index of working buffer of Double Buffer
                    //DI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady, &fstop);
                    DI_AsyncReTrigNextReady(wCardNumber, &HalfReady, &fstop,
                        &RdyTrigCnt
                    );
                    if(HalfReady){
                        DI_AsyncDblBufferToFile(wCardNumber);  //transfer to file
                        CurrentCard.DI_Status.wDblBufferIndex = \
                            (CurrentCard.DI_Status.wDblBufferIndex==0)? 1:0;
                    }
                    if((*dwAccessCnt!=CurrentCard.DI_Status.dwTransferCount)&&
                        (*dwAccessCnt!=0))
                    {
                        write_data_to_file(CurrentCard.DI_Status.pDAQbuf[\
                            CurrentCard.DI_Status.wDblBufferIndex],
                            CurrentCard.DI_Status.pwBuffer,
                            (U16)((CurrentCard_DIFunc.wResoluation+7)/8),
                            FALSE,
                            *dwAccessCnt
                        );
                    }
                }
                else{
                //Write Last Data from Memory to Disk File
                    write_data_to_file(
                        (U16 *)CurrentCard.DI_Status.pDAQbuf[0],
                        CurrentCard.DI_Status.pwBuffer,
                        (U16)((CurrentCard_DIFunc.wResoluation+7)/8),
                        FALSE,
                        *dwAccessCnt
                    );
                }
                fp = (FILE *)CurrentCard.DI_Status.pwBuffer;
                data_res = (U16)((CurrentCard_DIFunc.wResoluation+7)/8);
                cur_pos = ftell(fp);
                ch_scan = (cur_pos-60)/data_res;
                ch_scan = ch_scan/CurrentCard.DI_Status.wNumChans;
                fseek(fp, 15, SEEK_SET);
                fwrite(&ch_scan, sizeof(long), 1, fp);
                fseek(fp, 0, SEEK_END);
                fclose((FILE *)CurrentCard.DI_Status.pwBuffer);
            }

            //Reset Buffer Structure if AutoBufferReset is Enabled
            if((CurrentCard.wCardType==PCI_9222)||
                (CurrentCard.wCardType==PCI_9223))
            {
                if(iotConf.mData.m922xAI.fAutoResetBuf){
                    CurrentCard.DI_Status.pDAQbuf[BufId] = NULL;
                    CurrentCard.DI_Status.dwBufCount[BufId] = 0;
                }
            }
        }
        else{ //DblBufferMode
            if(!CurrentCard.DI_Status.bToFile){
                //Get the index of working buffer of Double Buffer
                DI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady);
                if(HalfReady){
                    *dwAccessCnt += CurrentCard.DI_Status.dwTransferCount;
                    //Removed, it is not necessary to mark next ready
                    //CurrentCard.DI_Status.wDblBufferIndex |= 0x02;
                }
            }
            else{
                //Get the index of working buffer of Double Buffer
                DI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady);
                //write data to file
                if(HalfReady){
                    //transfer to file
                    DI_AsyncDblBufferToFile(wCardNumber);
                    CurrentCard.DI_Status.wDblBufferIndex = \
                        (!CurrentCard.DI_Status.wDblBufferIndex)? 1:0;
                }
                if((*dwAccessCnt!=CurrentCard.DI_Status.dwTransferCount)&&
                    (*dwAccessCnt!=0))
                {
                    BufId = CurrentCard.DI_Status.wDblBufferIndex;
                    write_data_to_file(
                        CurrentCard.DI_Status.pDAQbuf[BufId],
                        CurrentCard.DI_Status.pwBuffer,
                        (U16)((CurrentCard_DIFunc.wResoluation+7)/8),
                        FALSE,
                        (*dwAccessCnt)
                    );
                }
                fp = (FILE *)CurrentCard.DI_Status.pwBuffer;
                data_res = (U16)((CurrentCard_DIFunc.wResoluation+7)/8);
                cur_pos = ftell(fp);
                ch_scan = (cur_pos - 60)/data_res;
                ch_scan = ch_scan/CurrentCard.DI_Status.wNumChans;
                fseek(fp, 15, SEEK_SET);
                fwrite(&ch_scan, sizeof(long), 1, fp);
                fseek(fp, 0, SEEK_END);
                fclose((FILE *)CurrentCard.DI_Status.pwBuffer);
            }

            if((CurrentCard.wCardType==PCI_9222)||
                (CurrentCard.wCardType==PCI_9223))
            {
                if(iotConf.mData.m922xAI.fAutoResetBuf){
                    CurrentCard.DI_Status.pDAQbuf[0] = NULL;
                    CurrentCard.DI_Status.pDAQbuf[1] = NULL;
                    CurrentCard.DI_Status.dwBufCount[0] = 0;
                    CurrentCard.DI_Status.dwBufCount[1] = 0;
                }
            }
        }
    }
    else{
        //Copy the DI data from internal buffer to user buffer
        if(!CurrentCard.DI_Status.bDblBufferMode){
            if((CurrentCard.wCardType!=PCI_7300A_RevA)&&
                (CurrentCard.wCardType!=PCI_7300A_RevB)&&
                (CurrentCard.wCardType!=PCI_7350))
            {
                if(CurrentCard.DI_Status.bToFile){
                    fwrite(
                        CurrentCard_DIFunc.pView[0],
                        (U32)((CurrentCard_DIFunc.wResoluation+7)/8),
                        CurrentCard.DI_Status.dwTransferCount,
                        (FILE *)CurrentCard.DI_Status.pwBuffer
                    );
                }
                else{
                    if(CurrentCard.DI_Status.pwBuffer){
                        memcpy(
                            CurrentCard.DI_Status.pwBuffer,
                            CurrentCard_DIFunc.pView[0],
                            CurrentCard.DI_Status.dwTransferCount*
                            (U32)((CurrentCard_DIFunc.wResoluation+7)/8)
                        );
                    }
                }
            }
            //06/07/09
            else if(CurrentCard.wCardType==PCI_7350){
                if(CurrentCard.DI_Status.bToFile){
                    byte_width = \
                        CurrentCard_DICurConfig.mData.m7350DIO.PortCount+1;
                    fwrite(
                        CurrentCard.DI_Status.p7300buf,
                        byte_width,
                        *dwAccessCnt,
                        (FILE *)CurrentCard.DI_Status.pwBuffer
                    );
                    munlock(
                        (const void *)CurrentCard.DI_Status.p7300buf,
                        CurrentCard.DI_Status.dwTransferCount*byte_width
                    );
                    free(CurrentCard.DI_Status.p7300buf);
                    CurrentCard.DI_Status.p7300buf = NULL;
                    /*re-write new scan count in file header*/
                    fp = (FILE *)CurrentCard.DI_Status.pwBuffer;
                    data_res = (U16)byte_width;
                    cur_pos = ftell(fp);
                    ch_scan = (cur_pos-60)/data_res;
                    fseek(fp, 15, SEEK_SET);
                    fwrite(&ch_scan, sizeof(long), 1, fp);
                    fseek(fp, 0, SEEK_END);

                    //yuan add 04/11/01 for multibuffer unlock ??
                    for(i=0; i<CurrentCard.DI_Status.p7300mbufCnt; i++){
                        if(CurrentCard.DI_Status.p7350mbuf[i].dwBufferAddr){
                            munlock(
                                (const void *)
                                CurrentCard.DI_Status.p7350mbuf[i].dwBufferAddr,
                                CurrentCard.DI_Status.p7350mbuf[i].dwAccessCount
                                *byte_width
                            );
                        }
                        CurrentCard.DI_Status.p7350mbuf[i].dwBufferAddr = \
                            (U32)NULL;
                        CurrentCard.DI_Status.p7350mbuf[i].dwAccessCount = 0;
                        CurrentCard.DI_Status.p7350mbuf[i].wBufferId = -1;
                    }
                    CurrentCard.DI_Status.p7300mbufCnt = 0;
                }
            }
            else{ //PCI-7300A
                if(CurrentCard.wCardType==PCI_7300A_RevA){
                    tmp = CurrentCard_DICurConfig.mData.m7300DIO.fPortWidth;
                    byte_width = pwidth[tmp]/8;
                }
                else{
                    tmp = CurrentCard_DICurConfig.mData.m7300bDIO.fPortWidth;
                    byte_width = pwidth[tmp]/8;
                }
                //00/06/02 jeffrey , for memory lock in user level
                munlock(
                    (const void *)CurrentCard.DI_Status.p7300buf,
                    CurrentCard.DI_Status.dwTransferCount*byte_width
                );
                if(CurrentCard.DI_Status.bToFile){
                    fwrite(
                        CurrentCard.DI_Status.p7300buf,
                        byte_width,
                        CurrentCard.DI_Status.dwTransferCount,
                        (FILE *)CurrentCard.DI_Status.pwBuffer
                    );
                    free(CurrentCard.DI_Status.p7300buf);
                }
                CurrentCard.DI_Status.p7300buf = NULL;

                //yuan add 04/11/01 for multibuffer unlock
                for(i=0; i<CurrentCard.DI_Status.p7300mbufCnt; i++){
                    if(CurrentCard.DI_Status.p7300mbuf[i].dwBufferAddr){
                        munlock(
                            (const void *)
                            CurrentCard.DI_Status.p7300mbuf[i].dwBufferAddr,
                            CurrentCard.DI_Status.p7300mbuf[i].dwAccessCount*\
                            byte_width
                        );
                    }
                    CurrentCard.DI_Status.p7300mbuf[i].dwBufferAddr = (U32) NULL;
                    CurrentCard.DI_Status.p7300mbuf[i].dwAccessCount = 0;
                    CurrentCard.DI_Status.p7300mbuf[i].wBufferId = -1;
                }
                CurrentCard.DI_Status.p7300mbufCnt = 0;
            }
            if(CurrentCard.DI_Status.bToFile)
                fclose((FILE *)CurrentCard.DI_Status.pwBuffer);
        }
        else{
            //yuan add 07/25/01
            if((CurrentCard.wCardType!=PCI_7300A_RevA)&&
                (CurrentCard.wCardType!=PCI_7300A_RevB))
            {
                data_res = (U16)((CurrentCard_DIFunc.wResoluation+7)/8);
            }
            else{
                tmp = CurrentCard_DICurConfig.mData.m7300DIO.fPortWidth;
                data_res = pwidth[tmp]/8;
            }
            tmpCount = CurrentCard_DIFunc.dwBufferSize/(U32)(data_res);
            tmpCount = tmpCount/2;

            if((iotAsync.dwAccessCount&0xffffffff)<tmpCount)
                *dwAccessCnt = iotAsync.dwAccessCount&0xffffffff;
            else
                *dwAccessCnt = (iotAsync.dwAccessCount&0xffffffff)-tmpCount;

            if(CurrentCard.DI_Status.bToFile){
                //Get the index of working buffer of Double Buffer
                DI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady);
                if(HalfReady){
                    DI_AsyncDblBufferTransfer(wCardNumber, NULL);//transfer to file
                    CurrentCard.DI_Status.wDblBufferIndex = \
                        (CurrentCard.DI_Status.wDblBufferIndex==0)? 1:0;
                }
                if(((*dwAccessCnt)!=(CurrentCard.DI_Status.dwTransferCount))&&
                    ((*dwAccessCnt)!=0))
                {
                    if((CurrentCard.wCardType)!=(PCI_7300A_RevA)&&
                        (CurrentCard.wCardType)!=(PCI_7300A_RevB))
                    {
                        BufId = CurrentCard.DI_Status.wDblBufferIndex;
                        fwrite(
                            CurrentCard_DIFunc.pView[BufId],
                            data_res,
                            *dwAccessCnt,
                            (FILE *)CurrentCard.DI_Status.pwBuffer
                        );
                    }
                }
                if(HalfReady)
                    *dwAccessCnt += CurrentCard.DI_Status.dwTransferCount;
                fp = (FILE *)CurrentCard.DI_Status.pwBuffer;
                cur_pos = ftell(fp);
                ch_scan = (cur_pos-60)/data_res;
                ch_scan = ch_scan/CurrentCard.DI_Status.wNumChans;
                fseek(fp, 15, SEEK_SET);
                fwrite(&ch_scan, sizeof(long), 1, fp);
                fseek(fp, 0, SEEK_END);
                fclose((FILE *)CurrentCard.DI_Status.pwBuffer);
            }
            else{ //yuan 07/25/01
                DI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady);
                if(HalfReady){
                    *dwAccessCnt += CurrentCard.DI_Status.dwTransferCount;
                    CurrentCard.DI_Status.wDblBufferIndex |= 0x02;
                }
            }
        }
    }

    //move to the place before memory copy
    /*
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_ASY_CLEAR, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    */
    //*dwAccessCnt = iotAsync.dwAccessCount;
    return NoError;
}


I16 DI_AsyncDblBufferHalfReady(
    U16     wCardNumber,
    BOOLEAN *bHalfReady
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bHalfReady = (BOOLEAN)iotDbf.wReady;
    CurrentCard.DI_Status.wDblBufferIndex = iotDbf.wViewIndex;

    return NoError;
}


I16 DI_AsyncDblBufferHalfReadyIndex(
    U16     wCardNumber,
    BOOLEAN *bHalfReady,
    int     *ViewIndex
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bHalfReady = (BOOLEAN)iotDbf.wReady;
    *ViewIndex = iotDbf.wViewIndex;
    CurrentCard.DI_Status.wDblBufferIndex = iotDbf.wViewIndex;

    return NoError;
}


I16 DI_AsyncMultiBufferNextReady(
    U16     wCardNumber,
    BOOLEAN *bNextReady,
    U16     *wBufferId
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bNextReady = (BOOLEAN)iotDbf.wReady;
    *wBufferId = CurrentCard.DI_Status.wDblBufferIndex = iotDbf.wViewIndex;

    return NoError;
}


I16 DI_AsyncDblBufferMode(
    U16     wCardNumber,
    BOOLEAN bEnable
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotDbf.wEnable = bEnable;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_DBF_MODE, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    CurrentCard.DI_Status.bDblBufferMode = bEnable;

    return NoError;
}


I16 DI_AsyncDblBufferTransfer(
    U16  wCardNumber,
    void *pwBuffer
    )
{
    I16 err;
    U16 DblBufId;
    volatile PPCI_DBXORUN pOvRun;
    U32 *dwCtr_eisr;
    U16 BufIdx;

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    //yuan add 06/18/04
    if((CurrentCard.wCardType==PCI_7300A_RevB)&&
        (CurrentCard.DI_Status.bChanRange)&&(!CurrentCard.DI_Status.bToFile))
    {
        volatile PPCI7300_DBXORUN pOvRun;
        pOvRun = (volatile PPCI7300_DBXORUN)(CurrentCard_AOFunc.pView[0]);
        //07/08/08
        CurrentCard.DI_Status.wDblBufferIndex &= 0xff;
        BufIdx = CurrentCard.DI_Status.wDblBufferIndex;
        if(pOvRun){
            if(!pOvRun->xFlag.f.overrun){
                //while (pOvRun->xFlag.f.lock);
                pOvRun->xFlag.f.lock = 1;
                if(!pOvRun->xFlag.f.extended){
                    if(pOvRun->dwCtr_isr[BufIdx]>=1)
                        (pOvRun->dwCtr_isr[BufIdx])-=1;
                }
                else{
                    if(CurrentCard.DI_Status.wDblBufferIndex<8){
                        if(pOvRun->dwCtr_isr[BufIdx]>=1)
                            (pOvRun->dwCtr_isr[BufIdx])-=1;
                    }
                    else{
                        dwCtr_eisr = (U32 *)(pOvRun+1);
                        if(dwCtr_eisr[BufIdx-8]>=1)
                            (dwCtr_eisr[BufIdx-8])-=1;
                    }
                }
                pOvRun->xFlag.f.lock = 0;
            }
        }
        return NoError;
    }

    DblBufId = CurrentCard.DI_Status.wDblBufferIndex;
    CurrentCard.DI_Status.wDblBufferIndex &= 0x01;
    BufIdx = CurrentCard.DI_Status.wDblBufferIndex;

    if(!CurrentCard.DI_Status.bDblBufferMode)
        return ErrorNotDoubleBufferMode;

    if(CurrentCard.DI_Status.bToFile){ //to file
        fwrite(
            CurrentCard_DIFunc.pView[BufIdx],
            (U32)((CurrentCard_DIFunc.wResoluation+7)/8),
            CurrentCard.DI_Status.dwTransferCount,
            (FILE *)CurrentCard.DI_Status.pwBuffer
        );
    }
    else{
        memcpy(
            pwBuffer,
            CurrentCard_DIFunc.pView[BufIdx],
            CurrentCard.DI_Status.dwTransferCount*
            (U16)((CurrentCard_DIFunc.wResoluation+7)/8)
        );
        //yuan add 07/25/01
        if((DblBufId>>1)&0x01){
            CurrentCard.DI_Status.wDblBufferIndex = (BufIdx==0)? 1:0;
        }
    }

    //yuan add 09/11/02
    if(CurrentCard.wCardType==PCI_7200){
        pOvRun = (volatile PPCI_DBXORUN)(CurrentCard_AOFunc.pView[0]);
        BufIdx = CurrentCard.DI_Status.wDblBufferIndex;
        if(pOvRun){
            if(!pOvRun->xFlag.f.overrun){
                pOvRun->xFlag.f.lock = 1;
                if(pOvRun->dwCtr_isr[BufIdx]>=1)
                    pOvRun->dwCtr_isr[BufIdx] -= 1;
                pOvRun->xFlag.f.lock = 0;
            }
        }
    }

    return NoError;
}


//yuan add 08/21/06
I16 DI_AsyncMultiBuffersHandled(
    U16 wCardNumber,
    U16 bufcnt,
    U16 *bufid
    )
{
    I16 err;
    U16 DblBufId;
    //volatile PPCI7300_DBXORUN pOvRun;
    U32 *dwCtr_eisr;

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!= NoError)
    {
        return err;
    }

    //01/14/08 modified
    if(bufcnt>MAXMBUFCNT)
        return ErrorInvalidBufferID;

    //yuan add 06/18/04
    if((CurrentCard.wCardType==PCI_7300A_RevB)&&
        (CurrentCard.DI_Status.bChanRange))
    {
        volatile PPCI7300_DBXORUN pOvRun;
        pOvRun = (volatile PPCI7300_DBXORUN)(CurrentCard_AOFunc.pView[0]);
        //CurrentCard.DI_Status.wDblBufferIndex &= 0x0f;
        if(pOvRun){
            int vi = 0;
            if(!pOvRun->xFlag.f.overrun){
                //while (pOvRun->xFlag.f.lock);
                pOvRun->xFlag.f.lock = 1;
                for(vi=0; vi<bufcnt; vi++){
                    DblBufId = bufid[vi];
                    if(DblBufId>=MAXMBUFCNT)
                        return ErrorInvalidBufferID;
                    if(DblBufId>=8){
                        if(pOvRun->xFlag.f.extended==0)
                            return ErrorInvalidBufferID;
                        dwCtr_eisr = (U32 *)(pOvRun+1);
                        if(dwCtr_eisr[DblBufId-8]>=1)
                            dwCtr_eisr[DblBufId-8]-=1;
                    }
                    else{
                        if(pOvRun->dwCtr_isr[DblBufId]>=1)
                            pOvRun->dwCtr_isr[DblBufId]-=1;
                    }
                }
                pOvRun->xFlag.f.lock = 0;
            }
        }
    }
    else if(CurrentCard.wCardType==PCI_7350){
        if(CurrentCard.DI_Status.bChanRange){
            volatile PPCI7350_DBXORUN pOvRun;
            pOvRun = \
                &(((DAS_IOT_CALCONST_7350 *)CurrentCard.pCAL_Const)->OvRun[0]);
            if(pOvRun){
                int vi = 0;
                if(!pOvRun->xFlag.f.overrun){
                    pOvRun->xFlag.f.lock = 1;
                    for(vi=0; vi<bufcnt; vi++){
                        DblBufId = bufid[vi];
                        if(DblBufId>=MAXMBUFCNT)
                            return ErrorInvalidBufferID;
                        if(pOvRun->dwCtr_isr[DblBufId]>=1)
                            pOvRun->dwCtr_isr[DblBufId] -= 1;
                    }
                    pOvRun->xFlag.f.lock = 0;
                }
            }
        }
    }
    else
        return ErrorFuncNotSupport;

    return NoError;
}


I16 DI_ContMultiBufferSetup(
    U16  wCardNumber,
    void *pwBuffer,
    U32  dwReadCount,
    U16  *BufferId
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_MBUF iotMbuf;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    U32 byte_width;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    if((CurrentCard.wCardType==PCI_7300A_RevB)||
        (CurrentCard.wCardType==PCI_7300A_RevA)||
        (CurrentCard.wCardType==PCI_7350))
    {
        if(CurrentCard.wCardType==PCI_7350)
            byte_width = iotConf.mData.m7350DIO.PortCount+1;
        else if(CurrentCard.wCardType==PCI_7300A_RevA)
            byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
        else
            byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;
        //record the user-level buffer for unlock in DI_AsyncClear
        iotMbuf.dwBufferAddr = (U32)pwBuffer;
        iotMbuf.dwAccessCount = dwReadCount;
        //yuan add 11/05/01
        //if(CurrentCard.wCardType==PCI_7350)
        //    iotMbuf.wBufferId = 1;
        //else
            iotMbuf.wBufferId = 0;
        err = mlock((const void *)pwBuffer, dwReadCount*byte_width);
        if(err!=0){ //errno defined in errno.h
            //yuan add 06/19/06
            iotMbuf.wErrorCode = 1; //checking
            goto lockio;
            if(errno==ENOMEM)
                printf("memory lock failed : errno = ENOMEM , please check the memory size\n");
            else if(errno==EPERM)
                printf("memory lock failed : error =  EPERM, mlock system call is only for ROOT\n");
            else if(errno==EINVAL)
                printf("memory lock failed : error = EINVAL, Invalate value\n");
            else
                printf("memory lock failed : errno = %d\n", errno);
            return ErrorLockMemory;
        }
lockio:
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET, &(iotMbuf))
            !=DAS_ERR_NO)
        {
            munlock((const void *)pwBuffer, dwReadCount*byte_width);
            return iotMbuf.wErrorCode;
        }
        *BufferId = iotMbuf.wBufferId;
        if(CurrentCard.wCardType==PCI_7350){
            CurrentCard.DI_Status.p7350mbuf[*BufferId].dwBufferAddr = \
                (U32)pwBuffer;
            CurrentCard.DI_Status.p7350mbuf[*BufferId].dwAccessCount = \
                dwReadCount;
            CurrentCard.DI_Status.p7350mbuf[*BufferId].wBufferId = \
                *BufferId;
        }
        else{
            CurrentCard.DI_Status.p7300mbuf[*BufferId].dwBufferAddr = \
                (U32)pwBuffer;
            CurrentCard.DI_Status.p7300mbuf[*BufferId].dwAccessCount = \
                dwReadCount;
            CurrentCard.DI_Status.p7300mbuf[*BufferId].wBufferId = \
                *BufferId;
        }
        CurrentCard.DI_Status.p7300mbufCnt += 1;
        //yuan add 10/22/02
        CurrentCard.DI_Status.bChanRange = 1;
    }
    else
        return ErrorFuncNotSupport;

    return NoError;
}


I16 DI_ContMultiBufferReset(
    U16 wCardNumber
    )
{
    DAS_IOT_MBUF iotMbuf;
    DAS_IOT_CONFIG iotConf;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    U32 byte_width;
    int i;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    if((CurrentCard.wCardType==PCI_7300A_RevB)||
        (CurrentCard.wCardType==PCI_7300A_RevA)||
        (CurrentCard.wCardType==PCI_7350))
    {
        iotMbuf.dwAccessCount = 0;
        //yuan add 11/05/01
        iotMbuf.wBufferId = 0;

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET, &(iotMbuf))
            !=DAS_ERR_NO)
        {
            return iotMbuf.wErrorCode;
        }

        if(CurrentCard.wCardType==PCI_7350)
            byte_width = iotConf.mData.m7350DIO.PortCount+1;
        else if(CurrentCard.wCardType==PCI_7300A_RevA)
            byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
        else
            byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;

        //yuan add 04/11/01 for multibuffer unlock
        for(i=0; i<CurrentCard.DI_Status.p7300mbufCnt; i++){
            if(CurrentCard.wCardType==PCI_7350){
                if(CurrentCard.DI_Status.p7350mbuf[i].dwBufferAddr){
                    munlock(
                        (const void *)
                        CurrentCard.DI_Status.p7350mbuf[i].dwBufferAddr,
                        CurrentCard.DI_Status.p7350mbuf[i].dwAccessCount*
                        byte_width
                    );
                }
                CurrentCard.DI_Status.p7350mbuf[i].dwBufferAddr = (U32)NULL;
                CurrentCard.DI_Status.p7350mbuf[i].dwAccessCount = 0;
                CurrentCard.DI_Status.p7350mbuf[i].wBufferId = -1;
            }
            else{
                if(CurrentCard.DI_Status.p7300mbuf[i].dwBufferAddr){
                    munlock(
                        (const void *)
                        CurrentCard.DI_Status.p7350mbuf[i].dwBufferAddr,
                        CurrentCard.DI_Status.p7350mbuf[i].dwAccessCount*
                        byte_width
                    );
                }
                CurrentCard.DI_Status.p7300mbuf[i].dwBufferAddr = (U32)NULL;
                CurrentCard.DI_Status.p7300mbuf[i].dwAccessCount = 0;
                CurrentCard.DI_Status.p7300mbuf[i].wBufferId = -1;
            }
        }
        CurrentCard.DI_Status.p7300mbufCnt = 0;
    }
    else
        return ErrorFuncNotSupport;

    //yuan add 10/22/02
    CurrentCard.DI_Status.bChanRange = 0;

    return NoError;
}


/*
 * pwBuffer is of no use while double-buffer mode enabled
 */
I16 DI_ContReadPort(
    U16  wCardNumber,
    U16  wPort,
    void *pwBuffer,
    U32  dwReadCount,
    F64  fSampleRate,
    U16  wSyncMode
    )
{
    I16 err;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_SG_CONT iotSCont;
    //00/06/02 jeffrey added
    U32 byte_width = 0;
    U8 needlock = 0;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotSCont, '\0', sizeof(DAS_IOT_SG_CONT));

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)){
        U16 BufferID = *((U16 *)pwBuffer);
        U32 SampIntrv = 0;
        if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 1,
            &wPort, dwReadCount*(U32)((CurrentCard_DIFunc.wResoluation+7)/8),
            TRUE))!=NoError)
        {
            return err;
        }

        if(((CurrentCard.DI_Status.bDblBufferMode)||(wSyncMode==SYNCH_OP))&&
            (iotConf.mData.m922xAI.fEnReTrig))
        {
            return ErrorInvalidTriggerMode;
        }

        if(iotConf.mData.m922xAI.fConvSrc==0){
            if((fSampleRate==0)||(fSampleRate>2000000))
                return ErrorInvalidSampleRate;
            else
                SampIntrv = (U32)(80000000/fSampleRate);
        }

        if(!CurrentCard.DI_Status.bDblBufferMode){
            if((BufferID>1)||(!CurrentCard.DI_Status.pDAQbuf[BufferID]))
                return ErrorInvalidBufferID;
            CurrentCard.DI_Status.pDAQbufHandle = (U32)(BufferID);
        }
        else{
            if((BufferID>1)||(!CurrentCard.DI_Status.pDAQbuf[0])||
                (!CurrentCard.DI_Status.pDAQbuf[1]))
            {
                return ErrorInvalidBufferID;
            }
            CurrentCard.DI_Status.pDAQbufHandle = 0;
        }
        CurrentCard.DI_Status.bToFile = FALSE;
        CurrentCard.DI_Status.wNumChans = 1;
        CurrentCard.DI_Status.dwTransferCount = dwReadCount;

        iotSCont.wAutoScan = 0;
        iotSCont.wSyncMode = wSyncMode;
        iotSCont.wChannelCount = 1;
        iotSCont.dwAccessCount = dwReadCount;
        iotSCont.dwChannelGain[0].wChannel = wPort;
        iotSCont.dwBufferAddr = BufferID;
        iotSCont.dwPacerValue[0] = SampIntrv;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CON_READ_SG, &(iotSCont))
            !=DAS_ERR_NO)
        {
            return iotSCont.wErrorCode;
        }
    }
    else{
        if(CurrentCard.wCardType==PCI_7200)
            byte_width = (CurrentCard_DIFunc.wResoluation+7)/8;
        if(CurrentCard.wCardType==PCI_7300A_RevA)
            byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
        if(CurrentCard.wCardType==PCI_7300A_RevB)
            byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;
        if(CurrentCard.wCardType==PCI_7350)
            byte_width = iotConf.mData.m7350DIO.PortCount+1;

        if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 1,
            &wPort, dwReadCount*byte_width, TRUE))!=NoError)
        {
            return err;
        }

        if(CurrentCard.DI_Status.bDblBufferMode)
            dwReadCount /= 2;

        CurrentCard.DI_Status.bToFile = FALSE;
        CurrentCard.DI_Status.dwTransferCount = dwReadCount;
        CurrentCard.DI_Status.pwBuffer = pwBuffer;

        //06/10/09
        if((CurrentCard.wCardType!=PCI_7350)||
            ((CurrentCard.wCardType==PCI_7350)&&
            (iotConf.mData.m7350DIO.OPMode!=1)&&
            (!iotConf.mData.m7350DIO.ExtClk)))
        {
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
                return ErrorInvalidSampleRate;
        }
        else if((CurrentCard.wCardType==PCI_7350)&&
            (iotConf.mData.m7350DIO.OPMode==1))
        {
            iotCont.dwPacerValue = 2;
        }

        if(CurrentCard.wCardType==PCI_7200){
            //Set Timer Cascade
            if(iotCont.dwPacerValue>65535)
                iotConf.mData.m7200DIO.fTimerCascade = 1;
            else
                iotConf.mData.m7200DIO.fTimerCascade = 0;

            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }
        if((CurrentCard.wCardType==PCI_7300A_RevA)||
            (CurrentCard.wCardType==PCI_7300A_RevB)||
            (CurrentCard.wCardType==PCI_7350))
        {
            //if(CurrentCard.wCardType==PCI_7350)
            //    byte_width = iotConf.mData.m7350DIO.PortCount+1;
            //else if(CurrentCard.wCardType==PCI_7300A_RevA)
            //    byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
            //else
            //    byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;

            err = mlock((const void *)pwBuffer, dwReadCount*byte_width);
            if(err!=0){ //errno defined in errno.h
                //yuan add 06/19/06
                needlock = 1; //checking
                goto lockio;
                if(errno==ENOMEM)
                    printf("memory lock failed : errno = ENOMEM , please check the memory size\n");
                else if(errno==EPERM)
                    printf("memory lock failed : error =  EPERM, mlock system call is only for ROOT\n");
                else if(errno==EINVAL)
                    printf("memory lock failed : error = EINVAL, Invalate value\n");
                else
                    printf("memory lock failed : errno = %d\n", errno);
                return ErrorLockMemory;
            }
lockio:
            //record the user-level buffer for unlock in DI_AsyncClear
            CurrentCard.DI_Status.p7300buf = pwBuffer;
        }

        if(CurrentCard.wCardType==PCI_7300A_RevA){
            //Set Buffer Address
            iotConf.mData.m7300DIO.dwBufferAddr = (U32)pwBuffer;
            //yuan add 11/30/99
            iotConf.mData.m7300DIO.fConfig = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                munlock((const void *)pwBuffer, dwReadCount*byte_width);
                //yuan add 04/16/01
                CurrentCard.DI_Status.p7300buf = NULL;
                return iotConf.wErrorCode;
            }
        }
        if(CurrentCard.wCardType==PCI_7300A_RevB){
            //Set Buffer Address
            iotConf.mData.m7300bDIO.dwBufferAddr = (U32)pwBuffer;
            //yuan add 11/30/99
            iotConf.mData.m7300bDIO.fConfig = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                munlock((const void *)pwBuffer, dwReadCount*byte_width);
                //yuan add 04/16/01
                CurrentCard.DI_Status.p7300buf = NULL;
                return iotConf.wErrorCode;
            }
        }
        if(CurrentCard.wCardType==PCI_7350){
            if(CurrentCard.DI_Status.bChanRange==0){
                /*Non-Multi-Buffer*/
                DAS_IOT_MBUF iotMbuf = {0};
                iotMbuf.dwBufferAddr = (U32)pwBuffer;
                if((iotConf.mData.m7350DIO.WaitTrig==1)&&
                    (iotConf.mData.m7350DIO.WaitSTrig==0))
                {
                    /*External Trigger*/
                    if(!CurrentCard.AI_Status.CurrentConfig.mData.dwValue[1]){
                        /*Infinite Trigger*/
                        iotMbuf.dwAccessCount = dwReadCount*2;
                        CurrentCard.DI_Status.dwTransferCount *= 2;
                    }
                    else{
                        iotMbuf.dwAccessCount = dwReadCount*
                           CurrentCard.AI_Status.CurrentConfig.mData.dwValue[1];
                        CurrentCard.DI_Status.dwTransferCount *=
                           CurrentCard.AI_Status.CurrentConfig.mData.dwValue[1];
                    }
                }
                else
                    iotMbuf.dwAccessCount = dwReadCount;
                iotMbuf.wBufferId = 0|0x8000; //0: DI; 1:DO; //0x8000 single buffer
                //iotMbuf.wErrorCode = ErrorConfigIoctl;
                if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET, &(iotMbuf))
                    !=DAS_ERR_NO)
                {
                    munlock((const void *)pwBuffer, dwReadCount*byte_width);
                    //yuan add 04/16/01
                    CurrentCard.DI_Status.p7300buf = NULL;
                    return iotMbuf.wErrorCode;
                }
            }
        }

        iotCont.dwAccessCount = dwReadCount;
        iotCont.wAutoScan = 0;
        iotCont.wSyncMode = wSyncMode;
        iotCont.wChannelCount = 1;
        iotCont.dwChannelGain[0].wChannel = wPort;
        //yuan add 06/20/06
        iotCont.dwChannelGain[0].wGain = needlock;
        iotCont.wIterations = 1;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CON_READ, &(iotCont))
            !=DAS_ERR_NO)
        {
            if((CurrentCard.wCardType==PCI_7300A_RevA)||
                (CurrentCard.wCardType==PCI_7300A_RevB)||
                (CurrentCard.wCardType==PCI_7350))
            {
                munlock((const void *)pwBuffer, dwReadCount*byte_width);
                //yuan add 04/16/01
                CurrentCard.DI_Status.p7300buf = NULL;
            }
            return iotCont.wErrorCode;
        }

        if((!CurrentCard.DI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            if((CurrentCard.wCardType!=PCI_7300A_RevA)&&
                (CurrentCard.wCardType!=PCI_7300A_RevB)&&
                (CurrentCard.wCardType!=PCI_7350))
            {
                if(CurrentCard.DI_Status.pwBuffer){
                    memcpy(
                        CurrentCard.DI_Status.pwBuffer,
                        CurrentCard_DIFunc.pView[0],
                        CurrentCard.DI_Status.dwTransferCount*
                        (U32)((CurrentCard_DIFunc.wResoluation+7)/8)
                    );
                }
            }
            else{
                //7300A / 7300B unlock the memory for SyncMode ,
                //the AsyncMode unlock memory in DI_AsyncClear()
                munlock((const void *)pwBuffer, dwReadCount*byte_width);
                //yuan add 04/16/01
                CurrentCard.DI_Status.p7300buf = NULL;
            }
        }
    }

    return NoError;
}


/*
 * pwBuffer is of no use while double-buffer mode enabled
 */
I16 DI_ContReadPortToFile(
    U16 wCardNumber,
    U16 wPort,
    U8  *fileName,
    U32 dwReadCount,
    F64 fSampleRate,
    U16 wSyncMode
    )
{
    I16 err;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    FILE *di_file;
    void *buf = NULL;
    U32 byte_width = 0;
    U8 svFileName[128], needlock = 0;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_SG_CONT iotSCont;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotSCont, '\0', sizeof(DAS_IOT_SG_CONT));

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    //07/07/08 added for 922x
    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)){
        U32 SampIntrv = 0;
        if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 1,
            &wPort, dwReadCount*(U32)((CurrentCard_DIFunc.wResoluation+7)/8),
            TRUE))!= NoError)
        {
            return err;
        }
        //Open File to be written
        sprintf((char*)svFileName, "%s.dat", fileName);
        if((di_file = fopen((char*)svFileName, "w+b"))==NULL)
            return ErrorOpenFile;
        //Sampling Rate Boundary
        if(iotConf.mData.m922xAI.fConvSrc==0){
            if((fSampleRate==0)||(fSampleRate>2000000))
                return ErrorInvalidSampleRate;
            else
                SampIntrv = (U32)(80000000/fSampleRate);
        }
        //Set Indicated Buffer - Always from First Buffer in ToFile Mode
        if(!CurrentCard.DI_Status.bDblBufferMode){
            if(!CurrentCard.DI_Status.pDAQbuf[0])
                return ErrorInvalidBufferID;
            CurrentCard.DI_Status.pDAQbufHandle = 0;
        }
        else{
            if((!CurrentCard.DI_Status.pDAQbuf[0])||
                (!CurrentCard.DI_Status.pDAQbuf[1]))
            {
                return ErrorInvalidBufferID;
            }
            CurrentCard.DI_Status.pDAQbufHandle = 0;
        }

        CurrentCard.DI_Status.bToFile = TRUE; //ToFile Mode
        CurrentCard.DI_Status.wNumChans = 1; //Number of channel
        CurrentCard.DI_Status.dwTransferCount = dwReadCount; //Transfer Count
        CurrentCard.DI_Status.pwBuffer = di_file; //File Pointer to be Written
        byte_width = (CurrentCard_DIFunc.wResoluation+7)/8; //Byte Size

        iotSCont.wAutoScan = 0; //Not Channel Scan
        iotSCont.wSyncMode = wSyncMode; //Sync. or Async. Mode
        iotSCont.wChannelCount = 1; //Channel Count
        iotSCont.dwAccessCount = dwReadCount; //Read Count
        iotSCont.dwChannelGain[0].wChannel = wPort; //Port Number
        iotSCont.dwBufferAddr = 0; //Handled Buffer ID - Always 0 in File Mode
        iotSCont.dwPacerValue[0] = SampIntrv; //Sample Interval

        //Write File Header - Identify by FileConvertor Utility
        write_header_to_file(wCardNumber, di_file, (U16)byte_width, 0, 1,
            (U8)wPort, dwReadCount, fSampleRate, 0
        );

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CON_READ_SG, &(iotSCont))
            !=DAS_ERR_NO)
        {
            return iotSCont.wErrorCode;
        }

        //Write Data from User Memory to Disk File while SYNC. Mode
        if((!CurrentCard.DI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            write_data_to_file(
                (U16 *)CurrentCard.DI_Status.pDAQbuf[0],
                di_file,
                (U16)byte_width,
                FALSE, /*1,*/
                CurrentCard.DI_Status.dwTransferCount
            );
            fclose(di_file);
        }
    }
    else{
        if(CurrentCard.wCardType==PCI_7200)
            byte_width = (CurrentCard_DIFunc.wResoluation+7)/8;
        if(CurrentCard.wCardType==PCI_7300A_RevA)
            byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
        if(CurrentCard.wCardType==PCI_7300A_RevB)
            byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;
        if(CurrentCard.wCardType==PCI_7350)
            byte_width = iotConf.mData.m7350DIO.PortCount+1;

        if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 1,
            &wPort, dwReadCount*byte_width, TRUE))!=NoError)
        {
            return err;
        }

        if(CurrentCard.DI_Status.bDblBufferMode)
            dwReadCount /= 2;

        //yuan add 10/11/99
        sprintf((char*)svFileName, "%s.dat", fileName);
        if((di_file = fopen((char*)svFileName, "w+b"))==NULL)
            return ErrorOpenFile;

        CurrentCard.DI_Status.bToFile = TRUE;
        CurrentCard.DI_Status.dwTransferCount = dwReadCount;
        CurrentCard.DI_Status.wNumChans = 1;
        CurrentCard.DI_Status.pwBuffer = di_file;

        //06/10/09
        if((CurrentCard.wCardType!=PCI_7350)||
            ((CurrentCard.wCardType==PCI_7350)&&
            (iotConf.mData.m7350DIO.OPMode!=1)&&
            (!iotConf.mData.m7350DIO.ExtClk)))
        {
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
                return ErrorInvalidSampleRate;
        }
        else if((CurrentCard.wCardType==PCI_7350)&&
            (iotConf.mData.m7350DIO.OPMode==1))
        {
            iotCont.dwPacerValue = 2;
        }

        if(CurrentCard.wCardType==PCI_7200){
            //byte_width = (CurrentCard_DIFunc.wResoluation+7)/8;
            //Set Timer Cascade
            if(iotCont.dwPacerValue>65535)
                iotConf.mData.m7200DIO.fTimerCascade = 1;
            else
                iotConf.mData.m7200DIO.fTimerCascade = 0;

            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }

            switch(iotConf.mData.m7200DIO.wTrigSource){
                case TRIG_EXT_STROBE:
                case TRIG_HANDSHAKE:
                    fSampleRate = 0;
                    break;
                case TRIG_INT_PACER:
                default:
                    break;
            }
        }
        //allocate memory space for 7300 series
        if((CurrentCard.wCardType==PCI_7300A_RevA)||
            (CurrentCard.wCardType==PCI_7300A_RevB)||
            (CurrentCard.wCardType==PCI_7350))
        {
            //00/06/02 jeffrey added
            //if(CurrentCard.wCardType==PCI_7300A_RevA)
            //    byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
            //else
            //    byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;
            if(CurrentCard.wCardType==PCI_7350)
                buf = (void *)malloc(dwReadCount*byte_width+7);
                if(buf==NULL)
                    return ErrorAllocateMemory;
                if((U32)buf&0x7){
                    buf = (void *)(((U32)buf)+(8-((U32)buf&0x7)));
            }
            else{
                buf = (void *)malloc(dwReadCount*byte_width);
                if(buf==NULL)
                    return ErrorAllocateMemory;
            }
            //00/06/02 jeffrey , for memory lock in user level
            //lock memory with mlock()
            err = mlock((const void *)buf, dwReadCount*byte_width);
            if(err!=0){ //errno defined in errno.h
                needlock = 1;
                goto lockio;
                if(errno==ENOMEM)
                    printf("memory lock failed : errno = ENOMEM , please check the memory size\n");
                else if(errno==EPERM)
                    printf("memory lock failed : error =  EPERM, mlock system call is only for ROOT\n");
                else if(errno==EINVAL)
                    printf("memory lock failed : error = EINVAL, Invalate value\n");
                else
                    printf("memory lock failed : errno = %d\n", errno);
                return ErrorLockMemory;
            }
lockio:
            //CurrentCard.DI_Status.p7300bufHandle = hMem;
            CurrentCard.DI_Status.p7300buf = buf;
            if(CurrentCard.wCardType==PCI_7300A_RevA){
                //Set Buffer Address
                iotConf.mData.m7300DIO.dwBufferAddr = (U32)buf;
                //yuan add 11/30/99
                iotConf.mData.m7300DIO.fConfig = 0;
                if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
                    !=DAS_ERR_NO)
                {
                    munlock(
                        (const void *)buf,
                        CurrentCard.DI_Status.dwTransferCount*byte_width
                    );
                    //yuan add 04/16/01
                    CurrentCard.DI_Status.p7300buf = NULL;
                    return iotConf.wErrorCode;
                }
                switch(iotConf.mData.m7300DIO.wTrigSource){
                    case TRIG_CLK_20MHZ:
                        fSampleRate = 20000000;
                        break;
                    case TRIG_CLK_10MHZ:
                        fSampleRate = 10000000;
                        break;
                    case TRIG_EXT_STROBE:
                    case TRIG_HANDSHAKE:
                        fSampleRate = 0;
                        break;
                    case TRIG_INT_PACER:
                    default:
                        break;
                }
            }
            if(CurrentCard.wCardType==PCI_7300A_RevB){
                //Set Buffer Address
                iotConf.mData.m7300bDIO.dwBufferAddr = (U32)buf;
                //yuan add 11/30/99
                iotConf.mData.m7300bDIO.fConfig = 0;
                if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
                    !=DAS_ERR_NO)
                {
                    munlock(
                        (const void *)buf,
                        CurrentCard.DI_Status.dwTransferCount*byte_width
                    );
                    //yuan add 04/16/01
                    CurrentCard.DI_Status.p7300buf = NULL;
                    return iotConf.wErrorCode;
                }
                switch(iotConf.mData.m7300DIO.wTrigSource){
                    case TRIG_CLK_20MHZ:
                        fSampleRate = 20000000;
                        break;
                    case TRIG_CLK_10MHZ:
                        fSampleRate = 10000000;
                        break;
                    case TRIG_EXT_STROBE:
                    case TRIG_HANDSHAKE:
                        fSampleRate = 0;
                        break;
                    case TRIG_INT_PACER:
                    default:
                        break;
                }
            }
            if(CurrentCard.wCardType==PCI_7350){
                if(CurrentCard.DI_Status.bChanRange==0){
                    /*Non-Multi-Buffer*/
                    DAS_IOT_MBUF iotMbuf = {0};
                    //byte_width = iotConf.mData.m7350DIO.PortCount+1;

                    if((iotConf.mData.m7350DIO.WaitTrig==1)&&
                        (iotConf.mData.m7350DIO.WaitSTrig==0))
                    {
                        /*External Trigger*/
                        if(!CurrentCard.AI_Status.CurrentConfig.mData.dwValue[1])
                        {
                            /*Infinite Trigger*/
                            iotMbuf.dwAccessCount = dwReadCount*2;
                            CurrentCard.DI_Status.dwTransferCount *= 2;
                        }
                        else{
                            iotMbuf.dwAccessCount = dwReadCount*
                            CurrentCard.AI_Status.CurrentConfig.mData.dwValue[1];
                            CurrentCard.DI_Status.dwTransferCount *=
                            CurrentCard.AI_Status.CurrentConfig.mData.dwValue[1];
                        }
                    }
                    else
                        iotMbuf.dwAccessCount = dwReadCount;

                    /*allocate memory for DI DMA*/
                    //hMem = GlobalAlloc(GMEM_FIXED, iotMbuf.dwAccessCount*byte_width+7);
                    //buf = (PVOID)GlobalLock(hMem);
                    //if(buf==NULL)
                    //    return ErrorAllocateMemory;
                    //if((ULONG)buf&0x7){
                    //    buf = (PVOID)(((ULONG)buf)+(8-((ULONG)buf&0x7)));
                    //}
                    //GlobalFix(hMem);
                    //CurrentCard.DI_Status.p7300bufHandle = hMem;
                    //CurrentCard.DI_Status.p7300buf = buf;
                    /*Set Descriptor*/
                    iotMbuf.dwBufferAddr = (U32)buf;
                    //iotMbuf.dwAccessCount = dwReadCount;
                    iotMbuf.wBufferId = 0|0x8000; //0: DI; 1:DO; //0x8000 single buffer
                    //iotMbuf.wErrorCode = ErrorConfigIoctl;
                    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET,
                        &(iotMbuf))!=DAS_ERR_NO)
                    {
                        munlock((const void *)buf, dwReadCount*byte_width);
                        return iotMbuf.wErrorCode;
                    }
                    if(iotConf.mData.m7350DIO.ExtClk)
                        fSampleRate = 0;
                }
            }
        }

        iotCont.dwAccessCount = dwReadCount;
        iotCont.wAutoScan = 0;
        iotCont.wSyncMode = wSyncMode;
        iotCont.wChannelCount = 1;
        iotCont.dwChannelGain[0].wChannel = wPort;
        iotCont.dwChannelGain[0].wGain = needlock;
        iotCont.wIterations = 1;

        write_header_to_file(wCardNumber, di_file, byte_width, 0, 1, wPort,
            dwReadCount, fSampleRate, 0
        );

        err = ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CON_READ, &(iotCont));
        if(err!=DAS_ERR_NO){
            if((CurrentCard.wCardType==PCI_7300A_RevA)||
                (CurrentCard.wCardType==PCI_7300A_RevB)||
                (CurrentCard.wCardType==PCI_7350))
            {
                munlock(
                    (const void *)buf,
                    CurrentCard.DI_Status.dwTransferCount*byte_width
                );
                //yuan add 04/16/01
                CurrentCard.DI_Status.p7300buf = NULL;
            }
            return iotCont.wErrorCode;
        }

        if((!CurrentCard.DI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            if((CurrentCard.wCardType!=PCI_7300A_RevA)&&
                (CurrentCard.wCardType!=PCI_7300A_RevB)&&
                (CurrentCard.wCardType!=PCI_7350))
            {
                fwrite(
                    CurrentCard_DIFunc.pView[0],
                    (U32)((CurrentCard_DIFunc.wResoluation+7)/8),
                    CurrentCard.DI_Status.dwTransferCount, di_file
                );
            }
            else{ //PCI-7300A
                fwrite(
                    buf,
                    byte_width,
                    CurrentCard.DI_Status.dwTransferCount,
                    di_file
                );
                munlock(
                    (const void *)buf,
                    CurrentCard.DI_Status.dwTransferCount*byte_width
                );
                free(buf);
                //yuan add 04/16/01
                CurrentCard.DI_Status.p7300buf = NULL;
            }
            fclose(di_file);
        }
    }

    return NoError;
}


I16 DI_ContMultiBufferStart(
    U16 wCardNumber,
    U16 wPort,
    F64 fSampleRate
    )
{
    //I16 err;
    //I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    CurrentCard.DI_Status.bToFile = FALSE;

    //06/10/09
    if((CurrentCard.wCardType!=PCI_7350)||
        ((CurrentCard.wCardType==PCI_7350)&&
        (iotConf.mData.m7350DIO.OPMode!=1)&&
        (!iotConf.mData.m7350DIO.ExtClk)))
    {
        //Set PacerValue and TimerCascade
        if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
            return ErrorInvalidSampleRate;
    }
    else if((CurrentCard.wCardType==PCI_7350)&&
        (iotConf.mData.m7350DIO.OPMode==1))
    {
        iotCont.dwPacerValue = 2;
    }

    if(CurrentCard.wCardType==PCI_7200){
        //byte_width = (CurrentCard_DIFunc.wResoluation+7)/8;
        //Set Timer Cascade
        if(iotCont.dwPacerValue>65535)
            iotConf.mData.m7200DIO.fTimerCascade = 1;
        else
            iotConf.mData.m7200DIO.fTimerCascade = 0;

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
            !=DAS_ERR_NO)
        {
            return iotConf.wErrorCode;
        }
        switch(iotConf.mData.m7200DIO.wTrigSource){
            case TRIG_EXT_STROBE:
            case TRIG_HANDSHAKE:
                fSampleRate = 0;
                break;
            case TRIG_INT_PACER:
            default:
                break;
        }
    }

    //yuan 07/31/00
    iotCont.wAutoScan = 0;
    iotCont.wSyncMode = ASYNCH_OP;
    iotCont.wChannelCount = 1;
    iotCont.dwChannelGain[0].wChannel = wPort;
    iotCont.dwChannelGain[0].wGain = 0;
    iotCont.wIterations = 0;

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CON_READ, &(iotCont))
        !=DAS_ERR_NO)
    {
        return iotCont.wErrorCode;
    }

    return NoError;
}


#ifdef WHOLE
//pwBuffer is not used with double-buffer mode enabled
I16 DI_ContReadMultiPorts(
    U16  wCardNumber,
    U16  wNumPorts,
    U16  *pwPorts,
    void *pwBuffer,
    U32  dwReadCount,
    F64  fSampleRate,
    U16  wSyncMode
    )
{
/*
    I16 err;

    if((err = fundamental_check(&CurrentCardInfo.DI_Func, wCardNumber,
        wNumPorts, pwPorts, dwReadCount*
        (U32)((CurrentCard_DIFunc.wResoluation+7)/8), TRUE))!=NoError)
    {
        return err;
    }

    check pwPorts to determine if AutoScan set on
    {
        if(!CurrentCardInfo.DI_Func.fAutoScan) then no AutoScan
        if(ports do not start/end port-0 and fEndingZero) then no AutoScan
        if(ports are continuous){
            check bScanDirection to see if port ascending/decending is allowed
            if yes then AutoScan else no AutoScan
        }
        else no AutoScan
    }
    ioc=DAS_IOC_DI_CON_READ perform DI
    in :
    {
        dwAccessCnt = dwReadCount;
        wSampCntr = calculate_SampCntr(fSampleRate);
        wXferMode = CurrentCard.DI_Status.wXferMode;
        wTrigMode = CurrentCard.DI_Status.wTrigMode;
        wAutoScan = AutoScan;
        wSyncMode = wSyncMode;
        wChannelCnt = wNumPorts;
        for(i=0; i<wNumPorts; i++)
            wChannelGain[i].wChannel = pwPorts[i];
    }
*/
}
#endif


I16 DI_ReadPort(
    U16 wCardNumber,
    U16 wPort,
    U32 *pValue
    )
{
    I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    //"Port Size" is different
    //09/22/08 modified
    switch(CurrentCard.wCardType){
        case PCI_6202:
            if(wPort==0){
                //ISO DI - 16 bits
                iotPio.wByteCount = 2;
            }
            else if(wPort==1){
                //TTL - 8 bits
                iotPio.wByteCount = 1;
            }
            break;
        case PCI_7350:
            iotPio.wByteCount = 1;
            break;
        case PCI_7396:
            if(wPort<Channel_P1)
                iotPio.wByteCount = 1;
            else
                iotPio.wByteCount = (CurrentCard_DIFunc.wResoluation+7)/8;
            break;
        case PCI_7442:
            if(wPort<2){
                //DI Channel - 32 bits Port
                iotPio.wByteCount = 4;
            }
            else if(wPort<4){
                //TTL port - 16 bits Port
                iotPio.wByteCount = 2;
            }
            break;
        case PCI_7443:
            if(wPort<4){
                //DI0 ~ DI3 channel - 32 bits
                iotPio.wByteCount = 4;
            }
            else if(wPort<6){
                //TTL port - 16 bits
                iotPio.wByteCount = 2;
            }
            break;
        case PCI_7444:
            iotPio.wByteCount = 2;
            break;
        default:
            iotPio.wByteCount = (CurrentCard_DIFunc.wResoluation+7)/8;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    *pValue = iotPio.dwValue;

    return NoError;
}


I16 DI_ReadLine(
    U16 wCardNumber,
    U16 wPort,
    U16 wLine,
    U16 *pwValue
    )
{
    I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, &wPort,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    iotPio.wLine_Gain = wLine;
    iotPio.wByteCount = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    *pwValue = (U16)iotPio.dwValue;

    return NoError;
}


I16 DI_InitialMemoryAllocated(
    U16 wCardNumber,
    U32 *MemSize
    )
{
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    //06/02/09 add for 7350
    if(CurrentCard.wCardType==PCI_7350){
        if(CurrentCard_DIFunc.dwBufferSize)
            *MemSize = (CurrentCard_DIFunc.dwBufferSize-128*20)/5;
        else
            *MemSize = 0;
    }
    else{
        *MemSize = CurrentCard_DIFunc.dwBufferSize/1024; //in KB
        if((CurrentCard.wCardType==PCI_7300A_RevA)||
            (CurrentCard.wCardType==PCI_7300A_RevB)||
            (CurrentCard.wCardType==PCI_9222)||
            (CurrentCard.wCardType==PCI_9223))
        {
            //07/07/08 modified
            if(*MemSize)
                *MemSize = (*MemSize-4)*256;
        }
    }

    return NoError;
}


I16 DI_ContStatus(
    U16 wCardNumber,
    U16 *Status
    )
{
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_STATUS iotStat;

    memset(&iotStat, '\0', sizeof(DAS_IOT_STATUS));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CON_STATUS, &(iotStat))
        !=DAS_ERR_NO)
    {
        return iotStat.wErrorCode;
    }
    *Status = iotStat.mStat.wValue;

    return NoError;
}


I16 write_header_to_file(
    U16  wCardNumber,
    FILE *fp,
    U16  data_width,
    U16  ad_range,
    U16  num_chan,
    U8   channel_no,
    U32  count,
    F64  fSampleRate,
    U16  num_of_channel_range
    )
{
    time_t cur_time;
    struct tm st_time;
    DAQFileHeader fheader;
    char ltime[9];
    char ldate[9];
    char tmpms[4]="000";

    strncpy(fheader.ID, "ADLinkDAQ1", 10);
    fheader.card_type = CurrentCard.wCardType;
    fheader.num_of_channel = num_chan;
    fheader.channel_no = channel_no;
    //00/08/02 jeffrey,
    //Check for 9812 , PCI-9812 can sample all 4-channel in parallel
    fheader.num_of_scan = count/num_chan;
    switch(data_width){
        case 1:
            fheader.data_width =  0;
            break;
        case 2:
            fheader.data_width =  1;
            break;
        case 4:
            fheader.data_width =  2;
            break;
        default:
            fheader.data_width =  1;
    }
    //02/16/09 add for 7350 24 bits
    if((CurrentCard.wCardType==PCI_7350)&&(data_width==3))
        fheader.data_width = 3;
    fheader.channel_order = 0;
    if(CurrentCard.wCardType==PCI_9112)
        fheader.channel_order = 1;
    if(num_of_channel_range)
        fheader.channel_order = 2;
    fheader.ad_range = ad_range;

    //00/02/17 jeffrey,
    //Check for 9812 , PCI-9812 can sample all 4-channel in parallel
    if((CurrentCard.wCardType==PCI_9812)||
        (CurrentCard.wCardType==PCI_9810)||
        (CurrentCard.wCardType==PCI_9221)||
        (CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        fheader.scan_rate = fSampleRate;
    }
    else
        fheader.scan_rate = fSampleRate/num_chan;

    fheader.num_of_channel_range = num_of_channel_range;
    time(&cur_time);
    memcpy(&st_time, localtime(&cur_time) , sizeof(struct tm));
    sprintf(ldate, "%02d/%02d/%02d", st_time.tm_mon+1, st_time.tm_mday,
        st_time.tm_year-100 //based on 2000
    );
    sprintf(ltime, "%02d:%02d:%02d", st_time.tm_hour, st_time.tm_min,
        st_time.tm_sec
    );

    strncpy(fheader.start_date, ldate, 8);
    strncpy(fheader.start_time, ltime, 8);
    strncpy(fheader.start_millisec, tmpms, 3);
    strncpy(fheader.reserved, "", 6);
    //strncpy((void *)(temp_header+HEAD_RESERVED), "", 6);
    fwrite(&fheader, sizeof(DAQFileHeader), 1, fp);

    return 0;
}


I16 DI_EventCallBack(
    U16  wCardNumber,
    I16  mode,
    I16  EventType,
    void (*callbackAddr)(int)
    )
{
    DAS_IOT_DUALINT iotDint;

    memset(&iotDint, '\0', sizeof(DAS_IOT_DUALINT));
    /*
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }
    */
    if(mode==1)
        iotDint.bInt1Enable = 1;
    else
        iotDint.bInt1Enable = 0;

    iotDint.bInt1Mode = EventType;

    switch(EventType){
    case 0:
        if(mode>0){ //enable the INT1
            iotDint.wErrorCode = SIGEVENT3;
            if(callbackAddr!=NULL){
                if(CurrentCard.i_sig_act.sa_handler==SIG_IGN){
                    iotDint.bInt1Status = TRUE; //fasync_helper is needed
                    CurrentCard.i_sig_act.sa_handler = callbackAddr;
                    CurrentCard.i_sig_act.sa_flags |= SA_RESTART;
                    if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL )==-1){
                        return ErrorOpenEventFailed;
                    }
                }
                else{ //signal_handle is asigned , so ignore the signal setting
                    iotDint.bInt1Status = FALSE; //ignore this signal setting
                }
            }
            else{ //only ENABLE INT , no signal is needed
                iotDint.bInt1Status = FALSE; //ignore this signal setting
            }
        }
        else{ //disable the INT1
            //reset the signal handler
            if(CurrentCard.i_sig_act.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.i_sig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGEVENT3, &(CurrentCard.i_sig_act), NULL)==-1){
                    //ignore the signal SIGEVENT3
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE; //ignore this signal setting
            }
        }
        break;
    case 1:
        iotDint.wErrorCode = SIGDBI;
        if(mode>0){ //enable the INT1
            if(callbackAddr!=NULL){
                if(CurrentCard.i_dbsig_act.sa_handler==SIG_IGN){
                    iotDint.bInt1Status = TRUE; //fasync_helper is needed
                    CurrentCard.i_dbsig_act.sa_handler = callbackAddr;
                    CurrentCard.i_dbsig_act.sa_flags |= SA_RESTART;
                    if(sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL)==-1)
                        return ErrorOpenEventFailed;
                }
                else{ //signal_handle is asigned , so ignore the signal setting
                    iotDint.bInt1Status = FALSE; //ignore this signal setting
                }
            }
            else{ //only ENABLE INT , no signal is needed
                iotDint.bInt1Status = FALSE; //ignore this signal setting
            }
        }
        else{ //disable the INT1
            //reset the signal handler
            if(CurrentCard.i_dbsig_act.sa_handler!=SIG_IGN){ //
                iotDint.bInt1Status = TRUE; //fasync_helper is needed
                CurrentCard.i_dbsig_act.sa_handler = SIG_IGN;
                if(sigaction(SIGDBI, &(CurrentCard.i_dbsig_act), NULL)==-1){
                    //ignore the signal SIGEVENT3
                    return ErrorOpenEventFailed;
                }
            }
            else{ //signal_handle is asigned , so ignore the signal setting
                iotDint.bInt1Status = FALSE; //ignore this signal setting
            }
        }
        break;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_SIGNAL_SET, &iotDint)<0){
        return iotDint.wErrorCode;
    }

    return NoError;
}


I16 DI_AsyncDblBufferOverrun(
    U16 wCardNumber,
    U16 op,
    U16 *overrunFlag
    )
{
    I16 err;
    //volatile PPCI_DBXORUN pOvRun;

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    //yuan add 09/11/02
    if(CurrentCard.wCardType==PCI_7200){
        volatile PPCI_DBXORUN pOvRun;
        pOvRun = (volatile PPCI_DBXORUN)(CurrentCard_AOFunc.pView[0]);
        switch(op){
            case 0: //check
                *overrunFlag = pOvRun->xFlag.f.overrun;
                break;
            case 1: //clear
                memset(pOvRun, '\0', sizeof(PCI_DBXORUN));
                break;
        }
    }
    else if(CurrentCard.wCardType==PCI_7300A_RevB){
        volatile PPCI7300_DBXORUN pOvRun;
        pOvRun = (volatile PPCI7300_DBXORUN)(CurrentCard_AOFunc.pView[0]);
        switch(op){
            case 0: //check
                *overrunFlag = pOvRun->xFlag.f.overrun;
                break;
            case 1: //clear
                if(pOvRun->xFlag.f.extended==1){
                    memset(pOvRun, '\0',
                        (sizeof(PCI7300_DBXORUN)+sizeof(U32)*(MAXMBUFCNT-8))
                    );
                    pOvRun->xFlag.f.extended = 1;
                }
                else
                    memset(pOvRun, '\0', sizeof(PCI7300_DBXORUN));
                break;
        }
    }
    else if((CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223))
    {
        volatile PPCI_DBXORUN pOvRun;
        //OvRun[2] - for DI (9222/9223)
        pOvRun = \
            &(((DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const)->OvRun[2]);
        switch(op){
            case 0:
                //Check - Get Overrun Flag from Kernel to Check if Overrun
                *overrunFlag = pOvRun->xFlag.f.overrun;
                break;
            case 1:
                //Clear - Clear all Contents of Overrun Structure
                memset(pOvRun, '\0', sizeof(PCI_DBXORUN));
                break;
        }
    }
    else if(CurrentCard.wCardType==PCI_7350){
        if(CurrentCard.DI_Status.bChanRange){
            volatile PPCI7350_DBXORUN pOvRun;
            pOvRun = \
                &(((DAS_IOT_CALCONST_7350 *)CurrentCard.pCAL_Const)->OvRun[0]);
            switch(op){
                case 0:
                    //Check - Get Overrun Flag from Kernel to Check if Overrun
                    *overrunFlag = pOvRun->xFlag.f.overrun;
                    break;
                case 1:
                    //Clear - Clear all Contents of Overrun Structure
                    memset(pOvRun, '\0', sizeof(PCI7350_DBXORUN));
                    break;
            }
        }
    }

    return NoError;
}


I16 di_922x_config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTriggerCnt,
    BOOLEAN AutoResetBuf
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }
    if((((TrigCtrl>>4)&0xf)==0)&&((TrigCtrl>>9)&0x1)){
        //Soft Trigger Source & ReTrigger
        return ErrorInvalidTriggerMode;
    }

    /*Conversion source*/
    iotConf.mData.m922xAI.fConvSrc = ConfigCtrl&0xf;
    /*Trigger source*/
    iotConf.mData.m922xAI.fTrigSrc = (TrigCtrl>>4)&0xf;
    /*Trigger polarity*/
    iotConf.mData.m922xAI.fTrigPol = (TrigCtrl>>8)&0x1;
    /*Enable ReTrigger*/
    iotConf.mData.m922xAI.fEnReTrig = (TrigCtrl>>9)&0x1;
    /*Auto reset buffer?*/
    iotConf.mData.m922xAI.fAutoResetBuf = AutoResetBuf;
    if(iotConf.mData.m922xAI.fEnReTrig)
        iotConf.mData.m922xAI.ReTriggerCnt = ReTriggerCnt;
    //iotConf.wErrorCode = ErrorConfigIoctl;
    CurrentCard.DI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    //CurrentCard.DI_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DI_9222_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTriggerCnt,
    BOOLEAN AutoResetBuf
    )
{
    return di_922x_config(wCardNumber, ConfigCtrl, TrigCtrl, ReTriggerCnt,
        AutoResetBuf);
}


I16 DI_9223_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTriggerCnt,
    BOOLEAN AutoResetBuf
    )
{
    return di_922x_config(wCardNumber, ConfigCtrl, TrigCtrl, ReTriggerCnt,
        AutoResetBuf);
}


I16 DI_AsyncDblBufferHandled(
    U16 wCardNumber
    )
{
    I16 err;
    volatile PPCI_DBXORUN pOvRun;

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(!CurrentCard.DI_Status.bDblBufferMode)
        return ErrorNotDoubleBufferMode;

    switch(CurrentCard.wCardType){
        case PCI_9222:
        case PCI_9223:
            //01/26/08
            pOvRun = \
                &(((DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const)->OvRun[2]);
            break;
        default:
            return ErrorFuncNotSupport;
    }
    if(pOvRun){
        if(!pOvRun->xFlag.f.overrun){
            pOvRun->xFlag.f.lock = 1;
            if(pOvRun->dwCtr_isr[CurrentCard.DI_Status.wDblBufferIndex]>=1)
                pOvRun->dwCtr_isr[CurrentCard.DI_Status.wDblBufferIndex] -= 1;
            pOvRun->xFlag.f.lock = 0;
        }
    }

    return NoError;
}


I16 DI_AsyncDblBufferToFile(
    U16 wCardNumber
    )
{
    I16 err;
    volatile PPCI_DBXORUN pOvRun;
    U16 wBuffIdx;

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    //07/31/08 removed
    //if(!CurrentCard.DI_Status.bDblBufferMode)
    //    return ErrorNotDoubleBufferMode;

    CurrentCard.DI_Status.wDblBufferIndex &= 0x01;
    wBuffIdx = CurrentCard.DI_Status.wDblBufferIndex;
    //Write Data to File
    write_data_to_file(
        (U16 *)CurrentCard.DI_Status.pDAQbuf[wBuffIdx], //Data Buffer
        (FILE *)CurrentCard.DI_Status.pwBuffer, //File pointer
        (U16)((CurrentCard_DIFunc.wResoluation+7)/8), //Byte Count per Sample
        FALSE,
        CurrentCard.DI_Status.dwTransferCount //Write Count to File
    );

    //Let Kernel Know that the Buffer is Handled
    switch(CurrentCard.wCardType){
        case PCI_9222:
        case PCI_9223:
            pOvRun = \
                &(((DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const)->OvRun[2]);
            break;
        default:
            return ErrorFuncNotSupport;
    }
    if(pOvRun){
        if(!pOvRun->xFlag.f.overrun){
            pOvRun->xFlag.f.lock = 1;
            if(pOvRun->dwCtr_isr[wBuffIdx]>=1)
                pOvRun->dwCtr_isr[wBuffIdx] -= 1;
            pOvRun->xFlag.f.lock = 0;
        }
    }

    return NoError;
}


I16 DI_SetTimeOut(
    U16 wCardNumber,
    U32 dwTimeOut
    )
{
    I16 err;
    DAS_IOT_SI iotSI;

    memset(&iotSI, '\0', sizeof(DAS_IOT_SI));

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    /*
     * Setup Structure to be Passed to Kernel
     */
    //TimeOut Period(ms) for DI Sync. Mode
    iotSI.dwScanInterval = dwTimeOut;
    iotSI.dwSampInterval = 0;
    //iotSI.wErrorCode = ErrorConfigIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_TIMEOUT_SET, &(iotSI))
        !=DAS_ERR_NO)
    {
        return iotSI.wErrorCode;
    }

    return NoError;
}


I16 DI_ContBufferSetup(
    U16  wCardNumber,
    void *pwBuffer,
    U32  dwReadCount,
    U16  *BufferId
    )
{
    DAS_IOT_MBUF2 iotMbuf = {0};
    I16 err;

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        dwReadCount, TRUE))!=NoError)
    {
        return err;
    }

    iotMbuf.dwBufferAddr = (U32)pwBuffer; //Buffer Address to be Setup
    iotMbuf.dwAccessCount = dwReadCount; //Buffer Size to be Setup
    iotMbuf.wDir = 2; //2: Let Kernel Know that DI Buffer to be Setup
    iotMbuf.wBufferId = 1; //1: Lock Memory - Compatiable for Linux Buffer Set
    //iotMbuf.wErrorCode = ErrorConfigIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_SET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }
    *BufferId = iotMbuf.wBufferId; //Buffer ID Got from Kernal Mode
    //Set Buffer Structure
    CurrentCard.DI_Status.pDAQbuf[*BufferId] = pwBuffer;
    CurrentCard.DI_Status.dwBufCount[*BufferId] = dwReadCount;

    return NoError;
}


I16 DI_ContBufferReset(
    U16 wCardNumber
    )
{
    DAS_IOT_MBUF2 iotMbuf = {0};
    I16 err;

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotMbuf.wDir = 2; //2: Let Kernel Know that DI Buffer to be Reset
    //iotMbuf.wErrorCode = ErrorConfigIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_SET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }
    //Reset DI Buffer Structure to Default
    CurrentCard.DI_Status.pDAQbuf[0] = NULL;
    CurrentCard.DI_Status.pDAQbuf[1] = NULL;
    CurrentCard.DI_Status.dwBufCount[0] = 0;
    CurrentCard.DI_Status.dwBufCount[1] = 0;

    return NoError;
}


I16 DI_AsyncReTrigNextReady(
    U16     wCardNumber,
    BOOLEAN *bReady,
    BOOLEAN *StopFlag,
    U16     *RdyTrigCnt
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf = {0};
    DAS_IOT_CONFIG iotConf;
    U8 fInfiniteTrigger = 0;

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }
    //02/25/09 for 7350
    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)){
        if(!iotConf.mData.m922xAI.fEnReTrig)
            return ErrorInvalidTriggerMode;
        if(iotConf.mData.m922xAI.ReTriggerCnt==0)
            fInfiniteTrigger = 1;
    }
    else if(CurrentCard.wCardType==PCI_7350){
        if(!iotConf.mData.m7350DIO.fReTrig)
            return ErrorInvalidTriggerMode;
        if(iotConf.mData.m7350DIO.fInfReTrig)
            fInfiniteTrigger = 1;
    }
    else
        return ErrorFuncNotSupport;

    /*
     * Setup Structure to be Passed to Kernel
     */
    //iotDbf.wErrorCode = ErrorDBHalfReadyIoctl;
    //Let Kernel Know DI_AsyncReTrigNextReady will be performed
    iotDbf.wEnable = 1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bReady = (BOOLEAN)(iotDbf.wReady&0x1);
    *StopFlag = (BOOLEAN)((iotDbf.wReady>>1)&0x1);

    /*
     * Return trigger count or buffer index
     */
    //02/25/09 for 7350
    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)){
        if(fInfiniteTrigger){
            //Infinit Trigger - Return Half Ready Buffer Index
            *RdyTrigCnt = iotDbf.wViewIndex;
            CurrentCard.DI_Status.wDblBufferIndex = iotDbf.wViewIndex;
        }
        else{
            //Finite Trigger - Return Comming Trigger Count
            *RdyTrigCnt = iotDbf.wViewIndex|(iotDbf.wEnable<<16);
        }
    }
    else if(CurrentCard.wCardType==PCI_7350){
        if(fInfiniteTrigger){
            //Infinit Trigger - Return Half Ready Buffer Index
            *RdyTrigCnt = ((iotDbf.wViewIndex|(iotDbf.wEnable<<16))%2)? 0:1;
            CurrentCard.DI_Status.wDblBufferIndex = *RdyTrigCnt;
        }
        else{
            //Finite Trigger - Return Comming Trigger Count
            *RdyTrigCnt = iotDbf.wViewIndex|(iotDbf.wEnable<<16);
            CurrentCard.DI_Status.wDblBufferIndex = *RdyTrigCnt;
        }
    }

    return NoError;
}


I16 DI_7350_Config(
    U16 wCardNumber,
    U16 wDIPortWidth,
    U16 wDIMode,
    U16 wDIWaitStatus,
    U16 wDIClkConfig
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if((wDIPortWidth!=8)&&(wDIPortWidth!=16)
        &&(wDIPortWidth!=24)&&(wDIPortWidth!=32))
    {
        return ErrorInvalidPortWidth;
    }
    else{
        U32 vi;
        U16 DIPortCount = 0;
        for(vi=0; vi<4; vi++){
            /*Check DIO Port Directions*/
            if(((iotConf.mData.m7350DIO.PortConfig>>vi)&0x1)==0){
                /*Port vi is enabled*/
                if(((iotConf.mData.m7350DIO.PortConfig>>(4+vi))&0x1)==0){
                    /*Port vi is configured as DI*/
                    DIPortCount++;
                }
            }
        }
        if(wDIPortWidth!=DIPortCount*8)
            return ErrorDIODataWidthError;
    }

    if(wDIMode>P7350_BurstHandShake)
        return ErrorInvalidTriggerMode;

    if(wDIWaitStatus>P7350_WAIT_SOFTTRG)
        return ErrorInvalidTriggerType;

    /*
     * Setup Structure to be Passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know DO_7350_Config will be performed
    iotConf.mData.m7350DIO.ConfigMode = 0;
    /*DI Port Width*/
    //0->8bits; 1->16bits; 2->24bits; 3->32bits
    iotConf.mData.m7350DIO.PortCount = wDIPortWidth/8-1;
    /*DI Mode*/
    iotConf.mData.m7350DIO.OPMode = wDIMode;
    /*Trigger Status*/
    if(wDIWaitStatus==P7350_WAIT_SOFTTRG){
        iotConf.mData.m7350DIO.WaitTrig = 1;
        iotConf.mData.m7350DIO.WaitSTrig = 1;
    }
    else if(wDIWaitStatus==P7350_WAIT_EXTTRG){
        iotConf.mData.m7350DIO.WaitTrig = 1;
        iotConf.mData.m7350DIO.WaitSTrig = 0;
    }
    else{
        iotConf.mData.m7350DIO.WaitTrig = 0;
        iotConf.mData.m7350DIO.WaitSTrig = 0;
    }
    /*Sampled Clock*/
    iotConf.mData.m7350DIO.ExtClk = wDIClkConfig&0x1;
    iotConf.mData.m7350DIO.ClkEdge = (wDIClkConfig>>1)&0x1;
    //04/23/09
    /*Export Sampled Clock?*/
    iotConf.mData.m7350DIO.en_expClk = (wDIClkConfig>>2)&0x1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DI_Status.CurrentConfig = iotConf;
    //CurrentCard.DI_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DI_7350_TrigHSConfig(
    U16 wCardNumber,
    U16 wTrigConfig,
    U16 wDI_IPOL,
    U16 wDI_REQSrc,
    U16 wDI_ACKSrc,
    U16 wDI_TRIGSrc,
    U16 wStartTrigSrc,
    U16 wPauseTrigSrc,
    U16 wSoftTrigOutSrc,
    U32 dwSoftTrigOutLength,
    U32 dwTrigCount
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if((wDI_REQSrc>P7350_AFI_7)||
        (wDI_ACKSrc>P7350_AFI_7)||
        (wDI_TRIGSrc>P7350_AFI_7)||
        (wStartTrigSrc>P7350_AFI_7)||
        (wPauseTrigSrc>P7350_AFI_7)||
        (wSoftTrigOutSrc>P7350_AFI_7))
    {
        return ErrorInvalidDioPort;
    }

    /*
     * Setup Structure to be Passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know DI_7350_TrigHSConfig will be performed
    iotConf.mData.m7350DIO.ConfigMode = 1;
    /*Trigger Configuration*/
    iotConf.mData.m7350DIO.PauseTrig = wTrigConfig&0x1;
    //Only Valid for Waiting Software Trigger
    iotConf.mData.m7350DIO.SoftTrigOut = (wTrigConfig>>1)&0x1;
    /*Trigger and HS Polarity*/
    iotConf.mData.m7350DIO.HSREQPol = wDI_IPOL&0x1;
    iotConf.mData.m7350DIO.HSACKPol = (wDI_IPOL>>1)&0x1;
    iotConf.mData.m7350DIO.HSTRIGPol = (wDI_IPOL>>2)&0x1;
    iotConf.mData.m7350DIO.StartTrigPol = (wDI_IPOL>>3)&0x1;
    iotConf.mData.m7350DIO.PauseTrigPol = (wDI_IPOL>>4)&0x1;
    /*Trigger and HS Source*/
    iotConf.mData.m7350DIO.HSREQSrc = wDI_REQSrc;
    iotConf.mData.m7350DIO.HSACKSrc = wDI_ACKSrc;
    iotConf.mData.m7350DIO.HSTRIGSrc = wDI_TRIGSrc;
    iotConf.mData.m7350DIO.StartTrigSrc = wStartTrigSrc;
    iotConf.mData.m7350DIO.PauseTrigSrc = wPauseTrigSrc;
    iotConf.mData.m7350DIO.SoftTrigOutSrc = wSoftTrigOutSrc;

    //Solve the issue of the size of the configuration structure is too small
    {
        //Get the Mapped IO Memory
        U32 LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);
        *((U32 *)(LcrAddr+0x110)) = dwSoftTrigOutLength;
        *((U32 *)(LcrAddr+0xE8)) = dwTrigCount;
        /*borrow from ai structure to store them*/
        CurrentCard.AI_Status.CurrentConfig.mData.dwValue[0] = \
            dwSoftTrigOutLength;
        CurrentCard.AI_Status.CurrentConfig.mData.dwValue[1] = \
            dwTrigCount;
    }

    //02/25/09
    if((iotConf.mData.m7350DIO.WaitTrig)&&(!iotConf.mData.m7350DIO.WaitSTrig)){
        /*Wait Hardware Trigger*/
        if(dwTrigCount==0){
            /*Infinite Triggers*/
            iotConf.mData.m7350DIO.fReTrig = 1;
            iotConf.mData.m7350DIO.fInfReTrig = 1;
        }
        else if(dwTrigCount==1){
            /*One Trigger*/
            iotConf.mData.m7350DIO.fReTrig = 0;
            iotConf.mData.m7350DIO.fInfReTrig = 0;
        }
        else{
            /*N Triggers*/
            iotConf.mData.m7350DIO.fReTrig = 1;
            iotConf.mData.m7350DIO.fInfReTrig = 0;
        }
    }
    else{
        iotConf.mData.m7350DIO.fReTrig = 0;
        iotConf.mData.m7350DIO.fInfReTrig = 0;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DI_Status.CurrentConfig = iotConf;
    //CurrentCard.DI_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DI_7350_ExtSampCLKConfig(
    U16 wCardNumber,
    U16 wCLK_Src,
    U16 wCLK_DDAMode,
    U16 wCLK_DPAMode,
    U16 wCLK_DDAVlaue,
    U16 wCLK_DPAVlaue
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if(!iotConf.mData.m7350DIO.ExtClk)
        return ErrorFuncNotSupport;

    if((wCLK_Src!=P7350_ECLK_IN)&&(wCLK_Src!=P7350_AFI_7))
        return ErrorInvalidDioPort;

    if(wCLK_DDAMode&0x1){
        if(wCLK_DDAVlaue>P7350_DDA_1R04NS)
            return ErrorUndefinedParameter;
    }
    if(wCLK_DPAMode&0x1){
        if(wCLK_DPAVlaue>P7350_DPA_337R5DG)
            return ErrorUndefinedParameter;
    }

    /*
     * Setup Structure to be passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know that DI_7350_ExtSampCLKConfig will be performed
    iotConf.mData.m7350DIO.ConfigMode = 2;
    iotConf.mData.m7350DIO.ExtClkSrc = (wCLK_Src==P7350_ECLK_IN)? 0:1;
    iotConf.mData.m7350DIO.DDAMode = wCLK_DDAMode;
    iotConf.mData.m7350DIO.DPAMode = wCLK_DPAMode;
    iotConf.mData.m7350DIO.DDAValue = wCLK_DDAVlaue;
    iotConf.mData.m7350DIO.DPAValue = wCLK_DPAVlaue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DI_Status.CurrentConfig = iotConf;
    //CurrentCard.DI_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DI_7350_ExportSampCLKConfig(
    U16 wCardNumber,
    U16 wCLK_Src,
    U16 wCLK_DPAMode,
    U16 wCLK_DPAVlaue
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    //04/23/09
    if(!iotConf.mData.m7350DIO.en_expClk)
        return ErrorFuncNotSupport;

    if((wCLK_Src!=P7350_ECLK_OUT)&&
        (wCLK_Src!=P7350_AFI_7)&&
        (wCLK_Src!=(P7350_ECLK_OUT|P7350_AFI_7)))
    {
        return ErrorInvalidDioPort;
    }

    if(wCLK_DPAMode&0x1){
        if(wCLK_DPAVlaue>P7350_DPA_337R5DG)
            return ErrorUndefinedParameter;
    }

    /*
     * Set Structure to be Passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know that DI_7350_ExportSampCLKConfig will be performed
    iotConf.mData.m7350DIO.ConfigMode = 3;
    if(wCLK_Src==P7350_ECLK_OUT){
        iotConf.mData.m7350DIO.expClkSrc = 1;
    }
    else if(wCLK_Src==P7350_AFI_7){
        iotConf.mData.m7350DIO.expClkSrc = 2;
    }
    else{
        //Both
        iotConf.mData.m7350DIO.expClkSrc = 3;
    }
    iotConf.mData.m7350DIO.eDPAMode = wCLK_DPAMode;
    iotConf.mData.m7350DIO.eDPAValue = wCLK_DPAVlaue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DI_Status.CurrentConfig = iotConf;
    //CurrentCard.DI_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DI_7350_SoftTriggerGen(
    U16 wCardNumber
    )
{
    I16 err;
    U32 LcrAddr;
    U32 RegData;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if(!iotConf.mData.m7350DIO.WaitSTrig)
        return ErrorInvalidTriggerType;

    LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);

    /*
     * Set DI Soft Trigger
     * Should mask DI_WAIT_TRIG (bit4) while setting software trigger (bit6)
     */
    //05/06/09 remove mask
    RegData = *((volatile U32 *)(LcrAddr+0x108));
    //RegData &= 0xffffffefL;
    RegData |= 0x40L;
    *((U32 *)(LcrAddr+0x108)) = RegData;
    //RegData &= ~(0x1L<<6);
    //*((ULONG *)(LcrAddr+0x108)) = RegData;

    return NoError;
}


/*
 * 08/26/09
 * Add to avoid burst hand shake issue
 */
I16 DI_7350_BurstHandShakeDelay(
    U16 wCardNumber,
    U8  bDelay
    )
{
    I16 err;
    U32 LcrAddr;
    U32 RegData;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DI_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if(iotConf.mData.m7350DIO.OPMode!=P7350_BurstHandShake)
        return ErrorInvalisOperationMode;

    LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);

    /*
     * Set DI burst hand shake delay count
     * offset: 0x104 - bit 16 ~ 23
     */
    RegData = *((volatile U32 *)(LcrAddr+0x104));
    RegData &= 0xff00ffffL;
    RegData |= ((bDelay<<16)&0x00ff0000);
    *((U32 *)(LcrAddr+0x104)) = RegData;

    return NoError;
}
